const fs4 = require('fs-extra');
//const {VertexAI} = require('@google-cloud/vertexai');
const Anthropic = require('@anthropic-ai/sdk');
const Groq = require("groq-sdk");
const OpenAI = require("openai");

class OpenAIError extends Error {
  constructor(message, type, param, code) {
    super(message);
    this.name = 'OpenAIError';
    this.type = type;
    this.param = param;
    this.code = code;
  }
}

var escapeRegExp = function(string) {
    return string.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1");
};

var replaceAll = function(string, find, replace) {
  return string.replace(new RegExp(escapeRegExp(find), 'g'), replace);
};

const escapeMyHtml = (unsafe) => {
    //do not escape now????
    //return unsafe;
    return unsafe.replaceAll('&', '&amp;').replaceAll('<', '&lt;').replaceAll('>', '&gt;').replaceAll('"', '&quot;').replaceAll("'", '&#039;');
};

const unescapeHtml = (unsafe) => {

    return unsafe.replaceAll('&amp;', '&').replaceAll('&lt;', '<').replaceAll('&gt;', '>').replaceAll('&quot;', '"').replaceAll('&#039;', "'");
};
    //faster testing
    //modelVersion = 'gpt-3.5-turbo';
    //gpt-3.5-turbo-16k
    //gpt-4
    //gpt-4-1106-preview
    //gpt-4-vision-preview
    
var CHATGPT = {
  LLM_PROVIDER: 'anthropic',
  LLM_MODEL: '',
  AVAILABLE_LLMS: ['anthropic', 'openai'],
  OPENAI_KEY: '',
  CUSTOM_OPENAI_KEY: 'XXX',
  OPENAI_ORG: '',  
  ANTHROPIC_KEY: '',
  MISTRAL_KEY: '',
  GEMINI_KEY: '',
  GROQ_KEY : '',
  DB_LOGGING: false,
  DEXIE_LOGGING: true,
  dexieLogger: null,
  FILE_LOGGING: false,
  NO_CACHE_REQUESTS : ['request_file_descriptions', 'single_file_full_change', 'single_file_partial_change', 'new_code_placement', 'single_function_full_change', 'prune_unlikely_project_files'],
  SHORT_TOKENS: ['request_file_descriptions', 'request_relevant_stub_lines', 'request_fix_stub_lines', 'request_relevant_files', 'formalize_code_changes', 'prune_unlikely_project_files'],
  openai: null,
  vertex: null,
  anthropic: null,
  mistral: null,
  did_init: false,
  inCodeBlock : false,
  inTextBlock : false,
  bufferText: "",
  htmlText: "",
  lastProcessTime : 0,
  processingInterval : 200,   
  lastLanguage: "javascript",
  codeblockIndex: 0,
  codeblockData: new Array(),
  codeblockFilenames: new Array(),  
  codeLength : 0,
  max_tokens: 3000,
  request_timeout: 900000,
  lastRequest: "",
  retryAttempts : 2,
  //put these in object for the request to have multiple requests?
  relevantContextJson: new Array(),
  searchTermArr: new Array(),
  stubFilesArr: new Array(),
  originalTaskText: "",  
  useInsertAlgorithm: true,  
  //gpt-4-0613 gpt-4-1106-preview gpt-3.5-turbo-1106  gpt-4-32k-0613 gpt-4-0314
  model_properties: {
      'request_relevant_files' : {
          'model': 'gpt-4-turbo',
          'tokens': 4096,
          'use_tokens': false
      },
      'request_relevant_line_numbers' : {
          'model': 'gpt-4-turbo',
          'tokens': 4096,
          'use_tokens': false
      },
      'request_relevant_stub_lines' : {
          'model': 'gpt-4-turbo',
          'tokens': 4096,
          'use_tokens': false          
      },
      'do_new_programming_task' : {
          'model': 'gpt-4-turbo',
          'tokens': 4096,
          'use_tokens': false
      },
      'do_additional_programming_task' : {
          'model': 'gpt-4-turbo',
          'tokens': 4096,
          'use_tokens': false
      },      
      'formalize_code_changes' : {
          'model': 'gpt-4-turbo',
          'tokens': 2048,
          'use_tokens': false
      },          
      'request_file_descriptions' : { 'model': 'gpt-4-turbo', 'tokens': 4096}      
  },
  anthropic_models: {
    'models': 'claude-3-5-sonnet-20240620,claude-3-haiku-20240307,claude-3-sonnet-20240229,claude-3-opus-20240229',
    't': "4096 max, 2048 medium"
  },  
  mistral_model_properties: {
      'request_relevant_files' : {
          'model': 'codestral-latest',
          'tokens': 4096,
          'use_tokens': false
      },
      'request_relevant_line_numbers' : {
          'model': 'codestral-latest',
          'tokens': 4096,
          'use_tokens': false
      },
      'request_relevant_stub_lines' : {
          'model': 'codestral-latest',
          'tokens': 4096,
          'use_tokens': false
      },      
      'do_new_programming_task' : {
          'model': 'codestral-latest',
          'tokens': 4096,
          'use_tokens' : true
      },
      'do_additional_programming_task' : {
          'model': 'codestral-latest',
          'tokens': 4096,
          'use_tokens' : true
      },      
      'formalize_code_changes' : {
          'model': 'codestral-latest',
          'tokens': 2048,
          'use_tokens' : true
      },    
      'request_file_descriptions' : { 'model': 'mistral-large-latest', 'tokens': 4096}
  },
  groq_model_properties: {
      'request_relevant_files' : {
          'model': 'gemma-7b-it',
          'tokens': 4096
      },
      'request_relevant_line_numbers' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 4096
      },
      'request_relevant_stub_lines' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 4096
      },      
      'do_new_programming_task' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 4096
      },
      'do_additional_programming_task' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 4096
      },      
      'formalize_code_changes' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 2048
      }, 
      'request_file_descriptions' : { 'model': 'mixtral-8x7b-32768', 'tokens': 4096}
  },  
  gemini_model_properties: {
      'request_relevant_files' : {
          'model': 'gemma-7b-it',
          'tokens': 4096
      },
      'request_relevant_line_numbers' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 4096
      },
      'request_relevant_stub_lines' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 4096
      },      
      'do_new_programming_task' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 4096
      },
      'do_additional_programming_task' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 4096
      },      
      'formalize_code_changes' : {
          'model': 'mixtral-8x7b-32768',
          'tokens': 2048
      }, 
      'request_file_descriptions' : { 'model': 'mixtral-8x7b-32768', 'tokens': 4096}
  },  
  recommended_model_properties: {
      'request_relevant_files' : ['claude-3-5-sonnet-20240620', 'llama3-70b-8192', 'claude-3-haiku-20240307', 'gpt-4-turbo', 'gemini-1.5-flash-latest', 'codestral-latest'],
      'request_relevant_line_numbers' : ['claude-3-5-sonnet-20240620', 'gpt-4o', 'claude-3-sonnet-20240229', 'gpt-4-turbo', 'gemini-1.5-flash-latest', 'codestral-latest', 'llama3-70b-8192'],
      'request_relevant_stub_lines' : ['claude-3-5-sonnet-20240620', 'gpt-4o', 'claude-3-sonnet-20240229', 'gpt-4-turbo', 'gemini-1.5-flash-latest', 'codestral-latest', 'llama3-70b-8192'],
      'do_new_programming_task' : ['claude-3-5-sonnet-20240620', 'gpt-4o', 'gemini-1.5-pro-latest', 'codestral-latest', 'gpt-4-turbo', 'llama3-70b-8192'],
      'do_additional_programming_task' : ['claude-3-5-sonnet-20240620', 'gpt-4o', 'gpt-4-turbo', 'gemini-1.5-pro-latest', 'codestral-latest', 'llama3-70b-8192'],
      'ask_codebase_question' : ['claude-3-5-sonnet-20240620', 'gpt-4o', 'gpt-4-turbo', 'gemini-1.5-pro-latest', 'codestral-latest', 'llama3-70b-8192'],
      'formalize_code_changes' : ['claude-3-5-sonnet-20240620', 'gpt-4o', 'gemini-1.5-flash-latest', 'codestral-latest', 'llama3-70b-8192'],
      'request_file_descriptions' : ['claude-3-5-sonnet-20240620', , 'gpt-4-turbo', 'gemini-1.5-flash-latest', 'codestral-latest', 'llama3-70b-8192'],
      'prune_unlikely_project_files' : ['claude-3-5-sonnet-20240620', 'gpt-4o', 'gemini-1.5-flash-latest', 'gpt-4-turbo', 'codestral-latest', 'llama3-70b-8192'],
      'single_file_full_change' : ['claude-3-5-sonnet-20240620', 'gpt-4o', 'gemini-1.5-flash-latest', 'gpt-4-turbo', 'codestral-latest', 'llama3-70b-8192'],
      'single_file_partial_change' : ['claude-3-5-sonnet-20240620', 'gpt-4o', 'gemini-1.5-flash-latest', 'gpt-4-turbo', 'codestral-latest', 'llama3-70b-8192'],      
      'single_function_full_change' : ['claude-3-5-sonnet-20240620', 'gemini-1.5-pro-latest', 'gpt-4-turbo', 'codestral-latest', 'llama3-70b-8192'],
      'new_code_placement' : ['claude-3-5-sonnet-20240620', 'gemini-1.5-pro-latest', 'gpt-4-turbo', 'codestral-latest', 'llama3-70b-8192']
  },
  estimateTokens: 0,
  originalFileListArr: new Array(),
  selectedProjects: new Array(),
  AVAIL_MODELS: new Array(),
  AVAIL_MODEL_LABELS: new Array(),   
  AVAIL_MODEL_PROVIDER: new Array(),
  getAvailableProviderModels : function(){
      
      let availModels = new Array();
      let availModelLabels = new Array();
      let availModelProv = new Array();
      
      STUBS.CUSTOM_API_URLS.forEach(alldata => {
          let dataSplit = alldata.split('|');
          let url = dataSplit[0];
          let model = 'N/A';
          if(dataSplit.length > 1){
              model = dataSplit[1];          
          }
          availModels.push(alldata);
          availModelLabels.push('Custom: ' + model);
          availModelProv.push('custom');
      });
    
      if(CHATGPT.AVAILABLE_LLMS.includes('anthropic')){
          availModels.push('claude-3-5-sonnet-20240620');
          availModelLabels.push('Claude 3.5 Sonnet (old)');
          availModelProv.push('anthropic');            
          
          availModels.push('claude-3-5-sonnet-20241022');
          availModelLabels.push('Claude 3.5 Sonnet (new)');
          availModelProv.push('anthropic');            
                    
          availModels.push('claude-3-opus-20240229');
          availModelLabels.push('Claude Opus');
          availModelProv.push('anthropic');         
                       
          availModels.push('claude-3-haiku-20240307');
          availModelLabels.push('Claude Haiku');
          availModelProv.push('anthropic');          
      }
      if(CHATGPT.AVAILABLE_LLMS.includes('openai')){
          
          availModels.push('gpt-4o');
          availModelLabels.push('GPT 4o (latest)');
          availModelProv.push('openai');          
          
          availModels.push('gpt-4o-mini');
          availModelLabels.push('GPT 4o Mini');
          availModelProv.push('openai');                    
          
          availModels.push('gpt-4-turbo');
          availModelLabels.push('GPT 4 Turbo');
          availModelProv.push('openai');          
          
          availModels.push('o1-preview');
          availModelLabels.push('o1-preview');
          availModelProv.push('openai');          
 
          availModels.push('o1-mini');
          availModelLabels.push('o1-mini');
          availModelProv.push('openai');          
           
      }
      if(CHATGPT.AVAILABLE_LLMS.includes('gemini')){
          
          availModels.push('gemini-1.5-pro-exp-0801');
          availModelLabels.push('Gemini Pro 1.5 (experimental)');
          availModelProv.push('gemini');                    
          
          availModels.push('gemini-1.5-pro-latest');
          availModelLabels.push('Gemini Pro 1.5 (latest)');
          availModelProv.push('gemini');          
          
          availModels.push('gemini-1.5-flash-latest');
          availModelLabels.push('Gemini Flash 1.5 (latest)');
          availModelProv.push('gemini');     
          
          availModels.push('gemini-1.5-flash-8b-exp-0924');
          availModelLabels.push('Gemini Flash 8b 1.5 (9-24)');
          availModelProv.push('gemini');     
          
      }
      if(CHATGPT.AVAILABLE_LLMS.includes('mistral')){
          /*
          availModels.push('mistral-large-latest');
          availModelLabels.push('Mistral Large');     
          availModelProv.push('mistral');          
          */
          availModels.push('codestral-latest');
          availModelLabels.push('Mistral Codestral');     
          availModelProv.push('mistral');              
      }
      if(CHATGPT.AVAILABLE_LLMS.includes('groq')){
          
          availModels.push('llama-3.1-8b-instant');
          availModelLabels.push('Llama 3.1 8b instant');
          availModelProv.push('groq');

          
          availModels.push('llama-3.1-70b-versatile');
          availModelLabels.push('Llama 3.1 70b');
          availModelProv.push('groq');
          
          
          availModels.push('llama3-70b-8192');
          availModelLabels.push('Llama 3 (8k context too small context window for tasks)');
          availModelProv.push('groq');

          availModels.push('mixtral-8x7b-32768');
          availModelLabels.push('Mixtral 8x7b 32k');
          availModelProv.push('groq');          
      }
      
      CHATGPT.AVAIL_MODELS = availModels;
      CHATGPT.AVAIL_MODEL_LABELS = availModelLabels;
      CHATGPT.AVAIL_MODEL_PROVIDER = availModelProv;
      
      console.log("AVAIL MODELS NOW");
      console.log(CHATGPT.AVAIL_MODELS);      
      return [availModels, availModelLabels, availModelProv];
  },
  updateAvailableProviders : function(){
      CHATGPT.AVAILABLE_LLMS = new Array();
      if(CHATGPT.ANTHROPIC_KEY && CHATGPT.ANTHROPIC_KEY.length > 0){
          CHATGPT.AVAILABLE_LLMS.push('anthropic');
      }
      if(CHATGPT.OPENAI_KEY && CHATGPT.OPENAI_KEY.length > 0){
          CHATGPT.AVAILABLE_LLMS.push('openai');
      }
      if(CHATGPT.GEMINI_KEY && CHATGPT.GEMINI_KEY.length > 0){
          CHATGPT.AVAILABLE_LLMS.push('gemini');
      }
      if(CHATGPT.GROQ_KEY && CHATGPT.GROQ_KEY.length > 0){
          CHATGPT.AVAILABLE_LLMS.push('groq');
      }      
      
      if(CHATGPT.MISTRAL_KEY && CHATGPT.MISTRAL_KEY.length > 0){
          CHATGPT.AVAILABLE_LLMS.push('mistral');
      }      
      
      console.log("AVAILABLE PROVIDERS NOW");
      console.log(CHATGPT.AVAILABLE_LLMS);
  },
  initVertex : async function(googleJsonPath){
    fs.readFile(googleJsonPath, 'utf8', (err, data) => {
      if (err) {
        console.error('Error reading credentials file:', err);
        return;
      }

      // Parse the JSON content
      const credentials = JSON.parse(data);

      // Initialize Vertex with your Cloud project and location
      CHATGPT.vertex = new VertexAI({
        project: credentials.project_id,
        location: 'us-central1', // or your preferred location
        credentials: credentials,
      });

      // Now you can use `vertexAI` to interact with Google Cloud services
    });      
  },
  initProvider : async function(tabIndex, taskType){
      console.log("IN INIT PROVIDER WITH TAB INDEX " + tabIndex);
      let currTab = TABS.TAB_ARR[tabIndex];
            
      let recommendedArr = CHATGPT.recommended_model_properties[taskType];
      
      //have to have the same format as programming_task because it is continuation
      if(taskType == 'do_additional_programming_task'){
          recommendedArr = CHATGPT.recommended_model_properties['do_new_programming_task'];
      }      
      
      if(taskType in STUBS.userOverrideLLMs && STUBS.userOverrideLLMs[taskType] !== undefined && STUBS.userOverrideLLMs[taskType] !== ''){          
          recommendedArr.unshift(STUBS.userOverrideLLMs[taskType]);
          console.log("Adding user override: " + STUBS.userOverrideLLMs[taskType]);
      }

      console.log("recommended arr");
      console.log(recommendedArr);
      
      //if user settings have an override then shift that to top... 
      let llmProvider = "";
      let customUrl = "";
      for(let r = 0; r < recommendedArr.length; r++){
          let theModel = recommendedArr[r];
          if(!theModel){
              console.log("empty/deleted custom model?");
              console.log(theModel);
              continue;
          }
          
          if(theModel.startsWith('http')){
              CHATGPT.LLM_MODEL = theModel;
              let modelParts = theModel.split("|");
              //add key and model
              customUrl = modelParts[0];
              CHATGPT.LLM_PROVIDER = "custom";
              if(modelParts.length > 1){
                  CHATGPT.CUSTOM_OPENAI_KEY = modelParts[2];
              }
              llmProvider = "openai";
              break;
          }
          if(llmProvider != ""){
              break;
          }
          for(let c = 0; c < CHATGPT.AVAIL_MODELS.length; c++){
              let cModel = CHATGPT.AVAIL_MODELS[c];
              let cProv = CHATGPT.AVAIL_MODEL_PROVIDER[c];
              if(cModel == theModel){                  
                  CHATGPT.LLM_MODEL = cModel;
                  CHATGPT.LLM_PROVIDER = cProv;
                  llmProvider = cProv;
                  break;
              }
          }

      }
      console.log("SET MODEL");
      console.log(CHATGPT.LLM_MODEL);   //http://
      console.log("SET PROVIDER");
      console.log(CHATGPT.LLM_PROVIDER);   //custom   
      /*
SET MODEL
chatgpt.js:411 http://localhost:1234/v1/
chatgpt.js:412 SET PROVIDER
chatgpt.js:413 custom      
      */
     //console.log("EXITING");

      if(llmProvider == ""){
          alert("No valid LLM provider, make sure you have at least one API key saved in Settings to continue.  If using a custom provider the url must start with http:");
          return;
      }
      
      TABS.TAB_ARR[tabIndex]['llm_provider'] = llmProvider;
      
      if(llmProvider == 'openai'){
          CHATGPT.init(customUrl);
      }
      else if(llmProvider == 'anthropic'){
          CHATGPT.initAnthropic();          
      }
      else if(llmProvider == 'mistral'){
          CHATGPT.initMistral();    
      }      
      else if(llmProvider == 'groq'){
          CHATGPT.initGroq();
      }           
      else if(llmProvider == 'gemini'){
          //alert("Not implemented");          
          console.log("gemini doesnt need init because using rest");
      }
      else if(llmProvider == 'mistral'){
          CHATGPT.initMistral();
      }
      
      //init in renderer.js now
      /*
      if(CHATGPT.DB_LOGGING){
          let theMysqlConfig = mysqlConfig;
          //console.log("initing database");
          //console.log(theMysqlConfig);
          await mysqlLogger.initDatabase(theMysqlConfig);        
      }    
      */
  },
  initAnthropic : function(){
      console.log("init anthropic");
      if(CHATGPT.anthropic != null){
          //already set
          return;
      }
      /*
        if(!myApiKeys){
            myApiKeys = apiKeys;
        }
        for (const apikey of Object.entries(myApiKeys)) {
            const [theName, theKey] = apikey;
            if(theName == 'anthropic'){
                CHATGPT.ANTHROPIC_KEY = theKey;
            }
        }      
        */
        if(CHATGPT.ANTHROPIC_KEY == ''){
            alert("no anthropic key");
            return;
        }
        
        CHATGPT.anthropic = new Anthropic({
            apiKey: CHATGPT.ANTHROPIC_KEY
        });        
        
        return true;        
  },
  initGroq : function(){
      console.log("init groq");
      if(CHATGPT.groq != null){
          //already set
          return;
      }
      /*
        if(!myApiKeys){
            myApiKeys = apiKeys;
        }
        for (const apikey of Object.entries(myApiKeys)) {
            const [theName, theKey] = apikey;
            if(theName == 'groq'){
                CHATGPT.GROQ_KEY = theKey;
            }
        }      
        */
        if(CHATGPT.GROQ_KEY == ''){
            alert("no groq key");
            return;
        }
        CHATGPT.groq = new Groq({
            apiKey: CHATGPT.GROQ_KEY,
            dangerouslyAllowBrowser: true
        });           
        
        return true;
        
  },   
  initMistral : function(){
      console.log("init mistral");
      if(CHATGPT.mistral != null){
          //already set
          return;
      }
      /*
        if(!myApiKeys){
            myApiKeys = apiKeys;
        }
        for (const apikey of Object.entries(myApiKeys)) {
            const [theName, theKey] = apikey;
            if(theName == 'mistral'){
                CHATGPT.MISTRAL_KEY = theKey;
            }
        }      
        */
        if(CHATGPT.MISTRAL_KEY == ''){
            alert("no mistral key");
            return;
        }
        CHATGPT.mistral = new MistralClient(CHATGPT.MISTRAL_KEY);             
        
        return true;
        
  },  
  init : function(customUrl = ""){ 
      if(CHATGPT.openai != null && customUrl == "" && CHATGPT.CUSTOM_OPENAI_KEY == 'XXX'){
          console.log("not creating new openai instance");
          //already set
          return;
      }
      console.log("creating new openai instance");
      /*
        if(!myApiKeys){
            myApiKeys = apiKeys;
        }
        

        for (const apikey of Object.entries(myApiKeys)) {
            const [theName, theKey] = apikey;
            if(theName == 'openai'){
                CHATGPT.OPENAI_KEY = theKey;
            }
        }
        */
        if(CHATGPT.OPENAI_KEY == '' && customUrl == ""){
            console.log("NO OPENAI KEY");
            alert("NO OPENAI KEY");
            return;
        }
        CHATGPT.inCodeBlock = false;
        CHATGPT.inTextBlock = false;
        if(customUrl == ""){
            let openaiConfig = {
              apiKey: CHATGPT.OPENAI_KEY,
              dangerouslyAllowBrowser: true         
            };
            console.log("standard instance " + customUrl);            
            if(CHATGPT.OPENAI_ORG != ""){
                //"org-YxcLU2YD2WyxucLJBdVLYT15"
                openaiConfig['organization'] = CHATGPT.OPENAI_ORG;
            }        
            CHATGPT.openai = new OpenAI(openaiConfig);
        }else{
            console.log("Using a custom url " + customUrl);
            let openaiConfig = {
              dangerouslyAllowBrowser: true,
              baseURL: customUrl,
              apiKey: CHATGPT.CUSTOM_OPENAI_KEY
            };     
            console.log("baseurl instance " + customUrl);
            if(customUrl.indexOf('openrouter') >= 0){
                openaiConfig.defaultHeaders = {
                  "HTTP-Referer": 'https://stubs.dev', 
                  "X-Title": "Stubs", 
                };
            }
            
            CHATGPT.openai = new OpenAI(openaiConfig);            
        }
  }, 
  estimateFileTokens : async function(messages){
        
        let fullPrompt = messages.join(' ');

        const encoded = encode(fullPrompt);

        console.log("token length");
        let totalToken = 0.0;
        for(let aToken in encoded){
            //let tokenAmount = (aToken / 10000);
            let tokenAmount = 0.483;
            //console.log("adding " + tokenAmount);
            totalToken += tokenAmount;
        }
        console.log("trying to get 6323");
        console.log(totalToken);             
        CHATGPT.estimateTokens = parseInt(totalToken);
        $('#token_estimate').html(CHATGPT.estimateTokens);
  },  
  testAnAPI : async function(apiType){
      alert("test disabled");
      /*
      if(apiType == 'openai'){
         CHATGPT.init();
      }
      else if(apiType == 'anthropic'){
          let good = CHATGPT.initAnthropic();
          if(!good){
              return;
          }
      }

  let testTask = `I'm using Croppie, a javascript cropping library.   I want to get the bounding left, top, and width, height of the area I wish to crop when I click id="crop_and_regen" button.  Can you write code to do that?  Here is my code:  

  function getImageSize(url) {
    return new Promise((resolve, reject) => {
      let img = new Image();
      img.crossOrigin = 'use-credentials';
      img.onload = function() {
        resolve({
          width: img.naturalWidth,
          height: img.naturalHeight
        });
      };
      img.onerror = function() {
        reject(new Error("Could not load image"));
      };
      img.src = url + '?v=' + new Date().getTime();
    });
  }      

  $(document).ready(function(){
    //croppie
      let theWidth, theHeight;
      let sImageUrl = $('#screenshot_image').attr('src');
      console.log("PROMISE FOR " + sImageUrl);
      let aPromise = getImageSize(sImageUrl);
      aPromise.then((retObj) => {
          theWidth = parseInt(retObj['width']);
          theHeight = parseInt(retObj['height']);
      console.log("PROMISE RES " + retObj);

          let aspectRatio = theHeight / theWidth;
          //50 / 100;
          let viewHeight = 330 * aspectRatio;
          let boundHeight = 450 * aspectRatio;

          console.log("setting up croppie");
          var el = document.getElementById('screenshot_image');
          var resize = new Croppie(el, {
              viewport: { width: 330, height: viewHeight },
              boundary: { width: 450, height: boundHeight },
              showZoomer: false,
              enableResize: true,
              enableOrientation: true,
              mouseWheelZoom: 'ctrl'
          });
          resize.bind({
              url: sImageUrl
          });
          

      });
      
      $('#crop_and_regen').on('click',function(){
          
          
      });
      
  });`;                    

        if(apiType == 'openai'){
            let theMessages = new Array();
            theMessages.push({
                role: "system",
                content: "You are a helpful coding assistant."
            });   
            
            theMessages.push({
                role: "user",
                content: testTask
            });            
            console.log("sending");
            console.log(theMessages);
            (async () => {
              let chatStream = await CHATGPT.streamChatCompletion(theMessages, 'do_new_programming_task');
              console.log("chatstream is");
              console.log(chatStream);
              //readStreamAndUpdateTextArea(chatStream);
            })();      
        }
        else if(apiType == 'anthropic'){
            let theMessages = new Array();
            
            theMessages.push({
                role: "user",
                content: testTask
            });                 
            let response = await CHATGPT.anthropic.messages.create({
              model: 'claude-3-opus-20240229',
              max_tokens: 4096,
              messages: theMessages
            });            
            console.log("response");            
            console.log(response);
        }else{
            alert("unknown");
        }
      */
  },
  processHtmlText : function (htmlText) {
    // Split the input HTML string into sections
    const sections = htmlText.split('`' + '`' + '`');
    let outputHtml = '';
    let inCodeBlock = false;
    let codeblockIndex = 0; // Initialize code block counter

    sections.forEach((section, index) => {
        if (inCodeBlock) {
            // Apply Highlight.js for code blocks and add id
            const highlightedCode = hljs.highlightAuto(section).value;
            outputHtml += `<div><pre class="hljs" id="code_${codeblockIndex}"><code>${highlightedCode}</code></pre></div>`;
            codeblockIndex++; // Increment code block counter
        } else {
            // Replace content within single backticks with escaped HTML
            section = section.replace(/`([^`]+)`/g, (match, p1) => {
                return `<span class="inline-code">${escapeMyHtml(p1)}</span>`;
            });

            // Replace line breaks with <br> tags for normal HTML
            const normalHtml = section.replace(/\n/g, '<br>');
            outputHtml += `<p>${normalHtml}</p>`;
        }
        inCodeBlock = !inCodeBlock; // Toggle inCodeBlock for alternating sections
    });
    
    //THIS IS WRONG... WE ARE NOT LOADING THE CURRRRRRRENT UID... WE ARE GOING TO INSERT INTO THE CORRECT ONE EVEN IF HIDDEN
    let currentTabUid = TABS.getCurrentUid();
    // Set the processed HTML to the target element
    $('#chatgptresponse_' + currentTabUid).html(outputHtml);
  },  
  updateHtmlText : async function(newText, tabUid){
      
      if(typeof(newText) === 'undefined'){
         console.log("returning false?  Is this a stop event?");
         return false;
      }      
      let tabIndex = TABS.getIndexByUid(tabUid);
      if(newText.indexOf('`') >= 0){
          CHATGPT.bufferText += newText;
          //console.log("===== PUTTING IN BUFFER " + newText);
          //console.log(CHATGPT.bufferText);
          return;
      }
      newText = CHATGPT.bufferText + newText;
      let updateBufferText = CHATGPT.bufferText;
      CHATGPT.bufferText = "";
      //console.log("===== PROCESSING ");
      //console.log(newText);
      let newHtml = newText;
      let updateHighlight = "";
      
      if(newText.indexOf('`' + '`' + '`') >= 0){
          if(CHATGPT.inTextBlock){
              //console.log("codeblock start");              
              //newHtml = replaceAll(newHtml, '\n', '<br>');
              //let htmlArr = newHtml.split('`' + '`' + '`');
              //newHtml = htmlArr[0] + '</p><pre id="code_' + CHATGPT.codeblockIndex + '"><code class="language-' + CHATGPT.lastLanguage + '">' + escapeHtml(htmlArr[1]);
              /*
              let currentFilename = CHATGPT.extractFilePath(updateBufferText, tabUid);            
              if(currentFilename != ''){    
                  console.log("SET codeblockFILENAME AT " + CHATGPT.codeblockIndex + " to " + currentFilename);
                  CHATGPT.codeblockFilenames[CHATGPT.codeblockIndex] = currentFilename;  
              }else{
              */
             //just use this html version which seems to work unlike above
              console.log("codeblockFilenames getting from all html so far");
              let fallbackHtml = replaceAll(CHATGPT.htmlText, '<br>', "\n");
              currentFilename = CHATGPT.extractFilePath(fallbackHtml, tabUid);   
              CHATGPT.codeblockFilenames[CHATGPT.codeblockIndex] = currentFilename; 
              console.log("set to " + currentFilename);
              
   
              newHtml = replaceAll(newHtml, '`' + '`' + '`', '</p><pre class="hljs" id="code_' + CHATGPT.codeblockIndex + '"><code>');
              CHATGPT.codeblockIndex++;
              CHATGPT.inTextBlock = false;
              CHATGPT.inCodeBlock = true;
              //dytc - 
          }
          else if(CHATGPT.inCodeBlock){
              //console.log("codeblock end");                            
              //newHtml = replaceAll(newHtml, '\n', '<br>');
              //let htmlArr = newHtml.split('`' + '`' + '`');
              //newHtml = htmlArr[0] + '</code></pre><p>' + htmlArr[1];              
              newHtml = replaceAll(newHtml, '`' + '`' + '`', '</code></pre><p>');    
              CHATGPT.inTextBlock = true;  
              let lastCodeIndex = CHATGPT.codeblockIndex - 1;
              let lastCodeSelector = "#code_" + lastCodeIndex + " code";
              updateHighlight = lastCodeSelector;              
          }else{
              alert("strange codestart");
              console.log("STRANGE");
              console.log("newHtml");
              console.log(newHtml);
              console.log("CHATGPT.htmlText");
              console.log(CHATGPT.htmlText);
              
              //newHtml = replaceAll(newHtml, '\n', '<br>');
              newHtml = replaceAll(newHtml, '`' + '`' + '`', '<pre class="hljs" id="code_' + CHATGPT.codeblockIndex + '"><code>');
              CHATGPT.codeblockIndex++;
              CHATGPT.inTextBlock = false;
              CHATGPT.inCodeBlock = true;              
          }
      }
      else if(newText == ""){
          //end of response
          newHtml = "";
      }
      else if(!CHATGPT.inTextBlock && !CHATGPT.inCodeBlock){
          //only happens at start
          newHtml = newHtml.replace(/`([^`]+)`/g, (match, p1) => {
              return `<span class="inline-code">${escapeMyHtml(p1)}</span>`;
          });

          newHtml = '<p>' + newHtml;
          CHATGPT.inTextBlock = true;
      }
      else if(CHATGPT.inTextBlock){
          newHtml = newHtml.replace(/`([^`]+)`/g, (match, p1) => {
              return `<span class="inline-code">${escapeMyHtml(p1)}</span>`;
          });
          newHtml = replaceAll(newHtml, '\n', '<br>');                  
      }
      else if(CHATGPT.inCodeBlock){
          //newHtml = replaceAll(escapeHtml(newHtml), '\n', '<br>');  
          let lastCodeIndex = CHATGPT.codeblockIndex - 1;
          let lastCodeSelector = "#code_" + lastCodeIndex + " code";
          updateHighlight = lastCodeSelector;          
      }
      
      
      
      let outputHtml = CHATGPT.htmlText + newHtml;
      if(CHATGPT.inCodeBlock){
          outputHtml += '</code></pre>';
      }
      if(CHATGPT.inTextBlock){
          outputHtml += '</p>';
      }
      CHATGPT.htmlText = CHATGPT.htmlText + newHtml;
      const currentTime = Date.now();
      if ((currentTime - CHATGPT.lastProcessTime) > CHATGPT.processingInterval) {
          CHATGPT.lastProcessTime = currentTime;
          CHATGPT.postProcessHtmlText(CHATGPT.htmlText, tabUid);
      }      
      
      
      
      //console.log("before output html");
      //console.log(outputHtml);     
  },  
  postProcessHtmlText : function(outputHtml, tabUid){
      //console.log("IN POST PROCESS HTML");
      let tabIndex = TABS.getIndexByUid(tabUid);
      let $editHtml = cheerio.load(outputHtml, null, false);
      TABS.TAB_ARR[tabIndex]['code_blocks'] = new Array();         
      $editHtml.root().find("code").each(function(i, elem){    
         let theCode = $editHtml(elem).html();
         //console.log("found code element");
         //console.log(theCode);
   
         theCode = unescapeHtml(theCode);
         TABS.TAB_ARR[tabIndex]['code_blocks'].push(theCode);               
         let highlightCode = hljs.highlightAuto(theCode).value;

         // Check if first line is a single word indicating code type
         let codeLines = theCode.split(/\r?\n/);
         if (codeLines.length > 1 && codeLines[0].trim().split(' ').length === 1) {
           let codeType = codeLines[0].trim().toLowerCase();
           if (['javascript', 'php', 'html', 'css'].includes(codeType)) {
             // Remove first line
             codeLines.shift();
             theCode = codeLines.join('\n');
             highlightCode = hljs.highlightAuto(theCode).value;
           }
         }
          
         $editHtml(elem).html(highlightCode);

         // Add copy button and header to code block
         let copyIndex = i;
         //console.log("Post ProcessCode blcok filenames");
         //console.log(CHATGPT.codeblockFilenames);
         let filename = CHATGPT.codeblockFilenames[i] || "";
         //console.log("filename at " + i);
         //console.log(filename);
         let copyButton = `<div class="copybuttoncontainer">
              <button id="copy_${tabUid}_${copyIndex}" class="copy-button"><i class="ti ti-copy"></i></button>
          </div>
          <div class="codemirror-cont-header">${filename}<i class="ti ti-chevron-down"></i></div>`;   
          
          $editHtml(elem).parent('pre').before(copyButton);
      });
      outputHtml = $editHtml.root().html();
      $('#chatgptresponse_' + tabUid).html(outputHtml);  
 
  }, 
  finalizeHtmlText : function(tabUid){
      //fuck it just add copy buttons afterwards
      setTimeout(function(){
          console.log("FINALIZING CODE HEADERS");
          let copyIndex = 0;
          $('#chatgptresponse_' + tabUid).find('.copybuttoncontainer').remove();
          $('#chatgptresponse_' + tabUid).find('.codemirror-cont-header').remove();
          
          $('#chatgptresponse_' + tabUid).find('pre').each(function(i, elem){
              console.log("pre elem id: ");
              console.log(elem.id);
              /*
               let copyButton = `<div class="copybuttoncontainer">
                    <button id="copy_${tabUid}_${copyIndex}" class="copy-button"><i class="ti ti-copy"></i></button>
                </div>
                <div class="codemirror-cont-header"><i class="ti ti-chevron-down"></i></div>`;   
              */
                // Add copy button, header, and filename to code block
                let copyIndex = i;
                console.log("copy index: " + copyIndex);
                
                let filename = CHATGPT.codeblockFilenames[i] || ""; // Get filename from codeblockFilenames array
                let copyButton = `<div class="copybuttoncontainer">
                     <button id="copy_${tabUid}_${copyIndex}" class="copy-button"><i class="ti ti-copy"></i></button>
                 </div>
                 <div class="codemirror-cont-header">${filename}<i class="ti ti-chevron-down"></i></div>`;                 
              
               $(elem).before(copyButton);
               copyIndex++;
          });          
          //$('#user_response_container').css('display', 'block');    
          $('#plusbuttoncontainer_' + tabUid).css('display', 'block');
          
      }, 100);      
  },
  providerChatCompletion: async function(myMessages, requestType, currTab, jsonMode){
      
      if(currTab['llm_provider'] == 'openai'){
          let response = await CHATGPT.normalChatCompletion(myMessages, requestType, currTab, jsonMode);
          return response;
      }else if(currTab['llm_provider'] == 'anthropic'){
          let response = await CHATGPT.normalAnthropicCompletion(myMessages, requestType, currTab, jsonMode);          
          return response;
      }
      else if(currTab['llm_provider'] == 'mistral'){
          let response = await CHATGPT.normalMistralCompletion(myMessages, requestType, currTab, jsonMode);          
          return response;
      }
      else if(currTab['llm_provider'] == 'groq'){
          let response = await CHATGPT.normalGroqCompletion(myMessages, requestType, currTab, jsonMode);          
          return response;
      }   
      else if(currTab['llm_provider'] == 'gemini'){
          let response = await CHATGPT.normalGeminiCompletion(myMessages, requestType, currTab, jsonMode);          
          return response;
      }       
      
      throw new Error('Uh Oh - bad LLM PROVIDER');
  },
normalGroqCompletion : async function(myMessages, requestType, currTab, jsonMode) {
    let fullText = '';
    CHATGPT.htmlText = "";
    CHATGPT.codeLength = 0;
  
    if(CHATGPT.DB_LOGGING){
        if(!CHATGPT.NO_CACHE_REQUESTS.includes(requestType)){
            let mockResponse = await mysqlLogger.getMockResponse(requestType, currTab['task_text'], "groq");
            if (mockResponse) {
              console.log('Mocking response:', mockResponse);
              return mockResponse; // or you can use `return { success: true, response: mockResponse };` depending on your requirements
            }       
        }
        else if(requestType == 'single_file_partial_change' || requestType == 'new_code_placement'){
            //we need to debug this by replaying 
            let mockResponse = await mysqlLogger.getMockByFullPrompt(requestType, fullPrompt, "groq");
            if (mockResponse) {
              console.log('GOT ' + requestType + ' FILE CHANGE MOCK RESPONSE:', mockResponse);
              return mockResponse; // or you can use `return { success: true, response: mockResponse };` depending on your requirements
            }           
        }
        
    }

    let fullPrompt = "";
    for (let anObj of myMessages) {
      if (anObj.role === 'user') {
        fullPrompt += anObj.content;
      }
    }

    let modelSettings = CHATGPT.groq_model_properties[requestType];

    if(CHATGPT.DB_LOGGING){
        //change TABS.CURRENT_TASK and CHATGPT.LLM_PROVIDER TO USE THE ONE FROM TABS
        mysqlLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider']);
    }
    if(CHATGPT.DEXIE_LOGGING){
        CHATGPT.dexieLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
    }    
    if(CHATGPT.FILE_LOGGING){
        const fileSending = `sendinggroq_${requestType}_${getFormattedDateTime()}.txt`;
        const logSendingPath = path.join('.', 'logs', fileSending);
        const jsonMyMessages = JSON.stringify(myMessages, null, 2);
        fs.writeFileSync(logSendingPath, jsonMyMessages, 'utf8');
    }

  const retArr = {};
  let assistantMessage = "";

  try {
    const response = await CHATGPT.groq.chat.completions.create({
        messages: myMessages,
        model: CHATGPT.LLM_MODEL,
        temperature: 0.1
    });

    if (!response || !response.choices || response.choices.length <= 0) {
      throw new Error('Bad message response body');
    }

    assistantMessage = response.choices[0].message.content;

    retArr.success = true;
    retArr.orig_response = assistantMessage;

    if (requestType === 'request_relevant_files') {
      retArr.response = CHATGPT.parseFilesList(assistantMessage, currTab);
    } else if (requestType === 'request_relevant_line_numbers') {
      
        let assistantJson = CHATGPT.parseMixedResponseToJsonArray(assistantMessage, "file_path");
        if(assistantJson.length <= 0){
            retArr.success = false;
            retArr.error = "BAD JSON";
            return retArr;
        }

        console.log("returning");
        console.log(assistantJson);
        retArr.response = assistantJson;
    } else if (requestType === 'request_relevant_stub_lines' || requestType == 'request_fix_stub_lines') {
        console.log("request_relevant_stub_lines assistant message: " );
        console.log(assistantMessage);
        let stubLineResp = CHATGPT.parseStubFileLines(assistantMessage); 
        if(stubLineResp.length <= 0){
            retArr.success = false;
            retArr.error = "BAD STUB LINES REPONSE";
            console.log(assistantMessage);
            return retArr;
        }

        console.log("returning Groq");
        console.log(stubLineResp);
        retArr.response = stubLineResp;
    } 
    else if(requestType == 'do_new_programming_task'){
        retArr.response = assistantMessage;
    }
    else if(requestType == 'single_file_full_change' || requestType == 'single_function_full_change' || requestType == 'single_file_partial_change'){
        retArr.response = CHATGPT.getJustCode(assistantMessage, currTab['just_files_linux']);              
    }
    else if(requestType == 'new_code_placement'){
        retArr.response = CHATGPT.parseNewCodePlacement(assistantMessage);
    }
    else if (requestType === 'formalize_code_changes') {
            console.log("formal change assistantMessage");
            console.log(assistantMessage);
            let assistantJson = CHATGPT.parseMixedResponseToJsonArray(assistantMessage, "contiguous_lines");
            if(assistantJson.length <= 0){
                retArr.success = false;
                retArr.error = "BAD JSON";
                return retArr;
            }            
            retArr.response = assistantJson;
    }else if (requestType === 'request_file_descriptions') {
      retArr.response = CHATGPT.parseFileDescriptions(assistantMessage);
    }
    else if (requestType === 'prune_unlikely_project_files') {
        retArr.response = CHATGPT.parsePruneUnlikelyProjectFiles(assistantMessage);
    }      
        
        let inputTokens = 0;
        let outputTokens = 0;
        if(response.usage){            
            retArr.input_tokens = response.usage.input_tokens;
            inputTokens = response.usage.input_tokens;
            retArr.output_tokens = response.usage.output_tokens;
            outputTokens = response.usage.output_tokens;
        }
    
        if(CHATGPT.DB_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            mysqlLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider']);
        }
        if(CHATGPT.DEXIE_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            CHATGPT.dexieLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL, inputTokens, outputTokens);
        }        
        
        if(CHATGPT.FILE_LOGGING){
            const fileName = `mistral_${requestType}_${getFormattedDateTime()}.txt`;
            const logFilePath = path.join('.', 'logs', fileName);
            fs.writeFileSync(logFilePath, assistantMessage, 'utf8');
        }
  } catch (error) {
    const retArr = {};
    console.error("API request error:", error);
    retArr.success = false;
    retArr.error = "API Error: " + error;
    return retArr;
  }

  return retArr;
},  
  normalChatCompletion: async function(myMessages, requestType, currTab, jsonMode){
    let fullText = '';
    CHATGPT.htmlText = "";  
    CHATGPT.codeLength = 0;
        
    let modelSettings = CHATGPT.model_properties[requestType];
    let fullPrompt = "";
    for (let anObj of myMessages) {
      if (anObj.role === 'user') {
        fullPrompt += anObj.content;
      }
    }       
    
    if(CHATGPT.DB_LOGGING){
        let extraTasks = "";
        if(currTab['more_tasks'].length > 0){
            for(let c = 0; c < currTab['more_tasks'].length; c++){
                extraTasks += currTab['more_tasks'][c];
            }
        }
        if(!CHATGPT.NO_CACHE_REQUESTS.includes(requestType)){
        
            let mockResponse = await mysqlLogger.getMockResponse(requestType, currTab['task_text'] + extraTasks, currTab['llm_provider']);
            if (mockResponse) {
              console.log('Mocking response:', mockResponse);
              return mockResponse; // or you can use `return { success: true, response: mockResponse };` depending on your requirements
            }     
        }
        else if(requestType == 'single_file_partial_change' || requestType == 'new_code_placement'){
            //we need to debug this by replaying 
            let mockResponse = await mysqlLogger.getMockByFullPrompt(requestType, fullPrompt, currTab['llm_provider']);
            if (mockResponse) {
              console.log('GOT ' + requestType + ' FILE CHANGE MOCK RESPONSE:', mockResponse);
              return mockResponse; // or you can use `return { success: true, response: mockResponse };` depending on your requirements
            }           
        }        
    }   

    

    //console.log("model: " + modelSettings['model'] + " tokens: " + modelSettings['tokens']);
    
    //let maxTokens = parseInt(document.getElementById('max_tokens_slider').value);

    //faster testing
    //modelVersion = 'gpt-3.5-turbo';
    //gpt-3.5-turbo-16k
    //gpt-4
    //gpt-4-1106-preview
    //gpt-4-vision-preview
    
    if(CHATGPT.DB_LOGGING){
        //change TABS.CURRENT_TASK and CHATGPT.LLM_PROVIDER TO USE THE ONE FROM TABS
        mysqlLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider']);
    }
    if(CHATGPT.DEXIE_LOGGING){
        CHATGPT.dexieLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
    }        
    if(CHATGPT.FILE_LOGGING){
        const fileSending = `sending_${requestType}_${getFormattedDateTime()}.txt`;
        const logSendingPath = path.join('.', 'logs', fileSending);
        const jsonMyMessages = JSON.stringify(myMessages, null, 2);
        fs.writeFileSync(logSendingPath, jsonMyMessages, 'utf8');    
    }
    const retArr = {};
    let assistantMessage = "";    
    let myTemp = 0.1;
    if(CHATGPT.LLM_MODEL.indexOf('o1') >= 0){     
        //o1 models need temp of 1.0?
        myTemp = 1.0;
    }
    
    
    try {
        // Use the timeout method for the API call
        //const completion = await CHATGPT.timeout(CHATGPT.request_timeout, CHATGPT.openai.createChatCompletion(requestObject));
        let requestObject = {
            messages: myMessages,
            temperature: myTemp
        };
        let customUrl = false;
        if(CHATGPT.LLM_MODEL.startsWith('http')){
            console.log("USING CUSTOM URL");
            let modelParts = CHATGPT.LLM_MODEL.split("|");
            if(modelParts.length > 0){
                requestObject.model = modelParts[1];
            }          
            customUrl = true;
            
        }else{
            requestObject.model = CHATGPT.LLM_MODEL;
        }
        
        const response = await Promise.race([
          CHATGPT.openai.chat.completions.create(requestObject),
          createTimeoutPromise(30000)
        ]);  
        
        //const response = await CHATGPT.openai.chat.completions.create(requestObject);        
        
        if (!response || !response.choices || response.choices.length <= 0) {
          throw new Error('Bad message response body');
        }
        if(customUrl){
            console.log("custom response");
            console.log(response);
        }

        assistantMessage = response.choices[0].message.content;

        retArr.success = true;
        retArr.orig_response = assistantMessage;     
        
        let inputTokens = 0;
        let outputTokens = 0;        
        if(response.usage){            
            retArr.input_tokens = response.usage.prompt_tokens;
            inputTokens = response.usage.prompt_tokens;
            retArr.output_tokens = response.usage.completion_tokens;
            outputTokens = response.usage.completion_tokens;
        }else{
            console.log("no usage data");
        }        
        retArr.model = CHATGPT.LLM_MODEL;
        let currentDate = new Date().toISOString().split('T')[0];       
        retArr.request_date = currentDate;
        
        //first create the stubs, send stubs to get_task_files
        if(requestType == 'request_relevant_files'){
            retArr.response = CHATGPT.parseFilesList(assistantMessage, currTab);
        }else if(requestType == 'request_relevant_line_numbers'){
            console.log("parsing");
            console.log(assistantMessage);
            let assistantJson = CHATGPT.parseMixedResponseToJsonArray(assistantMessage, "file_path");
            if(assistantJson.length <= 0){
                retArr.success = false;
                retArr.error = "BAD JSON";
                return retArr;
            }

            console.log("returning");
            console.log(assistantJson);
            retArr.response = assistantJson;      
        }else if (requestType === 'request_relevant_stub_lines' || requestType == 'request_fix_stub_lines') {
            let stubLineResp = CHATGPT.parseStubFileLines(assistantMessage); 
            if(stubLineResp.length <= 0){
                retArr.success = false;
                retArr.error = "BAD STUB LINES REPONSE";
                console.log(assistantMessage);
                return retArr;
            }

            console.log("returning ChatGPT");
            console.log(stubLineResp);
            retArr.response = stubLineResp;
        }
        else if(requestType == 'do_new_programming_task'){
            retArr.response = assistantMessage;
        }
        else if(requestType == 'single_file_full_change' || requestType == 'single_function_full_change' || requestType == 'single_file_partial_change'){
            retArr.response = CHATGPT.getJustCode(assistantMessage, currTab['just_files_linux']);      
        }
        else if(requestType == 'new_code_placement'){
            retArr.response = CHATGPT.parseNewCodePlacement(assistantMessage);
        }        
        else if(requestType == 'formalize_code_changes'){

            console.log("formal change assistantMessage");
            console.log(assistantMessage);
            let formalChangeArr = CHATGPT.parseFormalizedChanges(assistantMessage);

            if(formalChangeArr.length <= 0){
                console.log("invalid formal changes");
                retArr.success = false;
                retArr.error = "Invalid JSON for formalizing response: " + assistantMessage;     

                //we need to parse alternative

                return retArr;
            }        
            retArr.response = formalChangeArr;   
        }else if (requestType === 'request_file_descriptions') {
            retArr.response = CHATGPT.parseFileDescriptions(assistantMessage);
        }
        else if (requestType === 'prune_unlikely_project_files') {
            retArr.response = CHATGPT.parsePruneUnlikelyProjectFiles(assistantMessage);
        }          

        //console.log("retArr");
        //console.log(retArr);
        
        if(CHATGPT.DB_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            mysqlLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider']);
        }
        if(CHATGPT.DEXIE_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            CHATGPT.dexieLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL, inputTokens, outputTokens);
        }                
        if(CHATGPT.FILE_LOGGING){
            const fileName = `${requestType}_${getFormattedDateTime()}.txt`;
            const logFilePath = path.join('.', 'logs', fileName);
            fs.writeFileSync(logFilePath, assistantMessage, 'utf8');                    
        }
    } catch (error) {
        // Handle the error
        const retArr = {};
        if (error.message.startsWith("Timeout after") || error.message.startsWith("Request timed ")) {
            // Handle timeout-specific errors
            console.error("Request timed out:", error.message);
            console.log("REQUEST TIMEOUT ------------------");
            console.log("REQUEST TIMEOUT ------------------");
            console.log("REQUEST TIMEOUT ------------------");
            console.log("REQUEST TIMEOUT ------------------");            
            retArr.success = false;
            retArr.error = "Timeout Error: " + error.message;
        } else {
            // Handle other errors
            console.error("API request error:", error);
            retArr.success = false;
            retArr.error = "API Error: " + error;
        }
        return retArr;
    }        
    
    return retArr;     
  },
normalMistralCompletion : async function(myMessages, requestType, currTab, jsonMode) {
  let fullText = '';
  CHATGPT.htmlText = "";
  CHATGPT.codeLength = 0;
  
    if(CHATGPT.DB_LOGGING && !CHATGPT.NO_CACHE_REQUESTS.includes(requestType)){
        let mockResponse = await mysqlLogger.getMockResponse(requestType, currTab['task_text'], "mistral");
        if (mockResponse) {
          console.log('Mocking response:', mockResponse);
          return mockResponse; // or you can use `return { success: true, response: mockResponse };` depending on your requirements
        }       
    }

    let fullPrompt = "";
    for (let anObj of myMessages) {
      if (anObj.role === 'user') {
        fullPrompt += anObj.content;
      }
    }

    let modelSettings = CHATGPT.mistral_model_properties[requestType];

    if(CHATGPT.DB_LOGGING){
        //change TABS.CURRENT_TASK and CHATGPT.LLM_PROVIDER TO USE THE ONE FROM TABS
        mysqlLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider']);
    }
    if(CHATGPT.DEXIE_LOGGING){
        CHATGPT.dexieLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
    }        
    if(CHATGPT.FILE_LOGGING){
        const fileSending = `sendingmistral_${requestType}_${getFormattedDateTime()}.txt`;
        const logSendingPath = path.join('.', 'logs', fileSending);
        const jsonMyMessages = JSON.stringify(myMessages, null, 2);
        fs.writeFileSync(logSendingPath, jsonMyMessages, 'utf8');
    }

  const retArr = {};
  let assistantMessage = "";

  try {
      console.log("sending max tokens: " + modelSettings['tokens']);
    const response = await CHATGPT.mistral.chat({
      model: CHATGPT.LLM_MODEL,
      max_tokens: modelSettings['tokens'],
      temperature: 0.0,
      messages: myMessages
    });

    if (!response || !response.choices || response.choices.length <= 0) {
      throw new Error('Bad message response body');
    }

    assistantMessage = response.choices[0].message.content;

    retArr.success = true;
    retArr.orig_response = assistantMessage;
    
    let inputTokens = 0;
    let outputTokens = 0;        
    if(response.usage){            
        inputTokens = response.usage.prompt_tokens;
        outputTokens = response.usage.completion_tokens;
    }       
    
    retArr.model = CHATGPT.LLM_MODEL;
    let currentDate = new Date().toISOString().split('T')[0];       
    retArr.request_date = currentDate;    

    if (requestType === 'request_relevant_files') {
      retArr.response = CHATGPT.parseFilesList(assistantMessage, currTab);
    } else if (requestType === 'request_relevant_line_numbers') {
      
        let assistantJson = CHATGPT.parseMixedResponseToJsonArray(assistantMessage, "file_path");
        if(assistantJson.length <= 0){
            retArr.success = false;
            retArr.error = "BAD JSON";
            return retArr;
        }

        console.log("returning");
        console.log(assistantJson);
        retArr.response = assistantJson;
    }else if (requestType === 'request_relevant_stub_lines' || requestType == 'request_fix_stub_lines') {
            let stubLineResp = CHATGPT.parseStubFileLines(assistantMessage); 
            if(stubLineResp.length <= 0){
                retArr.success = false;
                retArr.error = "BAD STUB LINES REPONSE";
                console.log(assistantMessage);
                return retArr;
            }

            console.log("returning mistral");
            console.log(stubLineResp);
            retArr.response = stubLineResp;
    }
    else if(requestType == 'do_new_programming_task'){
        retArr.response = assistantMessage;
    }
    else if(requestType == 'single_file_full_change' || requestType == 'single_function_full_change' || requestType == 'single_file_partial_change'){
        retArr.response = CHATGPT.getJustCode(assistantMessage, currTab['just_files_linux']);           
    }
    else if(requestType == 'formalize_code_changes'){

            console.log("formal change assistantMessage");
            console.log(assistantMessage);
            let formalChangeArr = CHATGPT.parseFormalizedChanges(assistantMessage);

            if(formalChangeArr.length <= 0){
                console.log("invalid formal changes");
                retArr.success = false;
                retArr.error = "Invalid JSON for formalizing response: " + assistantMessage;     

                //we need to parse alternative

                return retArr;
            }        
            retArr.response = formalChangeArr;   
    }
    else if (requestType === 'request_file_descriptions') {
      retArr.response = CHATGPT.parseFileDescriptions(assistantMessage);
    }    
    else if (requestType === 'prune_unlikely_project_files') {
        retArr.response = CHATGPT.parsePruneUnlikelyProjectFiles(assistantMessage);
    }      
        if(CHATGPT.DB_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            mysqlLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider']);
        }
        if(CHATGPT.DEXIE_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            CHATGPT.dexieLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL, inputTokens, outputTokens);
        }            
        if(CHATGPT.FILE_LOGGING){
            const fileName = `mistral_${requestType}_${getFormattedDateTime()}.txt`;
            const logFilePath = path.join('.', 'logs', fileName);
            fs.writeFileSync(logFilePath, assistantMessage, 'utf8');
        }
  } catch (error) {
    const retArr = {};
    console.error("API request error:", error);
    retArr.success = false;
    retArr.error = "API Error: " + error;
    return retArr;
  }

  return retArr;
},
streamMistralCompletion : async function(myMessages, requestType, currTab) {
    
    if(CHATGPT.DB_LOGGING && currTab['has_ui']){          
        let mockResponse = await mysqlLogger.getMockResponse(requestType, currTab['task_text'], "mistral");
        if (mockResponse) {
          console.log('TASK COMPLETION Mocking response:', mockResponse);       
          let fullText = mockResponse['response'];
          CHATGPT.htmlText = fullText;
          CHATGPT.postProcessHtmlText(fullText, currTab['uid']);  
          if(requestType == 'do_new_programming_task'){
            let finalMessage = {"role": "assistant", "content": fullText};          
            myMessages.push(finalMessage);          
            let tabIndex = TABS.getIndexByUid(currTab['uid']);
            TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
          }      
          return mockResponse; 
        }      
    }        
  let fullText = '';
  CHATGPT.htmlText = "";
  CHATGPT.codeLength = 0;
  let modelSettings = CHATGPT.mistral_model_properties[requestType];
  let fullPrompt = "";
  for (let anObj of myMessages) {
      if (anObj.role === 'user') {
        fullPrompt += anObj.content;
      }
  }
   

  if(CHATGPT.DB_LOGGING){
    mysqlLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider']);
  }
  if(CHATGPT.DEXIE_LOGGING){
    CHATGPT.dexieLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
  }     
  if(CHATGPT.FILE_LOGGING){
    const fileSending = `sending_${requestType}_${getFormattedDateTime()}.txt`;
    const logSendingPath = path.join('.', 'logs', fileSending);
    const jsonMyMessages = JSON.stringify(myMessages, null, 2);
    fs.writeFileSync(logSendingPath, jsonMyMessages, 'utf8');      
  }

  const retArr = {
    success: true,
    total_tokens: 0, // This needs to be calculated if required
    orig_response: '',
    response: null
  };

  try {
      console.log("this is really mistral");
      //console.log(CHATGPT.mistral);
    const chatStreamResponse = await CHATGPT.mistral.chatStream({
      model: CHATGPT.LLM_MODEL,
      max_tokens: modelSettings['tokens'],
      temperature: 0.0,
      messages: myMessages
    });

    for await (const chunk of chatStreamResponse) {
      if (chunk.choices[0].delta.content !== undefined) {
        const streamText = chunk.choices[0].delta.content;
        fullText += streamText;
        CHATGPT.updateHtmlText(streamText, currTab['uid']);
      }
    }

    CHATGPT.finalizeHtmlText(currTab['uid']);
    retArr.orig_response = fullText;
    
    let finalMessage = { "role": "assistant", "content": fullText };
    myMessages.push(finalMessage);
    let tabIndex = TABS.getIndexByUid(currTab['uid']);
    TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
    console.log("TAB UID " + currTab['uid'] + " STORED MESSAGES: ");
    console.log(TABS.TAB_ARR[tabIndex]['messages']);
    
    let inputTokens = COST.estimateTokens(fullPrompt);
    let outputTokens = COST.estimateTokens(fullText);
    retArr.response = fullText; // Modify this based on the structure you need
    try{
        if(CHATGPT.DB_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            mysqlLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, fullText, retData, currTab['uid'], currTab['llm_provider']);
        }
        if(CHATGPT.DEXIE_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            CHATGPT.dexieLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, fullText, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL, inputTokens, outputTokens);
        }           
        if(CHATGPT.FILE_LOGGING){
            const fileName = `${requestType}_${getFormattedDateTime()}.txt`;
            const logFilePath = path.join('.', 'logs', fileName);
            fs.writeFileSync(logFilePath, fullText, 'utf8');      
        }
    }catch(le){
        console.log("log error");
        console.log(le);
    }
  } catch (e) {
    console.log("error");
    console.log(e);
    console.log("MESSAGEs WAS");
    console.log(myMessages);
    const retArr = {};
    retArr.success = false;
    retArr.error = e.toString();        
    return retArr;  
  }

  return retArr;
},  
streamGroqCompletion : async function(myMessages, requestType, currTab) {
    if(CHATGPT.DB_LOGGING && currTab['has_ui']){          
        let mockResponse = await mysqlLogger.getMockResponse(requestType, currTab['task_text'], "groq");
        if (mockResponse) {
          console.log('TASK COMPLETION Mocking response:', mockResponse);       
          let fullText = mockResponse['response'];
          CHATGPT.htmlText = fullText;
          CHATGPT.postProcessHtmlText(fullText, currTab['uid']);
          if(requestType == 'do_new_programming_task'){
            let finalMessage = {"role": "assistant", "content": fullText};          
            myMessages.push(finalMessage);          
            let tabIndex = TABS.getIndexByUid(currTab['uid']);
            TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
          }      
    
          
          return mockResponse; 
        }      
    }        
  let fullText = '';
  CHATGPT.htmlText = "";
  CHATGPT.codeLength = 0;
  let modelSettings = CHATGPT.groq_model_properties[requestType];
  let fullPrompt = "";
  for (let anObj of myMessages) {
    if (anObj.role === 'user') {
      fullPrompt += anObj.content;
    }
  }
  

  if(CHATGPT.DB_LOGGING){
    mysqlLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider']);
  }
  if(CHATGPT.DEXIE_LOGGING){
    CHATGPT.dexieLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
  }       
  if(CHATGPT.FILE_LOGGING){
    const fileSending = `sending_${requestType}_${getFormattedDateTime()}.txt`;
    const logSendingPath = path.join('.', 'logs', fileSending);
    const jsonMyMessages = JSON.stringify(myMessages, null, 2);
    fs.writeFileSync(logSendingPath, jsonMyMessages, 'utf8');      
  }

  const retArr = {
    success: true,
    total_tokens: 0, // This needs to be calculated if required
    orig_response: '',
    response: null
  };

  try {
      console.log("this is really groq");
      //console.log(CHATGPT.groq);

    const chatStreamResponse = await CHATGPT.groq.completions.create({
      model: CHATGPT.LLM_MODEL,
      temperature: 0.1,
      messages: myMessages,
      stream: true
    });

    for await (const chunk of chatStreamResponse){
      console.log("groq got chunk");
      console.log(groq);
      if (chunk.choices[0].delta.content !== undefined) {
        const streamText = chunk.choices[0].delta.content;
        fullText += streamText;
        CHATGPT.updateHtmlText(streamText, currTab['uid']);
      }
    }
    
    CHATGPT.finalizeHtmlText(currTab['uid']);    

    retArr.orig_response = fullText;
    retArr.response = fullText; // Modify this based on the structure you need    
    let finalMessage = { "role": "assistant", "content": fullText };
    myMessages.push(finalMessage);
    let tabIndex = TABS.getIndexByUid(currTab['uid']);
    TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
    console.log("TAB UID " + currTab['uid'] + " STORED MESSAGES: ");
    console.log(TABS.TAB_ARR[tabIndex]['messages']);    

    let inputTokens = COST.estimateTokens(fullPrompt);
    let outputTokens = COST.estimateTokens(fullText);
    
    if(CHATGPT.DB_LOGGING){
        let retData = JSON.stringify(retArr);
        console.log("uid: " + currTab['uid']);
        mysqlLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, fullText, retData, currTab['uid'], currTab['llm_provider']);
    }
    if(CHATGPT.DEXIE_LOGGING){
        let retData = JSON.stringify(retArr);
        console.log("uid: " + currTab['uid']);
        CHATGPT.dexieLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, fullText, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL, inputTokens, outputTokens);
    }        
    if(CHATGPT.FILE_LOGGING){
        const fileName = `${requestType}_${getFormattedDateTime()}.txt`;
        const logFilePath = path.join('.', 'logs', fileName);
        fs.writeFileSync(logFilePath, fullText, 'utf8');      
    }
  } catch (e) {
    console.log("error");
    console.log(e);
    console.log("MESSAGEs WAS");
    console.log(myMessages);
    const retArr = {};
    retArr.success = false;
    retArr.error = e.toString();    
    return retArr;  
  }

  return retArr;
},  
  normalAnthropicCompletion: async function(myMessages, requestType, currTab, jsonMode) {
      console.log("in anthropic chat completiion");
    let fullText = '';    
    CHATGPT.htmlText = "";
    CHATGPT.codeLength = 0;
    let fullPrompt = "";
    for (let anObj of myMessages) {
      if (anObj.role === 'user') {
        fullPrompt += anObj.content;
      }
    }
    
    if(CHATGPT.DB_LOGGING && currTab['has_ui']){  
        if(!CHATGPT.NO_CACHE_REQUESTS.includes(requestType)){
            let mockResponse = await mysqlLogger.getMockResponse(requestType, currTab['task_text'], "anthropic");
            if (mockResponse) {
              console.log('Mocking response:', mockResponse);
              return mockResponse; // or you can use `return { success: true, response: mockResponse };` depending on your requirements
            }      
        }
        else if(requestType == 'single_file_partial_change' || requestType == 'new_code_placement'){
            //we need to debug this by replaying 
            let mockResponse = await mysqlLogger.getMockByFullPrompt(requestType, fullPrompt, "anthropic");
            if (mockResponse) {
              console.log('GOT ' + requestType + ' FILE CHANGE MOCK RESPONSE:', mockResponse);
              return mockResponse; // or you can use `return { success: true, response: mockResponse };` depending on your requirements
            }           
        }
    }    
    
    let maxTokens = 4096;
    
    if(CHATGPT.SHORT_TOKENS.includes(requestType)){
        maxTokens = 2048;
    }    

    if(CHATGPT.DB_LOGGING){
        //change TABS.CURRENT_TASK and CHATGPT.LLM_PROVIDER TO USE THE ONE FROM TABS
        //console.log("logging full prompt");
        //console.log(fullPrompt);
        mysqlLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider']);
    }
    if(CHATGPT.DEXIE_LOGGING){
      CHATGPT.dexieLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
    }        
    if(CHATGPT.FILE_LOGGING){
        const fileSending = `sendinganthropic_${requestType}_${getFormattedDateTime()}.txt`;
        const logSendingPath = path.join('.', 'logs', fileSending);
        const jsonMyMessages = JSON.stringify(myMessages, null, 2);
        fs.writeFileSync(logSendingPath, jsonMyMessages, 'utf8');
    }
    console.log("after write log file");

    const retArr = {};
    let assistantMessage = "";
    
    if(jsonMode){
        myMessages.push({
            "role": "assistant",
            "content": "[{"
        });
    }
    else if(requestType == 'request_relevant_stub_lines' || requestType == 'formalize_code_changes' || requestType == 'request_file_descriptions' || requestType == 'request_fix_stub_lines'){
        myMessages.push({
            "role": "assistant",
            "content": "|"
        });        
    }
    
    try {
        let messageObject = {
            model: CHATGPT.LLM_MODEL,
            temperature: 0.1,
            max_tokens: maxTokens,
            messages: myMessages
        };
        if(requestType == 'request_relevant_line_numbers'){
            messageObject['system'] = "You help identify relevant line numbers in project files.  You should only return JSON, DO NOT add any other text.";
        }
        else if(requestType == 'request_relevant_stub_lines'){
            messageObject['system'] = "You help identify relevant line numbers in project files.  Only return relevant lines, do not explain the reasoning or how to do the task.";
            
        }
        
        const response = await CHATGPT.anthropic.messages.create(messageObject);
        //console.log("response");
        //console.log(response);

        if (!response || !response['content']) {
            throw new Error('Bad message response body');
        } 
        
        let messageType = response['content'][0]['type'];
        if(messageType == 'json'){
            assistantMessage = response['content'][0]['text'];
        }else{
            assistantMessage = response['content'][0][messageType];
        }
            
        //console.log("assistant message");
        //console.log(assistantMessage);

        retArr.success = true;
        retArr.orig_response = assistantMessage;
        
        let inputTokens = 0;
        let outputTokens = 0;
        if(response.usage){            
            retArr.input_tokens = response.usage.input_tokens;
            inputTokens = response.usage.input_tokens;
            retArr.output_tokens = response.usage.output_tokens;
            outputTokens = response.usage.output_tokens;
        }
        retArr.model = CHATGPT.LLM_MODEL;        

        if (requestType === 'request_relevant_files') {
            retArr.response = CHATGPT.parseFilesList(assistantMessage, currTab);
        } else if (requestType === 'request_relevant_line_numbers') {
            console.log("parsing");
            assistantMessage = "[{ " + assistantMessage;
            console.log(assistantMessage);
            let assistantJson = CHATGPT.parseMixedResponseToJsonArray(assistantMessage, "file_path");
            if(assistantJson.length <= 0){
                retArr.success = false;
                retArr.error = "BAD JSON";
                return retArr;
            }

            console.log("returning");
            console.log(assistantJson);
            retArr.response = assistantJson;
        } 
        else if (requestType === 'request_relevant_stub_lines' || requestType == 'request_fix_stub_lines') {
            //add a pipe to start
            assistantMessage = "|" + assistantMessage;
            let stubLineResp = CHATGPT.parseStubFileLines(assistantMessage); 
            if(stubLineResp.length <= 0){
                retArr.success = false;
                retArr.error = "BAD STUB LINES REPONSE";
                console.log(assistantMessage);
                return retArr;
            }

            console.log("anthropic returning");
            console.log(stubLineResp);
            retArr.response = stubLineResp;
        }         
        else if(requestType == 'do_new_programming_task'){
            retArr.response = assistantMessage;
        }
        else if(requestType == 'single_file_full_change' || requestType == 'single_function_full_change' || requestType == 'single_file_partial_change'){
            retArr.response = CHATGPT.getJustCode(assistantMessage, currTab['just_files_linux']);           
        }
        else if(requestType == 'new_code_placement'){
            retArr.response = CHATGPT.parseNewCodePlacement(assistantMessage);
        }        
        else if(requestType == 'formalize_code_changes'){
            assistantMessage = "|" + assistantMessage;
            console.log("formal change assistantMessage");
            console.log(assistantMessage);
            let formalChangeArr = CHATGPT.parseFormalizedChanges(assistantMessage);

            if(formalChangeArr.length <= 0){
                console.log("invalid formal changes");
                retArr.success = false;
                retArr.error = "Invalid JSON for formalizing response: " + assistantMessage;     

                //we need to parse alternative

                return retArr;
            }        
            retArr.response = formalChangeArr;   
        }
        else if (requestType === 'request_file_descriptions') {
            assistantMessage = "|" + assistantMessage;
            retArr.response = CHATGPT.parseFileDescriptions(assistantMessage);
        }  
        else if (requestType === 'prune_unlikely_project_files') {
            retArr.response = CHATGPT.parsePruneUnlikelyProjectFiles(assistantMessage);
        }            
        if(CHATGPT.DB_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            mysqlLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider']);
        }
        if(CHATGPT.DEXIE_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            CHATGPT.dexieLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL,inputTokens,outputTokens);
        }             
        if(CHATGPT.FILE_LOGGING){
            const fileName = `anthropic_${requestType}_${getFormattedDateTime()}.txt`;
            const logFilePath = path.join('.', 'logs', fileName);
            fs.writeFileSync(logFilePath, assistantMessage, 'utf8');
            console.log("done writing logs");
        }
        
    } catch (error) {
        const retArr = {};
        console.error("API request error:", error);
        retArr.success = false;
        retArr.error = "API Error: " + error;
        return retArr;
    }

    return retArr;
      
  },
  normalGeminiCompletion: async function(myMessages, requestType, currTab, jsonMode) {
      console.log("in gemini chat completiion");
    if(CHATGPT.DB_LOGGING && currTab['has_ui']){  
        if(!CHATGPT.NO_CACHE_REQUESTS.includes(requestType)){
            let mockResponse = await mysqlLogger.getMockResponse(requestType, currTab['task_text'], currTab['llm_provider']);
            if (mockResponse) {
              console.log('Mocking response:', mockResponse);
              return mockResponse; // or you can use `return { success: true, response: mockResponse };` depending on your requirements
            }      
        }
        else if(requestType == 'single_file_partial_change' || requestType == 'new_code_placement'){
            //we need to debug this by replaying 
            let mockResponse = await mysqlLogger.getMockByFullPrompt(requestType, fullPrompt, "gemini");
            if (mockResponse) {
              console.log('GOT ' + requestType + ' FILE CHANGE MOCK RESPONSE:', mockResponse);
              return mockResponse; // or you can use `return { success: true, response: mockResponse };` depending on your requirements
            }           
        }        
    }
    let fullText = '';    
    CHATGPT.htmlText = "";
    CHATGPT.codeLength = 0;
    let fullPrompt = "";
    
    //build google format content array out of myMessages
    let contentArr = [];
    for (let anObj of myMessages) {
      if (anObj.role === 'user') {
          
          let theText = anObj.content;
          fullPrompt += anObj.content;
          contentArr.push({
            parts: [{
              text: theText
            }],
            role: "user"
          });
      }else{
          let theText = anObj.content;
          contentArr.push({
            parts: [{
              text: theText
            }],
            role: "model"
          });          
      }
    }
    
    console.log("gemini content array");
    console.log(contentArr);

    
    //let modelSettings = CHATGPT.[requestType];
    //console.log("modelSettings");
    //console.log(modelSettings);
    if(CHATGPT.DB_LOGGING){
        //change TABS.CURRENT_TASK and CHATGPT.LLM_PROVIDER TO USE THE ONE FROM TABS
        mysqlLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider']);
    }
    if(CHATGPT.DEXIE_LOGGING){
      CHATGPT.dexieLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
    }        

    console.log("after write log file");

    const retArr = {};
    let assistantMessage = "";
    let inputTokens = 0;
    let outputTokens = 0;       

    try {
        const data = {
          contents: contentArr
        };        
        
        console.log("gemini model : " + CHATGPT.LLM_MODEL);
        
        let url = 'https://generativelanguage.googleapis.com/v1beta/models/' + CHATGPT.LLM_MODEL + ':generateContent?key=' + CHATGPT.GEMINI_KEY;
        let theresponse = null; 
     
        try{
            const response = await fetch(url, {
              method: 'POST',
              headers: {
                'Content-Type': 'application/json'
              },
              body: JSON.stringify(data)
            });

            theresponse = await response.json();
            console.log("response");
            console.log(theresponse);
            assistantMessage = theresponse.candidates[0].content.parts[0].text;

            console.log("gemini assistant message");
            console.log(assistantMessage);

            retArr.success = true;
            retArr.orig_response = assistantMessage;
            
            if (theresponse.usageMetadata) {
              retArr.input_tokens = theresponse.usageMetadata.promptTokenCount;
              inputTokens = theresponse.usageMetadata.promptTokenCount;
              retArr.output_tokens = theresponse.usageMetadata.candidatesTokenCount;
              outputTokens = theresponse.usageMetadata.candidatesTokenCount;
            }
            retArr.model = CHATGPT.LLM_MODEL;

            retArr.success = true;
            retArr.orig_response = assistantMessage;

            retArr.model = CHATGPT.LLM_MODEL;        
        }catch(mmmm){
            retArr.success = false;
            retArr.error = "GEMINI ERROR REPONSE";
            console.log(mmmm);
            return retArr;
        }

        if (requestType === 'request_relevant_files') {
            retArr.response = CHATGPT.parseFilesList(assistantMessage, currTab);
        } else if (requestType === 'request_relevant_line_numbers') {
            console.log("parsing");
            assistantMessage = "[{ " + assistantMessage;
            console.log(assistantMessage);
            let assistantJson = CHATGPT.parseMixedResponseToJsonArray(assistantMessage, "file_path");
            if(assistantJson.length <= 0){
                retArr.success = false;
                retArr.error = "BAD JSON";
                return retArr;
            }

            console.log("returning");
            console.log(assistantJson);
            retArr.response = assistantJson;
        } 
        else if (requestType === 'request_relevant_stub_lines' || requestType == 'request_fix_stub_lines') {
            let stubLineResp = CHATGPT.parseStubFileLines(assistantMessage); 
            if(stubLineResp.length <= 0){
                retArr.success = false;
                retArr.error = "BAD STUB LINES REPONSE";
                console.log(assistantMessage);
                return retArr;
            }

            console.log("anthropic returning");
            console.log(stubLineResp);
            retArr.response = stubLineResp;
        }         
        else if(requestType == 'do_new_programming_task'){
            retArr.response = assistantMessage;
        }
        else if(requestType == 'single_file_full_change' || requestType == 'single_function_full_change' || requestType == 'single_file_partial_change'){
            retArr.response = CHATGPT.getJustCode(assistantMessage, currTab['just_files_linux']);           
        }
        else if(requestType == 'new_code_placement'){
            retArr.response = CHATGPT.parseNewCodePlacement(assistantMessage);
        }        
        else if(requestType == 'formalize_code_changes'){
            console.log("formal change assistantMessage");
            console.log(assistantMessage);
            let formalChangeArr = CHATGPT.parseFormalizedChanges(assistantMessage);

            if(formalChangeArr.length <= 0){
                console.log("invalid formal changes");
                retArr.success = false;
                retArr.error = "Invalid JSON for formalizing response: " + assistantMessage;     

                //we need to parse alternative

                return retArr;
            }        
            retArr.response = formalChangeArr;   
        }
        else if (requestType === 'request_file_descriptions') {
            assistantMessage = "|" + assistantMessage;
            retArr.response = CHATGPT.parseFileDescriptions(assistantMessage);
        }        
        else if (requestType === 'prune_unlikely_project_files') {
            retArr.response = CHATGPT.parsePruneUnlikelyProjectFiles(assistantMessage);
        }          
        if(CHATGPT.DB_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            mysqlLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider']);
        }
        if(CHATGPT.DEXIE_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            CHATGPT.dexieLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, assistantMessage, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL, inputTokens, outputTokens);
        }             
        if(CHATGPT.FILE_LOGGING){
            const fileName = `anthropic_${requestType}_${getFormattedDateTime()}.txt`;
            const logFilePath = path.join('.', 'logs', fileName);
            fs.writeFileSync(logFilePath, assistantMessage, 'utf8');
            console.log("done writing logs");
        }
        
    } catch (error) {
        const retArr = {};
        console.error("Gemini API request error:", error);
        retArr.success = false;
        retArr.error = "Gemini API Error: " + error;
        return retArr;
    }

    return retArr;
      
  },  
  
streamAnthropicCompletion: async function(myMessages, requestType, currTab) {

    let taskText = currTab['task_text'];
    let lastPrompt = "";    
    if(requestType == 'do_additional_programming_task'){        
        for (let anObj of myMessages) {
          if (anObj.role === 'user') {
            lastPrompt = anObj.content;
          }
        }            
        console.log("Last prompt of additional message: " + lastPrompt);
        taskText = lastPrompt;
    }  
    if(CHATGPT.DB_LOGGING && currTab['has_ui']){          
        let mockResponse = await mysqlLogger.getMockResponse(requestType, taskText, "anthropic");
        if (mockResponse) {
          console.log('TASK COMPLETION Mocking response:', mockResponse);       
          let fullText = mockResponse['response'];
          CHATGPT.htmlText = fullText;
          CHATGPT.postProcessHtmlText(fullText, currTab['uid']);
          if(requestType == 'do_new_programming_task'){
            let finalMessage = {"role": "assistant", "content": fullText};          
            myMessages.push(finalMessage);          
            let tabIndex = TABS.getIndexByUid(currTab['uid']);
            TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
          }
          return mockResponse; 
        }      
    }    
  let fullText = '';
  CHATGPT.htmlText = "";
  CHATGPT.codeLength = 0;
  //let modelSettings = CHATGPT.anthropic_model_properties[requestType];
  //let useTokens = modelSettings['use_tokens'];
  let maxTokens = 4096;

  if(CHATGPT.SHORT_TOKENS.includes(requestType)){
      maxTokens = 2048;
  }  
  

  let requestBody = {
      messages: myMessages,
      model: CHATGPT.LLM_MODEL,
      temperature: 0.1,
      max_tokens: maxTokens,
      stream: true
    };
  let fullPrompt = "";
  let inputTokens = 0;
  let outputTokens = 0;
  for (let anObj of myMessages) {
    if (anObj.role === 'user') {
      fullPrompt += anObj.content;
    }
    if(anObj.usage){
        console.log("usage object");
        console.log(anObj.usage);
        if(anObj.usage.input_tokens){
            inputTokens += anObj.usage.input_tokens;
        }
        if(anObj.usage.input_tokens){
            outputTokens += anObj.usage.output_tokens;
        }        
    }
  }
  
  let savePrompt = fullPrompt;
  if(lastPrompt != ''){
      savePrompt = lastPrompt;
  }
   

  if(CHATGPT.DB_LOGGING){
    //change TABS.CURRENT_TASK and CHATGPT.LLM_PROVIDER TO USE THE ONE FROM TABS
    mysqlLogger.logRequest(1, requestType, taskText, savePrompt, "", null, currTab['llm_provider']);
  }
  if(CHATGPT.DEXIE_LOGGING){
    CHATGPT.dexieLogger.logRequest(1, requestType, currTab['task_text'], savePrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
  }      
  if(CHATGPT.FILE_LOGGING){      
    const fileSending = `sending_${requestType}_${getFormattedDateTime()}.txt`;
    const logSendingPath = path.join('.', 'logs', fileSending);
    const jsonMyMessages = JSON.stringify(myMessages, null, 2);
    fs.writeFileSync(logSendingPath, jsonMyMessages, 'utf8');      
  }

  const retArr = {
    success: true,
    total_tokens: 0, // This needs to be calculated if required
    orig_response: '',
    response: null
  };
  console.log("starting"); 

  try {
    //const stream = await CHATGPT.anthropic.messages.create(requestBody);
    
    let fullText = "";
    
    const stream = CHATGPT.anthropic.messages
      .stream(requestBody)
      .on('text', (text) => {
          //console.log("another");
          //console.log(text);     
          if(text){
            fullText += text;
            CHATGPT.updateHtmlText(text, currTab['uid']);              
          }
      });

    let finalObject = await stream.finalMessage();    
    fullText = finalObject['content'][0]['text'];

    if(finalObject.usage){            
        console.log("anthropic final object has usage");
        console.log(finalObject);
        
        retArr.input_tokens = finalObject.usage.input_tokens;
        retArr.output_tokens = finalObject.usage.output_tokens;
        inputTokens = finalObject.usage.input_tokens;
        outputTokens = finalObject.usage.output_tokens;
    }else{
        console.log("anthropic final object NOOOOOOOOO usage");
        console.log(finalObject);        
    }    
    CHATGPT.finalizeHtmlText(currTab['uid']);
    retArr.orig_response = fullText;
    retArr.response = fullText; // Modify this based on the structure you need
    
    let finalMessage = { "role": "assistant", "content": fullText };
    myMessages.push(finalMessage);
    let tabIndex = TABS.getIndexByUid(currTab['uid']);
    TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
    console.log("TAB UID " + currTab['uid'] + " STORED MESSAGES: ");
    console.log(TABS.TAB_ARR[tabIndex]['messages']);    
    
    try{
        if(CHATGPT.DB_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            mysqlLogger.logResponse(1, requestType, taskText, savePrompt, fullText, retData, currTab['uid'], currTab['llm_provider']);
        }
        if(CHATGPT.DEXIE_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            CHATGPT.dexieLogger.logResponse(1, requestType, currTab['task_text'], savePrompt, fullText, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL, inputTokens, outputTokens);
        }         
        if(CHATGPT.FILE_LOGGING){
            const fileName = `${requestType}_${getFormattedDateTime()}.txt`;
            const logFilePath = path.join('.', 'logs', fileName);
            fs.writeFileSync(logFilePath, fullText, 'utf8');      
        }
    }catch(le){
        console.log("log error");
        console.log(le);
    }        
  } catch (e) {
    console.log("error");
    console.log(e);
    console.log("MESSAGEs WAS");
    console.log(myMessages);
    const retArr = {};
    retArr.success = false;
    retArr.error = e.toString();    
    return retArr;    
    //let finalMessage = {"role": "assistant", "content": fullText};
    //CHATGPT.existingMessages.push(finalMessage);
  }

  console.log("returning retArr");
  console.log(retArr);

  return retArr;
},
streamChatCompletion: async function (myMessages, requestType, currTab) {
    
    let taskText = currTab['task_text'];
    let lastPrompt = "";    
    if(requestType == 'do_additional_programming_task'){        
        for (let anObj of myMessages) {
          if (anObj.role === 'user') {
            lastPrompt = anObj.content;
          }
        }            
        console.log("Last prompt of additional message: " + lastPrompt);
        taskText = lastPrompt;
    }    
    
    if(CHATGPT.DB_LOGGING && currTab['has_ui']){          
        let mockResponse = await mysqlLogger.getMockResponse(requestType, taskText, currTab['llm_provider']);
        if (mockResponse) {
          console.log('TASK COMPLETION Mocking response:', mockResponse);       
          let fullText = mockResponse['response'];
          CHATGPT.htmlText = fullText;
          CHATGPT.postProcessHtmlText(fullText, currTab['uid']);
          if(requestType == 'do_new_programming_task'){
            let finalMessage = {"role": "assistant", "content": fullText};          
            myMessages.push(finalMessage);          
            let tabIndex = TABS.getIndexByUid(currTab['uid']);
            TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
          }      
          
          return mockResponse; 
        }      
    }       
  let fullText = '';
  CHATGPT.htmlText = "";
  CHATGPT.codeLength = 0;
  let modelSettings = CHATGPT.model_properties[requestType];
  let fullPrompt = "";
  for (let anObj of myMessages) {
    if (anObj.role === 'user') {
      fullPrompt += anObj.content;
    }
  }
  let savePrompt = fullPrompt;
  if(lastPrompt != ''){
      savePrompt = lastPrompt;
  }
  

  if(CHATGPT.DB_LOGGING){
    mysqlLogger.logRequest(1, requestType, taskText, savePrompt, "", null, currTab['llm_provider']);
  }
  if(CHATGPT.DEXIE_LOGGING){
    CHATGPT.dexieLogger.logRequest(1, requestType, taskText, savePrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
  }       
  if(CHATGPT.FILE_LOGGING){
    const fileSending = `sending_${requestType}_${getFormattedDateTime()}.txt`;
    const logSendingPath = path.join('.', 'logs', fileSending);
    const jsonMyMessages = JSON.stringify(myMessages, null, 2);
    fs.writeFileSync(logSendingPath, jsonMyMessages, 'utf8');      
  }

  const retArr = {
    success: true,
    total_tokens: 0, // This needs to be calculated if required
    orig_response: '',
    response: null
  };
  let inputTokens = 0;
  let outputTokens = 0;  
  try {
      console.log("this is really chatgpt");
      console.log("MODEL: " + CHATGPT.LLM_MODEL);
      
      //
    let myTemp = 0.1;
    if(CHATGPT.LLM_MODEL.indexOf('o1') >= 0){     
        //o1 models need temp of 1.0?
        myTemp = 1.0;
    }
    
    let requestObject = {
      temperature: myTemp,
      messages: myMessages,
      stream: true      
    };
    if(CHATGPT.LLM_MODEL.startsWith('http')){
        console.log("USING CUSTOM URL");
        let modelParts = CHATGPT.LLM_MODEL.split("|");
        if(modelParts.length > 0){
            requestObject.model = modelParts[1];
        }     
    }else{
        requestObject.model = CHATGPT.LLM_MODEL;
        requestObject.stream_options = { include_usage: true };
    }
    const chatStreamResponse = await CHATGPT.openai.chat.completions.create(requestObject);
    

    console.log(requestObject.model + " getting chunks");
    for await (const chunk of chatStreamResponse){
      
      //console.log(chunk);
      if (chunk.choices && chunk.choices.length > 0) {
          // Check if delta object exists within the first choice
          if (chunk.choices[0].delta) {
            // Now it's safe to access delta.content
            const streamText = chunk.choices[0].delta.content;             
            fullText += streamText;
            CHATGPT.updateHtmlText(streamText, currTab['uid']);
          } else {
            console.log("Delta is undefined or missing.");
            // Optionally, handle the case where delta is missing
          }
          // Check if finish_reason is defined to determine if the stream has ended
          if (chunk.choices[0].finish_reason) {
            console.log(`Stream finished with reason: ${chunk.choices[0].finish_reason}`);
            //stream_options: {"include_usage": true}
            // Handle the end of the stream here, such as breaking the loop
            //break;
          }
        } else {
          console.log("Choices array is empty or undefined - could be usage chunk");
          if(chunk.usage){            
              retArr.input_tokens = chunk.usage.prompt_tokens;
              inputTokens = chunk.usage.prompt_tokens;
              retArr.output_tokens = chunk.usage.completion_tokens;
              outputTokens = chunk.usage.completion_tokens;              
          }               
          // Handle the case where choices array is empty or undefined
      }
    }
    
    CHATGPT.finalizeHtmlText(currTab['uid']);    

    retArr.orig_response = fullText;
    retArr.response = fullText; // Modify this based on the structure you need    
    let finalMessage = { "role": "assistant", "content": fullText };
    myMessages.push(finalMessage);
    let tabIndex = TABS.getIndexByUid(currTab['uid']);
    TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
    console.log("TAB UID " + currTab['uid'] + " STORED MESSAGES: ");
    console.log(TABS.TAB_ARR[tabIndex]['messages']);    
    try{
        if(CHATGPT.DB_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            mysqlLogger.logResponse(1, requestType, taskText, savePrompt, fullText, retData, currTab['uid'], currTab['llm_provider']);
        }
        if(CHATGPT.DEXIE_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            CHATGPT.dexieLogger.logResponse(1, requestType, taskText, savePrompt, fullText, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL,inputTokens,outputTokens);
        }        
        if(CHATGPT.FILE_LOGGING){
            const fileName = `${requestType}_${getFormattedDateTime()}.txt`;
            const logFilePath = path.join('.', 'logs', fileName);
            fs.writeFileSync(logFilePath, fullText, 'utf8');      
        }
    }catch(le){
        console.log("log error");
        console.log(le);
    }    
  } catch (e) {
    console.log("error");
    console.log(e);
    console.log("MESSAGEs WAS");
    console.log(myMessages);
    const retArr = {};
    retArr.success = false;
    retArr.error = e.toString();    
    return retArr;  
  }

  return retArr;
},
streamGeminiCompletion: async function (myMessages, requestType, currTab) {
    if(CHATGPT.DB_LOGGING && currTab['has_ui']){          
        let mockResponse = await mysqlLogger.getMockResponse(requestType, currTab['task_text'], currTab['llm_provider']);
        if (mockResponse) {
          console.log('TASK COMPLETION Mocking response:', mockResponse);       
          let fullText = mockResponse['response'];
          CHATGPT.htmlText = fullText;
          CHATGPT.postProcessHtmlText(fullText, currTab['uid']);
          if(requestType == 'do_new_programming_task'){
            let finalMessage = {"role": "assistant", "content": fullText};          
            myMessages.push(finalMessage);          
            let tabIndex = TABS.getIndexByUid(currTab['uid']);
            TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
          }      
          
          return mockResponse; 
        }      
    }       
  let fullText = '';
  CHATGPT.htmlText = "";
  CHATGPT.codeLength = 0;
  let modelSettings = CHATGPT.model_properties[requestType];
  let fullPrompt = "";
  for (let anObj of myMessages) {
    if (anObj.role === 'user') {
      fullPrompt += anObj.content;
    }
  }
    let newLine = "\n";
    //build google format content array out of myMessages
    let contentArr = [];
    for (let anObj of myMessages) {
      if (anObj.role === 'user') {
          
          let theText = anObj.content;
          contentArr.push({
            parts: [{
              text: theText
            }],
            role: "user"
          });
      }else{
          let theText = anObj.content;
          contentArr.push({
            parts: [{
              text: theText
            }],
            role: "model"
          });          
      }
    }  
    console.log("Gemini content array");  

  if(CHATGPT.DB_LOGGING){
    mysqlLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider']);
  }
  if(CHATGPT.DEXIE_LOGGING){
    CHATGPT.dexieLogger.logRequest(1, requestType, currTab['task_text'], fullPrompt, "", null, currTab['llm_provider'], CHATGPT.LLM_MODEL);
  }       
  if(CHATGPT.FILE_LOGGING){
    const fileSending = `sending_${requestType}_${getFormattedDateTime()}.txt`;
    const logSendingPath = path.join('.', 'logs', fileSending);
    const jsonMyMessages = JSON.stringify(myMessages, null, 2);
    fs.writeFileSync(logSendingPath, jsonMyMessages, 'utf8');      
  }

  const retArr = {
    success: true,
    total_tokens: 0, // This needs to be calculated if required
    orig_response: '',
    response: null
  };

  try {
    console.log("this is really gemini");
    const data = {
      contents: contentArr
    };          
    
      //console.log(CHATGPT.groq);
    const url = 'https://generativelanguage.googleapis.com/v1beta/models/' + CHATGPT.LLM_MODEL + ':streamGenerateContent?alt=sse&key=' + CHATGPT.GEMINI_KEY;

    try {
        const response = await fetch(url, {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json'
            },
            body: JSON.stringify(data)
        });

        const reader = response.body.getReader();
        const decoder = new TextDecoder('utf-8');
        let buffer = '';
        let fullText = '';

        while (true) {
            const { value, done } = await reader.read();
            if (done) {
                console.log('Stream finished');
                break;
            }

            let decodedString = decoder.decode(value, { stream: true });
            buffer += decodedString;

            while (buffer.includes('\n')) {
                let newlineIndex = buffer.indexOf('\n');
                let message = buffer.slice(0, newlineIndex);
                buffer = buffer.slice(newlineIndex + 1);

                if (message.startsWith('data: ')) {
                    message = message.slice(6);
                }

                if (message === '[START]' || message === '[DONE]') {
                    continue;
                }

                try {
                    if (message.trim() !== '') {
                        let json = JSON.parse(message);
                        if (json.candidates && json.candidates[0].content && json.candidates[0].content.parts) {
                            let text = json.candidates[0].content.parts[0].text;
                            fullText += text;
                            CHATGPT.updateHtmlText(text, currTab['uid']);
                            //console.log("Content:", text);
                        } else {
                            console.log("Skipped");
                        }
                    }
                } catch (e) {
                    console.log('Error parsing JSON:', e);
                    console.log('Message was:', message);
                }
            }
        }

        console.log('Final result:' + newLine + fullText);
        CHATGPT.postProcessHtmlText(fullText, currTab['uid']);  
    } catch (error) {
        console.error('Error:', error);
    }
    
    CHATGPT.finalizeHtmlText(currTab['uid']);    

    retArr.orig_response = fullText;
    retArr.response = fullText; // Modify this based on the structure you need    
    let finalMessage = { "role": "assistant", "content": fullText };
    myMessages.push(finalMessage);
    let tabIndex = TABS.getIndexByUid(currTab['uid']);
    TABS.TAB_ARR[tabIndex]['messages'] = myMessages;
    console.log("TAB UID " + currTab['uid'] + " STORED MESSAGES: ");
    console.log(TABS.TAB_ARR[tabIndex]['messages']);    
    
    let inputTokens = COST.estimateTokens(fullPrompt);
    let outputTokens = COST.estimateTokens(fullText);
    
    try{
        if(CHATGPT.DB_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);
            mysqlLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, fullText, retData, currTab['uid'], currTab['llm_provider']);
        }
        if(CHATGPT.DEXIE_LOGGING){
            let retData = JSON.stringify(retArr);
            console.log("uid: " + currTab['uid']);            
            CHATGPT.dexieLogger.logResponse(1, requestType, currTab['task_text'], fullPrompt, fullText, retData, currTab['uid'], currTab['llm_provider'], CHATGPT.LLM_MODEL, inputTokens, outputTokens);
        }        
        if(CHATGPT.FILE_LOGGING){
            const fileName = `${requestType}_${getFormattedDateTime()}.txt`;
            const logFilePath = path.join('.', 'logs', fileName);
            fs.writeFileSync(logFilePath, fullText, 'utf8');      
        }
    }catch(le){
        console.log("log error");
        console.log(le);
    }    
  } catch (e) {
    console.log("error");
    console.log(e);
    console.log("MESSAGEs WAS");
    console.log(myMessages);
    const retArr = {};
    retArr.success = false;
    retArr.error = e.toString();        
    return retArr;  
  }

  return retArr;
},
  fetchEngines: async function(){
    const responseEngines = await CHATGPT.openai.listEngines();
    console.log("response");
    console.log(responseEngines);
  }  
};
//lets put the prompts at the bottom of the file
CHATGPT.doNewProgrammingTask = async function(theTask, codeContextText, doStream, projectName, projectDescription, tabUid){
    
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "do_new_programming_task", currTab['has_image']);

    console.log("DOING " + currTab['llm_provider'] + " doNewProgrammingTask with tab UID: " + tabUid);
    //let codeFiles = await CHATGPT.prepareFileText();
    //console.log("files");
    //console.log(codeFiles);
    let newLine = "\n";
    
    //should have projectData instead of projectName
    let fullPrompt = "";
    let projectDesc = "";
    if(projectDescription && projectDescription.length > 0){
        projectDesc = "I am working on: " + projectDescription;
        fullPrompt = projectDesc + "\n\n";  
    }else{
        projectDesc = "I want you to help me with my project called " + projectName;
        fullPrompt = projectDesc + "\n\n";  
    }
    let projectRulesArr = currTab['checked_project_rules'];
    

    let messages = new Array();

    let languageFiles = "javascript/html/css";
    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){      
        messages.push({
            role: "system",
            content: "You are a helpful coding assistant that thinks through tasks, and writes working code to complete a task."
        });       
    }
    let databaseActiveText = "";
    let databaseSchemaText = "";
    if(currTab['use_schema']){
        databaseActiveText = " and database tables in that are used in this project";
        databaseSchemaText = "HERE ARE THE DATABASE TABLES AND THEIR FIELDS:" + newLine + currTab['schema_text'] + newLine;
    }
    
    let markdownDocumentActive = "";
    let markdownDocumentText = "";
    if(currTab['use_documents']){
        let relevantMarkdown = await getRelevantDocuments(theTask);
        if(relevantMarkdown.length > 0){
            markdownDocumentActive = " and some documentation that might be useful";
            markdownDocumentText = "POSSIBLE HELPFUL DOCUMENTATION: " + newLine + relevantMarkdown + newLine + newLine;
        }
    } 
    let referenceProjectText = "";
    if(currTab['use_projects'] && currTab['reference_project_text'].length > 0){
        referenceProjectText = currTab['reference_project_text'];
    }
    
    
    fullPrompt += "INSTRUCTIONS: I am going to give you a task and some project files" + databaseActiveText + markdownDocumentActive + ".  You are to update the code of any files needed to complete the task." + 
            " In your response you must output which file you are editing, a new line, and then use triple backticks around the code, for each code block in the response.  " + newLine + newLine;
    fullPrompt += "TASK: " + newLine + theTask + newLine + newLine;
    fullPrompt += "HERE ARE THE CODE FILES: " + newLine + codeContextText + newLine + newLine;
    fullPrompt += databaseSchemaText;
    fullPrompt += markdownDocumentText;  
    fullPrompt += referenceProjectText;
    if(CHATGPT.LLM_PROVIDER == 'openai'){
        fullPrompt += "It's important to only output the parts of the code that needs changes.  Do not print sections of the file that do not need updating, just give the new or modified code needed to complete the task.  Files that don't need changes should not be printed. Placeholders inside comments such as //... existing code can be used for existing code.  Please also explain the reasoning behind the changes in your response." + newLine;
    }else{
        fullPrompt += "Please do not print out the entire file, just give the code needed to complete the task.  Do not print files that don't need changes. Placeholders inside comments such as //... existing code can be used for existing code.  Please also explain the reasoning behind the changes in your response." + newLine;
    }
    CHATGPT.lastRequest = fullPrompt;
    
    
    if(currTab['llm_provider'] == 'anthropic' && currTab['has_image']){
        console.log("HAS IMAGE - YES");
        let contentArr = [];
        for(let c = 0; c < currTab['context_images'].length; c++){
            let newImage = {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": "image/jpeg",
                        "data": currTab['context_images'][c]
                    }
                };
                contentArr.push(newImage);
        }
        contentArr.push({
            "type": "text", 
            "text": fullPrompt
        });
        messages.push({
            role: "user", 
            content: contentArr
        });  
    }
    else if(currTab['llm_provider'] == 'openai' && currTab['has_image']){
        console.log("adding image to chat completion");
        let contentArr = [];
        for(let c = 0; c < currTab['context_images'].length; c++){
            let dataUrl = "data:image/jpeg;base64," + currTab['context_images'][c];
            let newImage = {
                    type: "image_url",
                    image_url: {
                        "url": dataUrl
                    }
                };
                
            contentArr.push(newImage);
        }
        contentArr.push({
            "type": "text", 
            "text": fullPrompt
        });    
        
        messages.push({
            role: "user", 
            content: contentArr
        });          
        
    }    
    else if(currTab['llm_provider'] == 'gemini' && currTab['has_image']){
        //todo - put something in messages so gemini knows to add an image of mimetype
    }
    else{
        //console.log("HAS IMAGE - NO");
        messages.push({
            role: "user",
            content: fullPrompt
        });      
    }    
    
    if(CHATGPT.FILE_LOGGING){
        const fileTask = `task_${getFormattedDateTime()}.txt`;
        const logTaskPath = path.join('.', 'logs', fileTask);
        fs.writeFileSync(logTaskPath, theTask, 'utf8');        
    }

    console.log("sending message");
    console.log(messages);
    if(!doStream){
        console.log("normal do_new_programming_task");
        let normalResponse = await CHATGPT.providerChatCompletion(messages, 'do_new_programming_task', currTab, false);
        let fullText = normalResponse.response;
        let finalMessage = {"role": "assistant", "content": fullText};          
        messages.push(finalMessage);          
        let tabIndex = TABS.getIndexByUid(currTab['uid']);
        TABS.TAB_ARR[tabIndex]['messages'] = messages;        
        return normalResponse;            
    }else{
        if(currTab['llm_provider'] == 'anthropic'){
            let streamResponse = await CHATGPT.streamAnthropicCompletion(messages, 'do_new_programming_task', currTab);
            console.log("chatstream response is");
            console.log(streamResponse);
            return streamResponse;            
        }else if(currTab['llm_provider'] == 'openai'){
            let streamResponse = await CHATGPT.streamChatCompletion(messages, 'do_new_programming_task', currTab);
            console.log("chatstream response is");
            console.log(streamResponse);
            return streamResponse;
        }else if(currTab['llm_provider'] == 'mistral'){
            let streamResponse = await CHATGPT.streamMistralCompletion(messages, 'do_new_programming_task', currTab);
            console.log("chatstream response is");
            console.log(streamResponse);
            return streamResponse;            
        }else if(currTab['llm_provider'] == 'groq'){
            let streamResponse = await CHATGPT.streamGroqCompletion(messages, 'do_new_programming_task', currTab);
            console.log("chatstream response is");
            console.log(streamResponse);
            return streamResponse;            
        }
        else if(currTab['llm_provider'] == 'gemini'){
            let streamResponse = await CHATGPT.streamGeminiCompletion(messages, 'do_new_programming_task', currTab);
            console.log("gemini chatstream response is");
            console.log(streamResponse);
            return streamResponse;            
            
        }
        else if(currTab['llm_provider'] == 'mistral'){
            let streamResponse = await CHATGPT.streamMistralCompletion(messages, 'do_new_programming_task', currTab);
            console.log("codestral chatstream response is");
            console.log(streamResponse);
            return streamResponse;              
        }
        else{
            alert("PROVIDER NOT FOUND");
            return null;
        }
        //readStreamAndUpdateTextArea(chatStream);

    }
}; 

CHATGPT.askCodebaseQuestion = async function(theQuestion, codeContextText, doStream, projectName, tabUid){
    
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "ask_codebase_question", currTab['has_image']);

    console.log("DOING " + currTab['llm_provider'] + " askCodebaseQuestion with tab UID: " + tabUid);
    let newLine = "\n";
    
    let fullPrompt = "";
    let projectDesc = "";
    if(projectName == 'scrapehawkext'){
        projectDesc = "I am building an Chrome extension in javascript.";            
        fullPrompt = projectDesc + "\n\n";            
    }else if(projectName == 'stubs'){
        projectDesc = "I am building an Electron application in javascript.";      
        fullPrompt = projectDesc + "\n\n";            
    }

    let messages = new Array();

    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){      
        messages.push({
            role: "system",
            content: "You are a helpful coding assistant that answers questions"
        });       
    }
    
    if(codeContextText != ""){
        fullPrompt += "INSTRUCTIONS: I am going to give you a question about my codebase and the code files.  Please answer the question as best you can based on the code provided. " 
                 + newLine;
        fullPrompt += "QUESTION: " + theQuestion + newLine + newLine;        
        fullPrompt += "HERE ARE THE CODE FILES: " + newLine + codeContextText + newLine + newLine;
    }
    else{
        //general question
        fullPrompt += theQuestion;
    }
    
    messages.push({
        role: "user",
        content: fullPrompt
    });

    console.log("sending message");
    console.log(messages);
    if(!doStream){
        console.log("normal askCodebaseQuestion");
        let normalResponse = await CHATGPT.providerChatCompletion(messages, 'ask_codebase_question', currTab, false);
        let fullText = normalResponse.response;
        let finalMessage = {"role": "assistant", "content": fullText};          
        messages.push(finalMessage);          
        let tabIndex = TABS.getIndexByUid(currTab['uid']);
        TABS.TAB_ARR[tabIndex]['messages'] = messages;        
        return normalResponse;            
    }else{
        if(currTab['llm_provider'] == 'anthropic'){
            let streamResponse = await CHATGPT.streamAnthropicCompletion(messages, 'ask_codebase_question', currTab);
            console.log("chatstream response is");
            console.log(streamResponse);
            return streamResponse;            
        }else if(currTab['llm_provider'] == 'openai'){
            let streamResponse = await CHATGPT.streamChatCompletion(messages, 'ask_codebase_question', currTab);
            console.log("chatstream response is");
            console.log(streamResponse);
            return streamResponse;
        }else if(currTab['llm_provider'] == 'mistral'){
            let streamResponse = await CHATGPT.streamMistralCompletion(messages, 'ask_codebase_question', currTab);
            console.log("chatstream response is");
            console.log(streamResponse);
            return streamResponse;            
        }else if(currTab['llm_provider'] == 'groq'){
            let streamResponse = await CHATGPT.streamGroqCompletion(messages, 'ask_codebase_question', currTab);
            console.log("chatstream response is");
            console.log(streamResponse);
            return streamResponse;            
        }
        else if(currTab['llm_provider'] == 'gemini'){
            let streamResponse = await CHATGPT.streamGeminiCompletion(messages, 'ask_codebase_question', currTab);
            console.log("chatstream response is");
            console.log(streamResponse);
            return streamResponse;            
            
        }
        else{
            alert("PROVIDER NOT FOUND");
            return null;
        }
    }
}; 

CHATGPT.doAdditionalProgrammingTask = async function(additionalText, tabUid, doStream, newCodeContext){
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "do_additional_programming_task");

    console.log("DOING " + currTab['llm_provider'] + " doAdditionalProgrammingTask with tab UID: " + tabUid);
    
    let newPrompt = additionalText;
    let newLine = "\n";
    if(newCodeContext.length > 0){
        newPrompt += newLine + newLine + "HERE ARE THE CURRENT CODE FILES: " + newLine + newCodeContext;
    }

    let messages = currTab['messages'];
    messages.push({
        role: "user",
        content: newPrompt
    });
    console.log("added messages:");
    console.log(messages);

    if(!doStream){
        console.log("normal do_additional_programming_task");
        let normalResponse = await CHATGPT.providerChatCompletion(messages, 'do_additional_programming_task', currTab, false);
        let fullText = normalResponse.response;
        let finalMessage = {"role": "assistant", "content": fullText};          
        messages.push(finalMessage);          
        let tabIndex = TABS.getIndexByUid(currTab['uid']);
        TABS.TAB_ARR[tabIndex]['messages'] = messages;        
        return normalResponse;            
    }else{
        // Create the chatgptprevious div if it doesn't exist
        if ($('#chatgptprevious_' + tabUid).length === 0) {
            $('<div>', {
                id: 'chatgptprevious_' + tabUid
            }).insertBefore('#chatgptresponse_' + tabUid);
            let prevMessageIndex = 0;
            $('<div id="previousresponse_' + tabUid + '_' + prevMessageIndex + '"><a href="javascript:void(0);" id="showprevious_' + tabUid + '" class="showprevious">Show Previous</a></div>').insertBefore('#chatgptresponse_' + tabUid);
        }
        
        // Save the previous response
        let previousResponse = $('#chatgptresponse_' + tabUid).html();
        $('#chatgptprevious_' + tabUid).append(previousResponse);
        $('#chatgptprevious_' + tabUid).css('display', 'none');
        
        // Clear the current response
        $('#chatgptresponse_' + tabUid).html("");        
        let chatStream = null;
        if(currTab['llm_provider'] == 'openai'){
            chatStream = await CHATGPT.streamChatCompletion(messages, 'do_additional_programming_task', currTab);
        }else if(currTab['llm_provider'] == 'anthropic'){
            chatStream = await CHATGPT.streamAnthropicCompletion(messages, 'do_additional_programming_task', currTab);   
        }else if(currTab['llm_provider'] == 'mistral'){
            chatStream = await CHATGPT.streamMistralCompletion(messages, 'do_additional_programming_task', currTab);   
        }else if(currTab['llm_provider'] == 'groq'){
            chatStream = await CHATGPT.streamGroqCompletion(messages, 'do_additional_programming_task', currTab);   
        }else{
            alert("PROVIDER NOT FOUND");
            return null;
        }
        console.log("chatstream response is");
        console.log(chatStream);
        return chatStream;
    }
};

CHATGPT.requestRelevantFiles = async function(theTask, filesListText, tabUid, isQuestion, referenceProjectData){
    console.log("DOING LLM requestRelevantFiles with tab uid " + tabUid);
    
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "request_relevant_files");
    //console.log("files");
    //console.log(filesListText);
    let newLine = "\n";

    let fullPrompt = "";
    /*
    if(!projectDesc){
        projectDesc = "PROJECT NAME: XXXX PROJECT DESC: I am building an Electron application in javascript.";
        fullPrompt = projectDesc + "\n\n";
    }
    */
   
    let taskWord = "task";
    let taskPhrase = "complete the task";
    if(isQuestion){
        taskWord = "question";
        taskPhrase = "answer the question";
    }

    
    let newLineSection = newLine + "=====================" + newLine;
    let llmProvider = CHATGPT.getLLMProviderForTask('request_relevant_files');
    if(currTab['has_image'] && (llmProvider == 'openai' || llmProvider == 'anthropic')){
        fullPrompt += "INSTRUCTIONS: " + newLineSection + 
                "I am going to give you a list of project files, a " + taskWord + " and an image.  I want you to rank the files, where the top file " + 
                "is the most likely to need changes to " + taskPhrase + ".  Use the image as extra information.  Since you only know the filepaths and not the source code, you have incomplete information, but do " + 
                "your best to provide a list of the full paths ranked, one per line. At most list 10 files, but ideally you would only list files that have a decent chance of needed to be updated to complete the task.  " + 
                "After the file paths I'd like you to describe what's in the image surrounded by quotes.  Those are the only two things I need.  A list of file paths, " + 
                "and then a description of image and why it's gives context for the " + taskWord + ".  Please do not write/say anything else." + newLine + newLine;                
    }else{
        fullPrompt += "INSTRUCTIONS: " + newLineSection + " I am going to give you a list of project files and a " + taskWord + ".  I want you to rank the files, where the top file " + 
                "is the most likely to need changes to complete the " + taskPhrase + ".  Since you only know the filepaths you have incomplete information, but I just want " + 
                "a list of the full paths ranked, one per line.  " + 
                "The list should only be files that have a decent chance (above 20%) of needing to be edited for the " + taskWord + ".  At a minimum provide at least 3 files (and a maximum of 10).  You are to return only the " + 
                "the list of filepaths. Do not return function names, just file paths." + newLine + newLine;        
    }
    
    if(referenceProjectData){
        //let originaProjectName = "";
        fullPrompt += "INSTRUCTIONS: " + newLineSection + " I have a " + taskWord + " and a related project " + referenceProjectData['project_name'] + " that I think might help.  I want you to rank files from that project, where the top file " + 
                "is the most likely to help give insight into the " + taskPhrase + " that I need to do on another project.  So I am using this codebase as a reference to help with the other project.   Since you only know the filepaths you have incomplete information, but I just want " + 
                "a list of the full paths ranked, one per line.  " + 
                "The list should only be files that have a decent chance of helping the " + taskWord + ".  At a minimum provide at least 3 files (and a maximum of 10).  You are to return only the " + 
                "the list of filepaths. Do not return function names, just file paths." + newLine + newLine;          
    }    
   
    let messages = new Array();

    if(isQuestion){
        fullPrompt += "QUESTION: " + newLineSection + theTask + newLine + newLine;
    }else{
        fullPrompt += "TASK: " + newLineSection + theTask + newLine + newLine;
    }
    fullPrompt += "PROJECT FILES: " + newLineSection + newLine + filesListText + newLine + newLine;
    
    if(referenceProjectData){
        fullPrompt += 'Now give me a list of file paths, one per line.  Only include a filepath if you think this file has some relevance.';
    }else{
        fullPrompt += 'Now give me a list of file paths, one per line.  Only include a filepath if the file has a decent chance (above 20%) it will need editing for this specific task.';        
    }    
    
    //put this in array for later
    CHATGPT.originalFileListArr = filesListText.split(/\r?\n/);
    //console.log("arr version");
    //console.log(CHATGPT.originalFileListArr);

    messages.push({
        role: "user",
        content: fullPrompt
    });

    //console.log("sending");
    //console.log(messages);
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'request_relevant_files', currTab, false);
    console.log("provider normal response is");
    console.log(normalResponse);
    return normalResponse;    
      //readStreamAndUpdateTextArea(chatStream);
    
}; 

CHATGPT.requestFileDescriptions = async function(projectName, projectDesc, fullFileSourceCode){
    console.log("DOING LLM requestFileDescriptions ");
    let currentTabUid = TABS.getCurrentUid();
    let tabIndex = TABS.getIndexByUid(currentTabUid);
    let currTab = TABS.TAB_ARR[tabIndex];
    await CHATGPT.initProvider(tabIndex, "request_file_descriptions");
    //console.log("files");
    //console.log(filesListText);
    let newLine = "\n";

    let fullPrompt = "";
    fullPrompt += "INSTRUCTIONS: I'm trying to document what each file does in my codebase.   I am going to give you a project name and description, and a list of project filepaths and their source code." +   
                "I want you to write a description for each file that I give you. " + 
                "The description should be one paragraph of text, and as concise as possible.  It should include the most important functionalities, and the most important interactions with other parts of the codebase. " + 
                "For your response I want you to give the filepath surrounded by pipes, a new line, then the description.  Do that for all files that I give you. Not all files from the codebase are being included so you have incomplete information, " + 
                "but do your best to document what the file does to the best of your ability.  Every file must have a description." + newLine;                

   
    let messages = new Array();


    fullPrompt += "Project Name: " + newLine + projectName + newLine;
    fullPrompt += "Project Desc: " + newLine + projectDesc + newLine;
    fullPrompt += "Project Files: " + newLine + fullFileSourceCode ;
    
    //console.log("arr version");
    //console.log(CHATGPT.originalFileListArr);

    messages.push({
        role: "user",
        content: fullPrompt
    });

    //console.log("sending");
    //console.log(messages);
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'request_file_descriptions', currTab, false);
    console.log("provider normal response is");
    console.log(normalResponse);
    return normalResponse;    
      //readStreamAndUpdateTextArea(chatStream);
    
}; 

CHATGPT.getLibrariesUsed = async function(){    
    //look at package json, look at other, index these docs with crawler
    //for this maybe communicate with my server to get latest documentation for common libraries
    //include in prompt   
};

 CHATGPT.getTerseFileDescriptions = async function(project, filesListText, tabUid){
     let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "get_terse_file_descriptions");
     
     console.log("DOING " + currTab['llm_provider'] + " getTerseFileDescriptions");
     
     let newLine = "\n";
     let fullPrompt = "INSTRUCTIONS: I am going to give you a list of all project files, and the full code of some files that I want to know the most relevant things done in the file.  I want you to give a one sentence describe of what " + newLine + 
             "each of the full code files do.  Your response should have the filename surrounded in pipes (e.g. |file1.js|) on a new line, and below that, the one sentence description. If the description lists how that file interacts with other files even better." + newLine + 
             "Do that for each file where the entire file is listed." + newLine + 
             "So example output could be: " + newLine +
             "|js/ui_events.js|" + newLine +
             "Contains handlers for buttons and forms from the main index.html file" + newLine +
             "|js/modals.js|" + newLine +
             "Displays and handles dialogs for settings and registration" + newLine +
             "|js/api.js|" + newLine +
             "Communicates with backend server for database inserts" + newLine + newLine;
             
     fullPrompt += "PROJECT: " + project + newLine + newLine;        
     fullPrompt += "FILES: " + newLine + filesListText;
     
     let messages = new Array();
     if(currTab['llm_provider'] == 'openai'){
         messages.push({
             role: "system",
             content: "You are an AI assistant that summarizes the purpose of code files."
         });       
     }
     messages.push({
         role: "user",
         content: fullPrompt
     });
     
     let normalResponse = await CHATGPT.providerChatCompletion(messages, 'get_terse_file_descriptions', currTab, false);
     console.log("chatgpt normal response is");
     console.log(normalResponse);
     return normalResponse;
 };       
        
CHATGPT.requestRelevantStubLines = async function(theTask, filesListText, stubFilesText, tabUid, isQuestion, referenceProjectData){
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    let multipleProjects = false;
    if(currTab['checked_projects'].length > 1){
        multipleProjects = true;
    }
    await CHATGPT.initProvider(tabIndex, 'request_relevant_stub_lines');
    
    console.log("DOING " + currTab['llm_provider'] + " requestRelevantStubLines");
    let newLine = "\n";
    let newLineSection = newLine + "=====================" + newLine;    
    
    let taskWord = "task";
    let taskPhrase = "complete the task";
    if(isQuestion){
        taskWord = "question";
        taskPhrase = "answer the question";
    }    
    let referenceText = "";

    let fullPrompt = "";
    let messages = new Array();
    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){  
        messages.push({
            role: "system",
            content: "You help identify relevant lines in project files related to a " + taskWord + ".  YOU DO NOT GIVE LINE NUMBERS, you give the FULL TEXT of each code line that is relevant to a task.  Do not write new code about the coding task, ONLY WRITE LINES THAT EXIST IN THE FILE."
        });       
    }    
    
    //SHOULD ALLOW NON-STUB FILES?  ADD TEXT BELOW
    //You can also add files in the files list that don't have stub files surrounded by pipes (e.g. |fileonlist.js|) if you believe they will be required for the task. 
    
    fullPrompt += "INSTRUCTIONS: " + newLineSection + "I am going to give you a list of all project files, and some abridged code files (stubs) that I believe are related to the " + taskWord + ". These stubs were generated using the Treesitter AST parser, which breaks up the code into portions based on the top-level control structures and functions. " +
        "I want you to pick lines from the stub files you want to be revealed so you can " + taskPhrase + ". The stubs may not be valid code, but that doesn't matter. Just try to guess the Stub lines around sections that might be relevant to the " + taskWord + ". " +
        "Your response should have the filename surrounded in pipes (e.g. |file1.js|) on a new line, and below that, list all the stub lines (the full text of a line) that are relevant, one per line. All files that have relevant lines should be listed. " +
        "For very long stub lines over 50 characters, just print the first 50 non-whitespace characters. You can have up to 15 relevant lines from each file. Remember, this is to expand sections, so if multiple sections of the stub file are related to the " + taskWord + ", please distribute the selection of lines you pick.  Only return filenames and LINES FROM STUB FILES. No other text should be added." + newLine;

    if(referenceProjectData){
        //let originaProjectName = "";
        fullPrompt += "INSTRUCTIONS: " + newLineSection + "I have a reference project " + referenceProjectData['project_name'] + " that I am using to help with a coding task in another project.  I am going to give you all files in that project, and some abridged code files (stubs) that I believe are related to the " + taskWord + ". These stubs were generated using the Treesitter AST parser, which breaks up the code into portions based on the top-level control structures and functions. " +
        "I want you to pick lines from the stub files you want to be revealed so that may be helpful to " + taskPhrase + ". The stubs may not be valid code, but that doesn't matter. Just try to guess the Stub lines around sections that might be relevant to the " + taskWord + ". " +
        "Your response should have the filename surrounded in pipes (e.g. |file1.js|) on a new line, and below that, list all the stub lines (the full text of a line) that are relevant, one per line. All files that have relevant lines should be listed. " +
        "For very long stub lines over 50 characters, just print the first 50 non-whitespace characters. You can have up to 15 relevant lines from each file. Remember, this is to expand sections, so if multiple sections of the stub file are related to the " + taskWord + ", please distribute the selection of lines you pick.  Only return filenames and LINES FROM STUB FILES. No other text should be added." + newLine;
        referenceText = "Reference ";
    }    
   
    if(isQuestion){
        fullPrompt += "QUESTION: " + newLineSection + theTask + newLine + newLine;
    }else{
        fullPrompt += "TASK: " + newLineSection + theTask + newLine + newLine;
    }
    
    
    
    fullPrompt += referenceText + "Project Files: " + newLine + filesListText + newLine;
    fullPrompt += "Stub Files: " + newLine + stubFilesText + newLine + ".";  

    messages.push({
        role: "user",
        content: fullPrompt
    });
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'request_relevant_stub_lines', currTab, false);
    //console.log("chatgpt normal response is");
    //console.log(normalResponse);
    return normalResponse;
};

CHATGPT.requestFixStubLines = async function(stubMessages, tabUid, doStream, newCodeContext){
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "do_additional_programming_task");

    console.log("DOING " + currTab['llm_provider'] + " doAdditionalProgrammingTask with tab UID: " + tabUid);
    
    let newPrompt = "You did not follow instructions.  Many lines you listed were not found in the Stub files.  This task is supposed to be for identifying " + 
            "markers inside the Stub files that are related to the task.  It is not for writing code, or solving the task.  Now give a list of files and their stub lines again, " +
            "but this time only include lines from the Stub Files.";  


    stubMessages.push({
        role: "user",
        content: newPrompt
    });
    console.log("new messages with request fix:");
    console.log(stubMessages);


    console.log("normal request_fix_stub_lines");
    let normalResponse = await CHATGPT.providerChatCompletion(stubMessages, 'request_fix_stub_lines', currTab, false);
    let fullText = normalResponse.response;
    let finalMessage = {"role": "assistant", "content": fullText};          
    stubMessages.push(finalMessage);              
    return normalResponse;                
};

CHATGPT.requestRelevantLineNumbers = async function(theTask, filesListText, stubFilesText, tabUid){
        
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "request_relevant_line_numbers");
    
    console.log("DOING " + currTab['llm_provider'] + " requestRelevantLineNumbers");
    
    //console.log("stubDataArr");
    //console.log(stubFilesText);
    let newLine = "\n";
    let newLineSection = newLine + "=====================" + newLine;  
    let fullPrompt = "";
    /*
    if(!projectDesc){
        projectDesc = "PROJECT NAME: XXXX PROJECT DESC: I am building an Electron application in javascript.";
        fullPrompt = projectDesc + "\n\n";
    }
    */
    let messages = new Array();
    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){  
        messages.push({
            role: "system",
            content: "You help identify relevant line numbers in project files.  You should only return JSON, DO NOT add any other text."
        });       
    }



    fullPrompt += "INSTRUCTIONS: " + newLineSection + "I am going to give you a list of all project files, and some abridged code files (stubs) I believe are related to the task.  I want you to pick lines from the stub files you want to be revealed " + 
            "so you can complete the task. The stubs may not be valid code, it doesn't matter, just try to guess the lines around sections that might be relevant to the task.  " + 
            "Pick all the areas that you think would be helpful to complete the task, up to 50 different lines.  If you think you should add something at the start or end of file for task, you can add the first or last line." + 
            "The JSON you return should be a list where each item has a 'file_path' which is path of the stub file, a 'line' field that is the text of the line, a 'line_number' which is the estimated line number.  If you think the task will need something new added to the file not related to a section, " +
            "add a 'line_number' 9999 and leave the line text blank.  Every entry must have the " + 
            "file path, line and line number.  Remember to only return JSON.  You are only to mark lines of existing code, do not do the task, I just want relevant lines marked." + newLine;
    fullPrompt += "TASK: " + theTask + newLine + newLine;
    fullPrompt += "Project Files: " + newLine + filesListText + newLine;
    fullPrompt += "Stub Files: " + newLine + stubFilesText + newLine + ".";  
    
    //Here is an example of JSON you might return: " + newLine;
    /*
    let someExample = [
        {"file_path": "math_operations.py", "line": "def add(a, b):"},
        {"file_path": "utils.py", "line": "def print_result(result):"},
        {"file_path": "utils.py", "line": "def get_input():"},
        {"file_path": "main.py", "line": "def main():"}
      ];
    */
   
    //put this in array for later
    CHATGPT.originalFileListArr = filesListText.split(/\r?\n/);
    //console.log("arr version");
    //console.log(CHATGPT.originalFileListArr);

    messages.push({
        role: "user",
        content: fullPrompt
    });

    //console.log("sending");
    //console.log(messages);
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'request_relevant_line_numbers', currTab, true);
    //console.log("chatgpt normal response is");
    //console.log(normalResponse);
    return normalResponse;
      //readStreamAndUpdateTextArea(chatStream);
    
}; 

CHATGPT.formalizeCodeChanges = async function(singleFile, theTask, filesListText, assistantResponse, tabUid, moreTasks, moreResponses, moreUpdated){

    //filesListText is the 
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "formalize_code_changes");
    
    console.log("DOING " + currTab['llm_provider'] + " formalizeCodeChanges");
    //console.log("stubDataArr");
    //console.log(stubFilesText);
    let newLine = "\n";

    let fullPrompt = "";
    let messages = new Array();

    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){  
        messages.push({
            role: "system",
            content: "You help formalize code changes from instructions taking into account the current code files.  You should only return the changes in the specified format, DO NOT add any other text.  You can create multiple changes for the same file when the changes are in different places of the file.  Think about what the change operation will do to the existing code, and if that code is correct after the change.  Doing the change should produce valid working code, so it's important to break changes up into logical parts."
        });       
    }
    
    // Check if there's only one file in filesListText
    let templateFile = 'formalize_code_changes.txt';
    if(singleFile){
        templateFile = 'formalize_code_changes_single.txt';
    }
    let files = filesListText.split(/\r?\n/).filter(line => line.trim() !== "");
    const fileCount = files.length;
    const firstFileName = files.length > 0 ? files[0].trim() : "";

    console.log("First filename " + firstFileName);

    
    console.log("FORMALIZE CHANGES USING TEMPLATE: " + templateFile);
    
    let thePrompt = fs.readFileSync(path.join(STUBS.appPath, 'prompts', templateFile), 'utf-8');

    let isOriginalText = "and the original code files";
    let isOriginalLabel = "ORIGINAL";
    let additionalText = "";

    if (moreTasks && moreTasks.length > 0) {
        let upText = " current";
        isOriginalLabel = "CURRENT";
        if (moreUpdated) {
            upText = " updated ";
            isOriginalLabel = "UPDATED";
        }
        isOriginalText = "any additional change requests and responses, and the " + upText + " code files";
        additionalText = "\n";
        for (let m = 0; m < moreTasks.length; m++) {
            let mTask = moreTasks[m];
            let mResp = moreResponses[m];
            additionalText += "UPDATED TASK REQUEST: \n" +
                mTask + "\n\nUPDATED DIRECTIONS: \n" + mResp + "\n\n";
        }
    }

    fullPrompt = CHATGPT.variableRender(thePrompt, {
        isOriginalText: isOriginalText,
        isOriginalLabel: isOriginalLabel,
        filesListText: filesListText,
        filename: firstFileName,
        theTask: theTask,
        assistantResponse: assistantResponse,
        additionalText: additionalText
    });
    messages.push({
        role: "user",
        content: fullPrompt
    });

    //console.log("sending");
    //console.log(messages);
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'formalize_code_changes', currTab, false);
    console.log("formalize response is");
    console.log(normalResponse);
    return normalResponse;            
            
};

CHATGPT.singleFilePartialChangeNoMessage = async function(filePath, originalSection, partialCode, tabUid){
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "single_file_partial_change");
    
    console.log("DOING " + currTab['llm_provider'] + " single_file_partial_change for file " + filePath);    
    
    let messages = new Array();
    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){  
        messages.push({
            role: "system",
            content: "You are a skilled software developer that gives the updated code only for the section requested. Only provide code, whatever you send will be copy pasted so do not provide anything but code."
        });       
    }
    
    if(originalSection == ""){
        alert("Request for original section empty returning assistant response verbatim");
        return partialCode;
    }    
    
    let newLine = "\n";
    let fullPrompt = "DIRECTIONS: I have been told to modify a section of code in a full source file, and have been given abridged updates or psuedocode that details the changes.  I want you to apply those updates the best you can keeping the rest of the original code intact" + newLine + newLine + 
    "ORIGINAL CODE SECTION:"  + newLine + 
    originalSection  + newLine + newLine + 
    "ABRIDGED UPDATES THAT DETAIL WHAT NEEDS CHANGING:" + newLine + 
    partialCode + newLine + newLine + 
    "NOW GIVE ME THE UPDATED ORIGINAL CODE SECTION. DO NOT INCLUDE ANY PLACEHOLDERS. INCLUDE ALL CODE FROM THE ORIGINAL SECTION INCLUDING COMMENTS UNLESS THE CHANGE IMPLIES REMOVAL/REPLACEMENT OF CODE. BEFORE THIS SECTION WAS WORKING AND I WANT IT TO CONTINUE WORKING WITH THE NEW CHANGES, I SHOULD BE ABLE TO COPY AND PASTE OVER THE SECTION OF CODE WITH YOUR RESPONSE TO GET THE CORRECT CODE." + newLine;

    messages.push({
        role: "user",
        content: fullPrompt
    });
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'single_file_partial_change', currTab, false);
    console.log("single file partial change response is");
    console.log(normalResponse);
    return normalResponse;       

};

CHATGPT.singleFilePartialChange = async function(filePath, partialOriginal, assistantMessage, tabUid){
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "single_file_partial_change");
    
    console.log("DOING " + currTab['llm_provider'] + " single_file_partial_change for file " + filePath);    
    
    let messages = new Array();
    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){  
        messages.push({
            role: "system",
            content: "You are a skilled software developer that gives the updated code only for the section requested. Only provide code, whatever you send will be copy pasted so do not provide anything but code."
        });       
    }
    
    if(partialOriginal == ""){
        alert("Warning - request for code changes empty, unexpected results likely");
    }
    
    let newLine = "\n";
    let fullPrompt = "DIRECTIONS: I have been given instructions on how to implement code changes to achieve my task. I am going to give you the instructions for the changes, and ONE PART OF THE FILE that I need updated. Only provide the new code for THAT PART REQUESTED. Do not worry about the other sections, they have already been updated." + newLine + newLine + 
    "FILE THAT STILL NEEDS UPDATING:"  + newLine + 
    filePath + newLine + newLine + 
    "PARTIAL CODE THAT STILL NEEDS UPDATING:"  + newLine + 
    partialOriginal  + newLine + newLine + 
    "INSTRUCTIONS FOR CODE CHANGES (ONLY FOCUS ON UPDATING THE CODE BLOCK GIVEN):" + newLine + 
    assistantMessage + newLine + newLine + 
    "NOW GIVE ME THE UPDATED PARTIAL CODE IN " + filePath + ", DO NOT INCLUDE ANY PLACEHOLDERS. INCLUDE ALL CODE FROM THE ORIGINAL PARTIAL CODE SNIPPET INCLUDING COMMENTS UNLESS THE CHANGE INVOLVED REMOVING CODE. BEFORE THIS SECTION WAS WORKING AND I WANT IT TO CONTINUE WORKING WITH THE NEW CHANGES, I SHOULD BE ABLE TO COPY AND PASTE OVER THE OLD PARTIAL CODE TO GET THE CORRECT CODE." + newLine;

    messages.push({
        role: "user",
        content: fullPrompt
    });
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'single_file_partial_change', currTab, false);
    console.log("single file partial change response is");
    console.log(normalResponse);
    return normalResponse;          
};


CHATGPT.codeUpdateType = async function(filePath, newCode, stubFile, assistantMessage, tabUid){
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "new_code_placement");
    
    console.log("DOING " + currTab['llm_provider'] + " code_update_type for file " + filePath);    
    
    let messages = new Array();
    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){  
        messages.push({
            role: "system",
            content: "You are a skilled software developer. Only provide the requested information."
        });       
    }
    
    let thePrompt = fs.readFileSync(path.join(STUBS.appPath, 'prompts', 'new_code_placement.txt'), 'utf-8');
    
    let fullPrompt = CHATGPT.variableRender(thePrompt, {
        stubFile: stubFile,
        newCode: newCode,
        assistantMessage: assistantMessage
    });

    messages.push({
        role: "user",
        content: fullPrompt
    });
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'new_code_placement', currTab, false);
    console.log("codeUpdateType response is");
    console.log(normalResponse);
    return normalResponse;      
};


CHATGPT.newCodePlacement = async function(filePath, newCode, stubFile, assistantMessage, tabUid){
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "new_code_placement");
    
    console.log("DOING " + currTab['llm_provider'] + " new_code_placement for file " + filePath);    
    
    let messages = new Array();
    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){  
        messages.push({
            role: "system",
            content: "You are a skilled software developer. Only provide the requested information."
        });       
    }
    
    let thePrompt = fs.readFileSync(path.join(STUBS.appPath, 'prompts', 'new_code_placement.txt'), 'utf-8');
    
    let fullPrompt = CHATGPT.variableRender(thePrompt, {
        stubFile: stubFile,
        newCode: newCode,
        assistantMessage: assistantMessage
    });

    messages.push({
        role: "user",
        content: fullPrompt
    });
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'new_code_placement', currTab, false);
    console.log("CHATGPT.newCodePlacement response is");
    console.log(normalResponse);
    return normalResponse;          
};


CHATGPT.singleFileFullChange = async function(filePath, originalCode, assistantMessage, tabUid){
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "single_file_full_change");
    
    console.log("DOING " + currTab['llm_provider'] + " single_file_full_change for file " + filePath);    
    
    let newLine = "\n";

    let fullPrompt = "";
    /*
    if(!projectDesc){
        projectDesc = "PROJECT NAME: XXXX PROJECT DESC: I am building an Electron application in javascript.";
        fullPrompt = projectDesc + "\n\n";
    }
    */
    let messages = new Array();

    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){  
        messages.push({
            role: "system",
            content: "You are a skilled software developer that gives the updated code for ONE FILE when requested.  Only provide code, whatever you send will be copy pasted so do not provide anything but code."
        });       
    }
    
    /*
     * 
     * 
     * Want to change this  
     *   add-below-line
     *   add-below-block
     */

    fullPrompt += "DIRECTIONS: I have been given instructions on how to implement code changes to achieve my task.  I am going to give you the " + 
            "instructions for the changes, and ONE FILE that I need updated. Only provide the new code for the file requested.  Do not worry about the other files, they have already been updated. " + newLine + 
            "FILE THAT STILL NEEDS UPDATING: " + newLine + 
            filePath + newLine +
            "INSTRUCTIONS FOR CODE CHANGES (ONLY FOCUS ON " + filePath + " CHANGES): " + newLine + 
            assistantMessage + newLine + newLine + 
            "ORIGINAL CODE FOR " + filePath + ":" + newLine + 
            originalCode + 
            "NOW GIVE ME COMPLETE UPDATED CODE FOR " + filePath + ", DO NOT INCLUDE ANY PLACEHOLDERS.  INCLUDE ALL CODE AND COMMENTS FROM THE ORIGINAL FILE UNLESS THE CHANGE INVOLVED REMOVING CODE.  I WANT A FULL WORKING FILE, THAT I CAN COPY AND PASTE OVER THE OLD FILE.";
    messages.push({
        role: "user",
        content: fullPrompt
    });

    //console.log("sending");
    //console.log(messages);
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'single_file_full_change', currTab, false);
    console.log("single file full change response is");
    console.log(normalResponse);
    return normalResponse;          
};


CHATGPT.singleFunctionFullChange = async function(filePath, functionCode, functionName, assistantMessage, tabUid){
    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "single_function_full_change");
    
    console.log("DOING " + currTab['llm_provider'] + " single_function_full_change for file " + filePath);    
    
    let newLine = "\n";

    let fullPrompt = "";
    /*
    if(!projectDesc){
        projectDesc = "PROJECT NAME: XXXX PROJECT DESC: I am building an Electron application in javascript.";
        fullPrompt = projectDesc + "\n\n";
    }
    */
    let messages = new Array();

    if(CHATGPT.LLM_PROVIDER == 'openai' && CHATGPT.LLM_MODEL.indexOf('o1') < 0){  
        messages.push({
            role: "system",
            content: "You are a skilled software developer that gives the updated code for A FUNCTION when requested.  Only provide code for the function, whatever you send will be copy pasted so provide complete code and nothing else."
        });       
    }
    
    /*
     * 
     * 
     * Want to change this  
     *   add-below-line
     *   add-below-block
     */

    fullPrompt += "DIRECTIONS: I have been given instructions on how to implement code changes to achieve my task.  I am going to give you the " + 
            "instructions for the changes, and ONE FUNCTION that I need updated. Only provide the new code for the function requested.  Do not worry about the other parts of the code, they have already been updated. " + newLine + 
            "FUNCTION NAME: " + functionName + newLine + 
            "IN FILE: " + filePath + newLine + 
            "INSTRUCTIONS FOR CODE CHANGES (ONLY GIVE ME THE COMPLETE CODE FOR " + functionName + "): " + newLine + 
            assistantMessage + newLine + newLine + 
            "ORIGINAL CODE FOR " + functionName + ":" + newLine + 
            functionCode + 
            "NOW GIVE ME COMPLETE UPDATED CODE FOR THE FUNCTION " + filePath + ", DO NOT INCLUDE ANY PLACEHOLDERS.  INCLUDE ALL CODE FROM THE ORIGINAL FUNCTION UNLESS THE CHANGE INVOLVED REMOVING CODE.  I WANT A FULL WORKING FUNCTION, THAT I CAN COPY AND PASTE IN THE OLD FILE.";
    messages.push({
        role: "user",
        content: fullPrompt
    });

    //console.log("sending");
    //console.log(messages);
    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'single__full_change', currTab, false);
    console.log("single file full change response is");
    console.log(normalResponse);
    return normalResponse;          
};
                                           
CHATGPT.pruneUnlikelyProjectFiles = async function(ignoreFile, filesListText, totalAmount, targetAmount, excludeAmount, pruneInfo, tabUid){

    let currTab = TABS.getTabByUid(tabUid);    
    let tabIndex = TABS.getIndexByUid(tabUid);
    await CHATGPT.initProvider(tabIndex, "prune_unlikely_project_files");
    
    console.log("DOING " + currTab['llm_provider'] + " prune_unlikely_project_files");
    //console.log("stubDataArr");
    //console.log(stubFilesText);
    let newLine = "\n";

    let fullPrompt = "";
    let messages = new Array();
    

    fullPrompt += "INSTRUCTIONS: I want you to help me exclude " + excludeAmount + " more project files based on how unlikely our programmers will edit those files while doing tasks. We already have an ignore file that excludes unlikely files " + 
            "and I want you to add to that file.  The amount of files after applying the current .gitignore are " + totalAmount + " and I want to reduce that number to " + targetAmount + " by giving me new rules.  Files like library files, minified files, images and other unnecessary files do not need to be edited by a programmer to do tasks.  They should be the ones you try to exclude first with your rules. " + 
            "After that exclude others not seeming to be in active development until you reduce the files that would be included to the target amount." + newLine +             
            "If let's say 5 of the files needed to be excluded are in the same folder, the git rule " + newLine + 
            "could just exclude the whole folder.  Or you could write rule with wildcard that excludes all of a type of file.  But be careful, do not write rules that would exclude ANY files that need to be kept. You are only to exclude the " + excludeAmount + " files that are least likely to need to be edited for tasks.  I am going to give you the current included files list, the existing .gitignore, the target amount of files that should be left after the rules. " + 
            "You are to only give me NEW RULES that will exclude that amount of files." + newLine +              
            "HERE ARE THE PROJECT FILE PATHS: " + newLine + 
            filesListText + newLine + newLine +
            "THE CURRENT IGNORE RULES: " + newLine + 
            ignoreFile + newLine + newLine +
            "TARGET AMOUNT DESIRED: " + targetAmount + newLine + newLine + 
            "Remember only respond with the NEW RULES (one per line) that can be added to the current .gitignore to get the target amount of files, keeping the most likely ones";
    
    messages.push({
        role: "user",
        content: fullPrompt
    });

    
    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'prune_unlikely_project_files', currTab, false);
    console.log("formalize response is");
    console.log(normalResponse);
    return normalResponse;            
};


CHATGPT.parseFilesList = function(assistantMessage, currTab){
    
    let responseLines = assistantMessage.split(/\r?\n/);
    let origFullPathArr = CHATGPT.originalFileListArr;
    let origShortPathArr = origFullPathArr.map(fullPath => path.basename(fullPath));
    let finalFiles = new Array();
    let imageDescriptionText = "";
    let imageDescriptionStarted = false;
    for(let y = 0; y < responseLines.length; y++){
        let aLine = responseLines[y];
        let fullPathIndex = origFullPathArr.indexOf(aLine);
        let shortPathIndex = origShortPathArr.indexOf(aLine);
        
        if(currTab && currTab['has_image'] && aLine.indexOf('"') >= 0){
            imageDescriptionStarted = true;
            imageDescriptionText += aLine;
        }else if(currTab && currTab['has_image'] && imageDescriptionStarted){
            imageDescriptionText += aLine;
        }

        if(fullPathIndex !== -1){
            if(!finalFiles.includes(origFullPathArr[fullPathIndex])){
                finalFiles.push(origFullPathArr[fullPathIndex]);
            }
        }else if(shortPathIndex !== -1){
            if(!finalFiles.includes(origFullPathArr[shortPathIndex])){
                finalFiles.push(origFullPathArr[shortPathIndex]);
            }            
        }else{
            for(let z = 0; z < origShortPathArr.length; z++){
                if(aLine.indexOf(origShortPathArr[z]) !== -1){
                    if(!finalFiles.includes(origFullPathArr[z])){
                        finalFiles.push(origFullPathArr[z]);
                        break; // exit the loop after finding a match
                    }                         
                }
            }       
        }           
    }
    
    if(currTab && currTab['has_image']){
        let tabIndex = TABS.getIndexByUid(currTab['uid']);
        TABS.TAB_ARR[tabIndex]['image_description'] = imageDescriptionText;
        console.log("IMAGE DESCRIPTION TEXT: " + imageDescriptionText);
    }
    
    console.log('final files are');
    console.log(finalFiles);
    
    return finalFiles;        
};

CHATGPT.parseFileDescriptions = function(assistantMessage, badGroups) {
  const groups = [];
  const lines = assistantMessage.split("\n");
  let currentGroup = null;

  for (let line of lines) {
    let lineTrim = line.trim();
    if ((badGroups && lineTrim.indexOf("|") >= 0) || (lineTrim.startsWith("|") && lineTrim.endsWith("|"))) {
      if (currentGroup) {
        groups.push(currentGroup);
      }
      currentGroup = {
        file_path: lineTrim.slice(1, -1),
        description: ""
      };
    } else if (currentGroup && line) {
      currentGroup.description += line;
    }
    
  }

  if (currentGroup) {
    groups.push(currentGroup);
  }
  
  if(groups.length <= 0){
      return CHATGPT.parseFileDescriptions(assistantMessage, true);
  }
  

  return groups;
};

CHATGPT.parsePruneUnlikelyProjectFiles = function(assistantMessage) {
  const lines = assistantMessage.split("\n");
  // Exclude any lines that have more than 2 spaces or are empty after .trim()
  const prunedLines = lines.filter(line => {
    const trimmedLine = line.trim();
    const spaceCount = (trimmedLine.match(/ /g) || []).length;
    return trimmedLine !== "" && spaceCount <= 2;
  });
  
  return prunedLines;
};

CHATGPT.parseNewCodePlacement = function(fullText) {
    console.log("parsing new code placement");
    console.log(fullText);
    
    const lines = fullText.split(/\r?\n/).filter(line => line.trim() !== '');
    const validOperations = ['outside-above', 'outside-below', 'inside'];
    
    let stubLine = '';
    let operation = 'outside-above'; // Default operation
    
    for (let line of lines) {
        line = line.trim();
        
        if (!stubLine && !validOperations.some(op => line.includes(op))) {
            stubLine = line;
        }
        
        for (let op of validOperations) {
            if (line.includes(op)) {
                operation = op;
                break;
            }
        }
        
        if (stubLine && operation !== 'outside-above') {
            break;
        }
    }
    
    if (!stubLine) {
        return null;
    }
    
    return {
        stub_line: stubLine,
        operation: operation
    };
};


CHATGPT.getJustCode = function(fullText, filepathArr = []) {
    const lines = fullText.split(/\r?\n/);
    const hasCodeBlock = lines.some(line => line.indexOf('```') >= 0);
    
    let newText = "";
    let inCodeBlock = false;
    let skipFirstLine = false;

    // Check if the first line contains a filepath or starts with "Here's the" or "Here is the" and contains "code"
    if (lines.length > 0) {
        const firstLine = lines[0].trim().toLowerCase();
        skipFirstLine = filepathArr.some(path => firstLine.includes(path.toLowerCase())) ||
                        ((firstLine.startsWith("here's the") || firstLine.startsWith("here is the")) && 
                         firstLine.includes("code"));
    }

    for (let i = 0; i < lines.length; i++) {
        let line = lines[i];

        if (i === 0 && skipFirstLine) {
            continue;
        }

        if (hasCodeBlock) {
            if (line.indexOf('```') >= 0) {
                inCodeBlock = !inCodeBlock;
                continue;
            }
            if (inCodeBlock) {
                newText += line + '\n';
            }
        } else {
            // If there's no code block, include all lines except empty ones at the start
            if (newText.trim() !== "" || line.trim() !== "") {
                newText += line + '\n';
            }
        }
    }

    return newText.trim();
};

CHATGPT.parseFormalizedChanges = function(fullText){
  const changes = [];
  fullText = fullText.replace(/\r\n/g, '\n');
  const lines = fullText.split(/\r?\n/);
  let newLine = "\n";
  let currentChange = null;

  for (let line of lines) {
      
    if(line.indexOf('```') >= 0){
        //for some reason the llm is saying 
        //```javascript
        console.log("Skipping the ``` line");
        continue;
    }  
    let trimLine = line.trim();

    if (trimLine.startsWith('|') && trimLine.endsWith('|')) {
      if (currentChange) {
        currentChange['contiguous_lines'] = currentChange['contiguous_lines'].join(newLine);
        currentChange['replace_lines'] = currentChange['replace_lines'].join(newLine);
        
        changes.push(currentChange);
      }
      currentChange = {
        file_path: trimLine.slice(1, -1),
        contiguous_lines: [],
        insertion_type: '',
        replace_lines: [],
      };
    } else if (trimLine.startsWith(':') && trimLine.endsWith(':')) {
         currentChange.insertion_type = line.slice(1, -1);
    } else if (currentChange.insertion_type != "") {
         currentChange.replace_lines.push(line);
    } else {
         currentChange.contiguous_lines.push(line);
    }
  }

  if (currentChange) {
    currentChange['contiguous_lines'] = currentChange['contiguous_lines'].join(newLine);
    currentChange['replace_lines'] = currentChange['replace_lines'].join(newLine);      
    changes.push(currentChange);
  }

  return changes;    
};


CHATGPT.parseStubFileLines = function(response) {
  console.log("parsing stub files");
  console.log(response);
  const lines = response.split("\n");
  const result = [];
  let currentFile = null;
  let currentFileExt = null;

  for (let line of lines) {
    line = line.trim();
    //console.log("line");
    //console.log(line);
    //add else if with one word and period, known extension, as fallback?
    if (line.startsWith("|") && line.endsWith("|")) {
      currentFile = line.slice(1, -1);
      let fileExtension = path.extname(currentFile);
      currentFileExt = fileExtension.slice(1);  
    }     
    else if (currentFile && line) {
      let functionName = extractFunctionFromLine(line, currentFileExt);
      let lineItem = {
        file_path: currentFile,
        line: line,
        function_name: functionName
      };
      //console.log("adding item: ", lineItem);
      result.push(lineItem);
    }
  }

  return result;
};


CHATGPT.timeout = function(ms, promise) {
    return new Promise((resolve, reject) => {
        setTimeout(() => {
            reject(new Error("Timeout after " + ms + " ms"));
        }, ms);
        promise.then(resolve).catch(reject);
    });
};

CHATGPT.parseLineNumbers = async function (relevantLines, projectName, stubsData, modifiedContext) {
    const CONTEXT_LINES = 100;
    if(!stubsData){
          stubsData = STUBS.stubsData;
    }
    const projectStubData = stubsData[projectName];
    const output = [];
    const fileLineNumbers = {};    
    CHATGPT.relevantContextJson = new Array();
    let stub = null;
    let removeFiles = new Array();
    let badCount = new Array();
    
    // Process modified context
    if (modifiedContext) {
        for (const modifiedFile of modifiedContext) {
            const { filename, the_state } = modifiedFile;
            if (the_state === 'fullfile') {
                const stub = projectStubData.find(s => s.filename === filename);
                if (stub) {
                    let fileContent = await fs.readFile(stub.full_path, 'utf-8');
                    fileContent = fileContent.replace(/\r\n/g, '\n');
                    fileLineNumbers[stub.full_path] = {
                        shortPath: stub.file_path,
                        fullPath: stub.full_path,
                        lineNumbers: [],
                        fullSource: fileContent,
                        total_line_count: stub.line_count
                    };
                    // Add all line numbers to get the full file
                    for (let i = 0; i < stub.line_count; i++) {
                        fileLineNumbers[stub.full_path].lineNumbers.push(i);
                    }
                }
            } 
            else if (the_state === 'remove') {
                removeFiles.push(stub.full_path);
                // Remove the file from fileLineNumbers if it exists
                //const stub = projectStubData.find(s => s.filename === filename);
                //if (stub && fileLineNumbers[stub.full_path]) {
                //    delete fileLineNumbers[stub.full_path];
                //}
            }
        }
    }    
    
    let parsedTrees = {};
    // Collect all the relevant line numbers for each file
    for (const relevantLine of relevantLines) {
        let { file_path, line } = relevantLine;
        console.log("processing filepath: " + file_path);
        console.log("processing line: " + line);
        try{
            // Replace newlines with slashes in file_path
            file_path = file_path.replace(/\n/g, '/');        

            let stub = projectStubData.find(s => path.normalize(s.file_path) === path.normalize(file_path) || s.filename === path.basename(file_path));

            if (!stub) {
                console.log('No matching stub found for', file_path);
                console.log('!!!!!!!!!!!!!!!!!!!!!!!!!!!', file_path);
                console.log('!!!!!!!!!!!!!!!!!!!!!!!!!!!', file_path);
                console.log('!!!!!!!!!!!!!!!!!!!!!!!!!!!', file_path);

                continue;
            }
            if(removeFiles.includes(file_path)){
                continue;
            }

            const fileContent = readFileWithCache(stub.full_path);
            const fileLines = fileContent.split(/\r?\n/);
            
            const exactLineNumber = fileLines.findIndex(l => l.trim() === line.trim());
            let lineNumber = -1;
            let isExact = true;
            if(exactLineNumber >= 0){
                lineNumber = exactLineNumber;
                isExact = true;
            }else{
                lineNumber = fileLines.findIndex(l => l.includes(line.trim()));
            }
            if (lineNumber === -1) {
                let fuzzyMatchItem = findFuzzyMatchIndex(fileContent, line.trim());
                console.log('Line not found in file', line, stub.full_path);
                console.log("Fuzzy Wuzzy", fuzzyMatchItem);
                console.log('!!!!!!!!!!!!!!!!!!!!!!!!!!!', file_path);
                
                let alreadyBad = false;
                if(badCount.length > 0){
                    for(let b = 0; b < badCount.length; b++){
                        if(badCount[b]['path'] == stub.full_path){
                            alreadyBad = true;
                        }
                    }
                }
                
                
                if (!fileLineNumbers[stub.full_path] || alreadyBad) {
                    console.log('DOING FALLBACK TO SEND WHOLE/MOST FILE!!!!!!!!!!', file_path);
                    //pushing all lines up to 1000
                    //workaround if the stupid LLM wants to WRITE NEW CODE instead of give me the stub lines

                    console.log("ONLY DO THIS IF NO OTHER LINES FOUND SINCE THE LLMS ARE MAKING UP CODE TO ADD BASED ON TASK");                    
                    fileLineNumbers[stub.full_path] = {
                        shortPath: stub.file_path,
                        fullPath: stub.full_path,
                        lineNumbers: [],
                        fullSource: fileContent,
                        total_line_count: stub.line_count
                    };
                    let limitCount = 50;
                    /*
                    if(TABS.TAB_ARR[TABS.CURRENT_INDEX]['llm_provider'] == 'anthropic'){
                        limitCount = 50;
                    }
                    */
                   
                    let theLineCount = fileLineNumbers[stub.full_path].total_line_count;
                    if(theLineCount <= 0){
                        let stubFullFile = readFileWithCache(stub.full_path);
                        let lineCount = stubFullFile.toString().split(/\r?\n/).length;
                        if(lineCount > 0){
                            theLineCount = lineCount;
                        }else{
                            theLineCount = 5000;
                        }
                    }
                   
                    for(let m = 0; m < limitCount; m++){
                        let aLineNum = 95 * m;
                        //console.log("total line count: " + theLineCount + " VS " + aLineNum);
                        
                        if (aLineNum < theLineCount && !fileLineNumbers[stub.full_path].lineNumbers.includes(aLineNum)) {
                            fileLineNumbers[stub.full_path].lineNumbers.push(aLineNum);
                        }                       
                    }

                    console.log("after workaround for file: " + stub.full_path);
                    console.log(fileLineNumbers);                    
                }                        
                
                badCount.push({
                    path: stub.full_path,
                    line: line.trim()
                });
                
                continue;
            }

            if (!fileLineNumbers[stub.full_path]) {
                fileLineNumbers[stub.full_path] = {
                    shortPath: stub.file_path,
                    fullPath: stub.full_path,
                    lineNumbers: [],
                    fullSource: fileContent,
                    total_line_count: stub.line_count
                };
            }

            //fileLineNumbers[stub.full_path].lineNumbers.push(lineNumber);
            if (!fileLineNumbers[stub.full_path].lineNumbers.includes(lineNumber)) {
                fileLineNumbers[stub.full_path].lineNumbers.push(lineNumber);
                
                if(isExact){
                    console.log("EXACT MATCH SO GETTING CONTROL STRUCTURE");
                    let preparsedTree = null;
                    if(!(stub.full_path in parsedTrees)){
                        preparsedTree = await DGRAPH.parseAnyFile(stub.full_path, fileContent);
                        parsedTrees[stub.full_path] = preparsedTree;
                    }else{
                        preparsedTree = parsedTrees[stub.full_path];
                    }
                    let exactLineText = line.trim();
                    console.log("exact line text: " + exactLineText);
                    let controlStructure = await DGRAPH.getControlStructureForLine(stub.full_path, exactLineText, fileContent, null, preparsedTree);
                    if(controlStructure != null){
                        let theLineCount = fileLineNumbers[stub.full_path].total_line_count;
                        if(theLineCount <= 0){
                            let stubFullFile = readFileWithCache(stub.full_path);
                            let lineCount = stubFullFile.toString().split(/\r?\n/).length;
                            if(lineCount > 0){
                                theLineCount = lineCount;
                            }else{
                                theLineCount = 100000;
                            }
                        }                        
                        let theDiff = controlStructure.endLine - controlStructure.startLine;
                        let diffPoints = Math.ceil(theDiff / 100);
                        for(let m = 1; m <= diffPoints; m++){
                            let extraLineNum = (100 * m) + lineNumber;
                            if (extraLineNum < theLineCount && !fileLineNumbers[stub.full_path].lineNumbers.includes(extraLineNum)) {
                                fileLineNumbers[stub.full_path].lineNumbers.push(extraLineNum);
                            }                       
                        }                        
                    }
                }
                
                
            }        
        }catch(eee){
            console.log("REALLY BAD LINE --- START ERROR");
            console.log(eee);
            console.log(file_path);
            console.log(line);
            console.log("REALLY BAD LINE --- END ERROR");
        }
    }
    if(CHATGPT.searchTermArr && CHATGPT.searchTermArr.length > 0){
    // New loop: Iterate over each stub file and then search for each text
        for (const stub of projectStubData) {
            // Check if the stub file is in the important files list
            if (!CHATGPT.stubFilesArr.some(stubFileName => stub.full_path.includes(stubFileName))) {
                continue; // Skip this file if it's not in the list
            }

            let fileContent = await fs.readFile(stub.full_path, 'utf-8');
            fileContent = fileContent.replace(/\r\n/g, '\n');
            const fileLines = fileContent.split(/\r?\n/);

            for (const searchText of CHATGPT.searchTermArr) {
                fileLines.forEach((line, lineNumber) => {
                    if (line.includes(searchText)) {
                        if (!fileLineNumbers[stub.full_path]) {
                            fileLineNumbers[stub.full_path] = {
                                shortPath: stub.file_path,
                                fullPath: stub.full_path,
                                lineNumbers: [],
                                fullSource: fileContent
                            };
                        }

                        if (!fileLineNumbers[stub.full_path].lineNumbers.includes(lineNumber)) {
                            fileLineNumbers[stub.full_path].lineNumbers.push(lineNumber);
                        }
                    }
                });
            }
        }
    }
    
    let allRelevantText = "";
    // Build the relevant source for each file
    for (const filePath in fileLineNumbers) {
        const { lineNumbers, fullSource, shortPath } = fileLineNumbers[filePath];
        const fileLines = fullSource.split(/\r?\n/);

        // Sort line numbers and remove duplicates
        const uniqueLineNumbers = Array.from(new Set(lineNumbers)).sort((a, b) => a - b);

        let relevantSource = '';
        let hiddenLineCount = 0;
        let visibleLineCount = 0; // Initialize visible line count
        fileLines.forEach((line, index) => {
            let isRelevant = uniqueLineNumbers.some(lineNumber => 
                index >= lineNumber - CONTEXT_LINES && index <= lineNumber + CONTEXT_LINES);

            if (isRelevant) {
                if (hiddenLineCount > 0) {
                    relevantSource += `//${hiddenLineCount} lines hidden\n`;
                    hiddenLineCount = 0;
                }
                relevantSource += line + '\n';
                visibleLineCount++; // Increment visible line count
            } else {
                hiddenLineCount++;
            }
        });

        // Handle the case where the last lines are hidden
        if (hiddenLineCount > 0) {
            relevantSource += `//${hiddenLineCount} lines hidden\n`;
        }
        allRelevantText += '\n';
        allRelevantText += shortPath + '\n';
        allRelevantText += relevantSource;
        allRelevantText += '\n';
        let totalLineCount = fileLines.length;

        output.push({
            file_path: filePath,
            short_path: shortPath,
            full_source: fullSource,
            relevant_source: relevantSource,
            relevant_lines: uniqueLineNumbers,
            total_lines: totalLineCount,
            visible_line_count: visibleLineCount 
        });
        
        CHATGPT.relevantContextJson.push({
            short_path: shortPath, 
            file_content: relevantSource
        });      
    }
      
    if(CHATGPT.FILE_LOGGING){
        const fileName = `finalcontext_${getFormattedDateTime()}.txt`;
        const logFilePath = path.join('.', 'logs', fileName);
        fs.writeFileSync(logFilePath, allRelevantText, 'utf8');    
    }
    
    /*
    let relevantContextJson = JSON.stringify(CHATGPT.relevantContextJson);
    const fileNameJson = `finalcontextjson_${getFormattedDateTime()}.txt`;
    const logFilePath2 = path.join('.', 'logs', fileNameJson);
    fs.writeFileSync(logFilePath2, relevantContextJson, 'utf8');    
    */
    
    let charCount = allRelevantText.length;
    let lineCount = allRelevantText.split(/\r?\n/).length;
       
    return { all_relevant_text: allRelevantText, all_data: output, char_count: charCount, line_count: lineCount, bad_lines: badCount };
};



// Function to format the current date and time for the file name
function getFormattedDateTime() {
  const now = new Date();
  return now.toISOString()
    .replace(/:/g, '-')
    .replace(/\..+/, '')
    .replace('T', '_');
}

//Backup of the 
CHATGPT.parseCodeInstructions = function(instructions) {
    const codeBlocks = instructions.split('`' + '`' + '`');
    let results = [];

    for (let i = 1; i < codeBlocks.length; i += 2) {
        let header = codeBlocks[i - 1].trim().split(/\r?\n/).pop();
        let code = codeBlocks[i].trim();
        let [filename, fileType] = header.split(/\r?\n/).pop().split('.');

        // Inferring the file type from the extension if not explicitly provided
        fileType = fileType || CHATGPT.inferFileType(filename);

        results.push({
            filename: filename,
            code: code,
            type: fileType
        });
    }

    return results;
}

CHATGPT.inferFileType = function(filename) {
    const extension = filename.split('.').pop();
    switch(extension) {
        case 'js': return 'javascript';
        case 'html': return 'html';
        case 'css': return 'css';
        // Add more cases for different file extensions as needed
        default: return 'unknown';
    }
};

CHATGPT.startMilestonePlannerConversation = function(){
    
};

CHATGPT.testStreaming = async function(){
    let currTab = TABS.TAB_ARR[TABS.CURRENT_INDEX];    

    let theTask = "Tell me which colors are in the text in this image named " + randomLetters(5) + ".jpg.  If you don't see any image then tell me.  Otherwise say what's in the image and the text color";  
    TABS.TAB_ARR[TABS.CURRENT_INDEX]['task_text'] = theTask;
    console.log("DOING CHATGPT testStreaming"); 
    await CHATGPT.initProvider(TABS.CURRENT_INDEX, "do_new_programming_task");

    //let codeFiles = await CHATGPT.prepareFileText();
    //console.log("files");
    //console.log(codeFiles);
    let newLine = "\n";

    let fullPrompt = "";

    let messages = new Array();
    fullPrompt += theTask;
    
    if(currTab['llm_provider'] == 'anthropic' && currTab['has_image']){
        console.log("adding image to chat completion");
        let contentArr = [];
        for(let c = 0; c < currTab['context_images'].length; c++){
            let newImage = {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": "image/jpeg",
                        "data": currTab['context_images'][c]
                    }
                };
                contentArr.push(newImage);
        }
        contentArr.push({
            "type": "text", 
            "text": fullPrompt
        });
        messages.push({
            role: "user", 
            content: contentArr
        });   
    }
    else if(currTab['llm_provider'] == 'openai' && currTab['has_image']){
        console.log("adding image to chat completion");
        let contentArr = [];
        for(let c = 0; c < currTab['context_images'].length; c++){
            let dataUrl = "data:image/jpeg;base64," + currTab['context_images'][c];
            let newImage = {
                    "type": "image_url",
                    "image_url": {
                        "url": dataUrl
                    }
                };
            contentArr.push(newImage);
        }
        contentArr.push({
            "type": "text", 
            "text": fullPrompt
        });
        messages.push({
            role: "user", 
            content: contentArr
        });          
        console.log("messages after images added");
        console.log(messages);
    }
    else{
        messages.push({
            role: "user",
            content: fullPrompt
        });        
    }
   
     let chatStream = null;
    if(currTab['llm_provider'] == 'openai'){
        chatStream = await CHATGPT.streamChatCompletion(messages, 'do_new_programming_task', currTab);        
    }else if(currTab['llm_provider'] == 'anthropic'){
        chatStream = await CHATGPT.streamAnthropicCompletion(messages, 'do_new_programming_task', currTab);   
    }
    else if(currTab['llm_provider'] == 'mistral'){
        console.log("really streaming mistral");
        chatStream = await CHATGPT.streamMistralCompletion(messages, 'do_new_programming_task', currTab);   
    }
    else if(currTab['llm_provider'] == 'groq'){
        console.log("really streaming groq");
        chatStream = await CHATGPT.streamGroqCompletion(messages, 'do_new_programming_task', currTab);   
    }    
    console.log("chatstream is");
    console.log(chatStream);
    //readStreamAndUpdateTextArea(chatStream);   
};

CHATGPT.extractFilePath = function(bufferText, tabUid) {
    let tabIndex = TABS.getIndexByUid(tabUid);
    let fileArr = TABS.TAB_ARR[tabIndex]['just_files_linux'];
    
    const extractFilePath = (line) => {      
        const backtickMatch = line.match(/`([^`\s]+?\.\w+)`/g);
        if (backtickMatch) {
            const lastMatch = backtickMatch[backtickMatch.length - 1];
            return lastMatch.slice(1, -1);
        }      
        
        const parts = line.split(/[/\\]/);
        const lastPart = parts[parts.length - 1];
        if (lastPart.includes('.') && lastPart.split('.').pop().match(/\w+/)) {
            let lSplit = line.split(' ');  
            for(let s = 0; s < lSplit.length; s++){
                let aPart = lSplit[s];
                if(aPart.includes('.')){
                    let sanitizedPath = aPart.replace(/[<>:"|?*\x00-\x1F]/g, '');
                    return sanitizedPath.trim();
                }
            }
            
            return line.trim();
        }
        return '';
    };
    
    const matchFileFromList = (line, existingFilenames) => {      
        for (let existingFile of existingFilenames) {
            if (line.includes(existingFile)) {
                return existingFile;  // Return as soon as an exact match is found
            }
        }     
        return '';
    };

    // Split the bufferText into lines
    const lines = bufferText.split(/\r?\n/);
    
    for(let j = lines.length - 1; j >= 0; j--) {
        const result = matchFileFromList(lines[j], fileArr);
        if (result) {
            return result;  // Return the first non-empty result (last valid filepath)
        }    
    }
    
    // Iterate through lines in reverse order
    for (let i = lines.length - 1; i >= 0; i--) {
        const result = extractFilePath(lines[i]);
        if (result) {
            return result;  // Return the first non-empty result (last valid filepath)
        }
    }

    return '';  // Return empty string if no valid filepath found
};

CHATGPT.getLLMProviderForTask = function(taskType) {
    let recommendedArr = CHATGPT.recommended_model_properties[taskType];
    
    if(taskType == 'do_additional_programming_task'){
        recommendedArr = CHATGPT.recommended_model_properties['do_new_programming_task'];
    }      
    
    if(taskType in STUBS.userOverrideLLMs && STUBS.userOverrideLLMs[taskType] !== undefined && STUBS.userOverrideLLMs[taskType] !== ''){          
        recommendedArr.unshift(STUBS.userOverrideLLMs[taskType]);
        console.log("Adding user override: " + STUBS.userOverrideLLMs[taskType]);
    }

    console.log("recommended arr");
    console.log(recommendedArr);
    
    let llmProvider = "";
    for(let r = 0; r < recommendedArr.length; r++){
        let theModel = recommendedArr[r];
        if(llmProvider != ""){
            break;
        }
        for(let c = 0; c < CHATGPT.AVAIL_MODELS.length; c++){
            let cModel = CHATGPT.AVAIL_MODELS[c];
            let cProv = CHATGPT.AVAIL_MODEL_PROVIDER[c];
            if(cModel == theModel){                  
                return cProv;
            }
        }
    }
    return null;
};

CHATGPT.testNormalCompletion = async function(theTask){
    let currTab = TABS.TAB_ARR[TABS.CURRENT_INDEX];
    
    TABS.TAB_ARR[TABS.CURRENT_INDEX]['task_text'] = theTask;
    //;  
    console.log("DOING CHATGPT testNorma");
    await CHATGPT.initProvider(TABS.CURRENT_INDEX, "do_new_programming_task");    

    //let codeFiles = await CHATGPT.prepareFileText();
    //console.log("files");
    //console.log(codeFiles);
    let newLine = "\n";

    let fullPrompt = "";

    let messages = new Array();
    /*
    if(currTab['llm_provider'] == 'openai'){
        messages.push({
            role: "system",
            content: "You are a helpful coding assistant that writes full working code."
        });       
    }     
     */
    fullPrompt += theTask;
    if(currTab['llm_provider'] == 'anthropic' && currTab['has_image']){
        console.log("adding image to chat completion");
        let contentArr = [];
        for(let c = 0; c < currTab['context_images'].length; c++){
            let newImage = {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": "image/jpeg",
                        "data": currTab['context_images'][c]
                    }
                };
                contentArr.push(newImage);
        }
        contentArr.push({
            "type": "text", 
            "text": fullPrompt
        });
        messages.push({
            role: "user", 
            content: contentArr
        });   
    }
    else if(currTab['llm_provider'] == 'openai' && currTab['has_image']){
        console.log("adding image to chat completion");
        let contentArr = [];
        for(let c = 0; c < currTab['context_images'].length; c++){
            let dataUrl = "data:image/jpeg;base64," + currTab['context_images'][c];
            let newImage = {
                    "type": "image_url",
                    "image_url": {
                        "url": dataUrl
                    }
                };
            contentArr.push(newImage);
        }
        contentArr.push({
            "type": "text", 
            "text": fullPrompt
        });
        messages.push({
            role: "user", 
            content: contentArr
        });          
        console.log("messages after images added");
        console.log(messages);
    }
    else{
        messages.push({
            role: "user",
            content: fullPrompt
        });        
    }
    


    let normalResponse = await CHATGPT.providerChatCompletion(messages, 'do_new_programming_task', currTab, false);
    
    let assistantMessage = normalResponse['response'];
    
    console.log("assistantMessage: ");
    console.log(assistantMessage);
    
    CHATGPT.processHtmlText(assistantMessage);    
    
    return normalResponse; 
    //console.log("chatstream is");
    //console.log(chatStream);
    //readStreamAndUpdateTextArea(chatStream);
       
    
};

CHATGPT.variableRender = function(template, variables) {
    let result = template;
    for (let key in variables) {
        if (variables.hasOwnProperty(key)) {
            let searchStr = '{{{' + key + '}}}';
            let replaceStr = String(variables[key]);
            result = result.split(searchStr).join(replaceStr);
        }
    }
    return result;
};

CHATGPT.SWITCH_LLM = function(tabUid) {
    let currTab = TABS.getTabByUid(tabUid);
    let tabIndex = TABS.getIndexByUid(tabUid);
    if(CHATGPT.AVAILABLE_LLMS.length == 1){
        //cant switch
        return false;
    }
  // Find the current index of LLM_PROVIDER in AVAILABLE_LLMS
  let currentIndex = CHATGPT.AVAILABLE_LLMS.indexOf(currTab['llm_provider']);
  if(currentIndex < 0){
      currentIndex = 0;
  }
  // Calculate the next index
  let nextIndex = (currentIndex + 1) % CHATGPT.AVAILABLE_LLMS.length;

  // Switch LLM_PROVIDER to the next one in AVAILABLE_LLMS
  TABS.TAB_ARR[tabIndex]['llm_provider'] = CHATGPT.AVAILABLE_LLMS[nextIndex];
  return true;
};


CHATGPT.testDatabaseLogger = async function(mysqlConfig){
    
    let requestType = "testType";
    let currentTask = "testTask";
    let fullPrompt = "testPrompt";
    let assistantMessage = "testMessage";
    let retData = "testRetData";
    await mysqlLogger.initDatabase(mysqlConfig);    
    //mysqlLogger.logRequest(1, requestType, currentTask, fullPrompt, assistantMessage, retData);
    mysqlLogger.logResponse(1, requestType, TABS.CURRENT_TASK, fullPrompt, assistantMessage, retData, "test");

    
};

CHATGPT.badFormalizeResponse = function(assistantMessage){
    let allLines = assistantMessage.split(/\r?\n/);
    let firstLine = allLines[0].trim();
    if(firstLine === '{' || firstLine == '}'){
        return true;
    }
    if(allLines.length > 1){
        let secondLine = allLines[1].trim();
        if(secondLine === '{' || secondLine == '}'){
            return true;
        }        
    }
    
    return false;
};

CHATGPT.setTabs = function(globalTabObject){
    global.TABS = globalTabObject;
};


if (!process.type) {
    module.exports = CHATGPT;
}