const cssWhat = require('css-what');
var cssjs = require("jotform-css.js");

//uncomment these to use in test_treesit.js
//const fs = require('fs-extra');
//const path = require('path');
//const Parser = require('web-tree-sitter');

// Create a new parser instance
let parser;
var DGRAPH = DGRAPH || {};
DGRAPH.parser = null;
DGRAPH.codeExtensions = ['php', 'js', 'sh', 'c', 'cs', 'cpp', 'elm', 'go', 'java', 'lua', 'py', 'rb', 'rs', 'sv', 'dart', 'json', 'ts', 'jsx', 'tsx', 'kt', 'swift', 'm', 'h', 'groovy', 'r', 'pl', 'f90', 'f95', 'f03', 'f08', 'vb', 'asm', 's', 'bas', 'pas', 'scala', 'clj', 'coffee', 'erl', 'hs', 'ex', 'exs', 'jl', 'nim', 'ml', 'fs', 'fsx', 'ps1', 'bat', 'cmd', 'vbs', 'ahk', 'au3', 'lisp', 'scm', 'rkt', 'sol', 'tf', 'hcl', 'dockerfile', 'blade', 'proto', 'graphql'];
DGRAPH.markupExtensions = ['html', 'tpl', 'md', 'vue', 'xml', 'pug', 'haml', 'jade', 'slim', 'liquid', 'hbs', 'ejs', 'njk', 'tex', 'twig', 'mustache', 'jinja', 'jinja2', 'handlebars', 'erb', 'rhtml', 'swig', 'phtml', 'latte', 'volt', 'mako', 'marko', 'dust', 'nunjucks', 'kit', 'dot', 'eco', 'ect', 'jst'];
DGRAPH.styleExtensions = ['css', 'styl', 'scss', 'sass', 'less', 'sty', 'cls'];
DGRAPH.stubExtensions = ['js', 'jsx', 'php', 'sh', 'c', 'cs', 'cpp', 'cc', 'cxx', 'hpp', 'hxx','css', 'elm', 'html', 'htm', 'tpl', 'java', 'lua', 'md', 'markdown','py', 'rb', 'rs', 'json', 'vue', 'yaml', 'yml', 'dart', 'go', 'ts', 'tsx', 'swift', 'kt'];
DGRAPH.webfileExtensions = ['html', 'tpl', 'css', 'mustache'];
DGRAPH.TREE_CACHE = {};
DGRAPH.PARSE_CALL_COUNT = 0;
DGRAPH.PARSE_CALL_LIMIT = 100; // Set the limit to 100 calls
DGRAPH.PARSE_FILE_STACK = []; // Track the stack of files being parsed

// Function to extract functions from a file
// Function to extract functions from a file
DGRAPH.extractFunctions = async function(filePath, content) {
  if (DGRAPH.isMinifiedFile(content)) {
    console.log(`Skipping minified file: ${filePath}`);
    return [];
  }    
  await DGRAPH.initParser();
  const tree = await DGRAPH.parseAnyFile(filePath, content);
  const functions = [];
  const maxDepth = 30; // Set the maximum depth for traversal

  if (!tree || !tree.rootNode) {
    console.error(`Failed to parse file: ${filePath}`);
    return [];
  }  
  if (filePath.endsWith(".min.js")) {
      console.log("Skipping minified file");
      return [];
  }  
  if (filePath.endsWith(".json")) {
      console.log("Skipping json file");
      return [];
  }    
  
  console.log("DOING FILEPATH : " + filePath);

  function traverseNode(node, depth = 0) {
    if (depth > maxDepth) {
      return; // Stop traversing if depth exceeds maxDepth
    }
    
    const nodeType = node.type;
    const startLine = node.startPosition.row;
    const endLine = node.endPosition.row;
    const lineText = content.split(/\r?\n/)[startLine];

    if (nodeType === 'function_declaration') {
      //console.log(`1111 Node Type: ${nodeType}`);
      const functionNameNode = node.childForFieldName('name');
      if (functionNameNode) {
        const functionName = functionNameNode.text;
        functions.push({
          name: functionName,
          line_text: lineText,
          start_line: startLine,
          end_line: endLine,
          extract_needed: false,
          selectors: []
        });
      } else {
        let functionName = extractFunctionFromLine(lineText);
        if (functionName !== '') {
          functions.push({
            name: functionName,
            line_text: lineText,
            start_line: startLine,
            end_line: endLine,
            extract_needed: true,
            selectors: []
          });
        }
      }      
    } else if (nodeType === 'function') {
      //console.log(`2222 Node Type: ${nodeType}`);
      const functionNameNode = node.childForFieldName('name');
      if (functionNameNode) {
        const functionName = functionNameNode.text;
        functions.push({
          name: functionName,
          line_text: lineText,
          start_line: startLine,
          end_line: endLine,
          extract_needed: false,
          selectors: []
        });
      } else {
        let functionName = extractFunctionFromLine(lineText);
        if (functionName !== '') {
          functions.push({
            name: functionName,
            line_text: lineText,
            start_line: startLine,
            end_line: endLine,
            extract_needed: true,
            selectors: []
          });
        }
      }         
    } else if (nodeType === 'method_definition') {
      //console.log(`333 Node Type: ${nodeType}`);
      const methodNameNode = node.childForFieldName('name');
      if (methodNameNode) {
        let functionName = methodNameNode.text;
        functions.push({
          name: functionName,
          line_text: lineText,
          start_line: startLine,
          end_line: endLine,
          extract_needed: false,
          selectors: []
        });
      } else {
        let functionName = extractFunctionFromLine(lineText);
        if (functionName !== '') {
          functions.push({
            name: functionName,
            line_text: lineText,
            start_line: startLine,
            end_line: endLine,
            extract_needed: true,
            selectors: []
          });
        }
      }      
    } else {
      // console.log(`Skip Node Type: ${nodeType}`);        
    }

    for (const child of node.namedChildren) {
      traverseNode(child, depth + 1); // Increment depth when traversing children
    }
  }

  traverseNode(tree.rootNode);

  return functions;
};

function extractStringsFromLine(line) {
  const regex = /['"]([\w-$.#[\]:>+~=,\s]+)['"]/g;
  const matches = line.match(regex);
  if (matches) {
    return matches.map(match => match.slice(1, -1));
  }
  return [];
}

// Function to process files and build the function data structure
// Function to process files and build the function data structure
DGRAPH.buildFunctionDataStructure = async function(files) {
  const functionDataStructure = {};
  const selectorDataStructure = {};

  // Use Promise.all to wait for all file extractions to complete
  const filePromises = files.map(async (file) => {
    let theExt = path.extname(file).slice(1);  
    
    if(!DGRAPH.codeExtensions.includes(theExt)){
        //console.log("skipping building functions, not code file: " + file);
        let empty = new Array();
        return { file, empty };
    }
      
    //console.log("extracting file: " + file);
    
    const content = readFileWithCache(file);
    const functions = await DGRAPH.extractFunctions(file, content);
    selectorDataStructure[file] = [];

    for (let f = 0; f < functions.length; f++) {
      let line = functions[f].line_text;
      let lineNum = functions[f].start_line;

      if (line.includes('querySelector') || line.includes('$') || line.includes('addEventListener')) {
        //console.log("testing: " + functions[f].line_text);
        if (line.includes("$(") || line.includes('addEventListener')) {
          functions[f].name = "event";
        }

        // Extract potential selectors from the line
        const selectors = extractStringsFromLine(line);
        let validSelectors = [];
        let invalidSelectors = ['body', 'click', 'mousedown', 'mouseup', 'mousemove', 'mouseover', 'mouseout', 'mouseenter', 'mouseleave', 'dblclick',
          'keydown', 'keyup', 'keypress', 'focus', 'blur', 'change', 'input', 'submit', 'reset', 'load', 'scroll', 'resize', 'touchstart', 'dragstart', 'drag', 'contextmenu'];

        // Check each selector for validity
        selectors.forEach(selector => {
          if (isValidSelector(selector) && !invalidSelectors.includes(selector)) {
            validSelectors.push(selector);
            let selectorObject = {
              line: lineNum,
              selector: selector
            };
            selectorDataStructure[file].push(selectorObject);
          }
        });

        functions[f]['selectors'] = validSelectors;
        //console.log("valid selectors");
        //console.log(validSelectors);
      }

      if (line.includes('getElementById')) {
        let regex = /getElementById\((['"])(.+?)\1\)/;
        let match = line.match(regex);
        if (match) {
          let idSelector = "#" + match[2];
          functions[f]['selectors'].push(idSelector);
        }
      }
    }

    return { file, functions };
  });

  // Wait for all file extractions to complete
  const fileResults = await Promise.all(filePromises);

  // Process the file results and build the function data structure
  fileResults.forEach(({ file, functions }) => {
    functionDataStructure[file] = functions;
  });

  return [functionDataStructure, selectorDataStructure];
};

function extractStringsFromLine(line) {
  const regex = /['"]([\w-$.#[\]:>+~=,\s]+)['"]/g;
  const matches = line.match(regex);
  if (matches) {
    return matches.map(match => match.slice(1, -1));
  }
  return [];
}

function isValidSelector(selector) {
  try {
    cssWhat.parse(selector);
    return true;
  } catch (error) {
    return false;
  }
}

DGRAPH.initParser = async function() {
  if (DGRAPH.parser != null) {
    return;
  }
  await Parser.init();
  DGRAPH.parser = new Parser();
};

DGRAPH.getRelatedFiles = async function(lineNumberArr, projectName) {
  let relatedFiles = new Set();

  for (let lineNumber of lineNumberArr) {
    let [file, line] = lineNumber.split(':');
    let dependencies = await getDependencyGraph(projectName);

    for (let dep of dependencies) {
      if (dep.short_caller === file && dep.line_caller === parseInt(line)) {
        relatedFiles.add(dep.short_callee);
      }
      if (dep.short_callee === file && dep.line_callee === parseInt(line)) {
        relatedFiles.add(dep.short_caller);
      }
    }
  }

  return Array.from(relatedFiles);
};

DGRAPH.parseAnyFile = async function(filePath, fileData){
    
    //REMOVE THIS     //REMOVE THIS 
    //REMOVE THIS 
    //REMOVE THIS 
    //REMOVE THIS 
    //REMOVE THIS 
    //REMOVE THIS 
    //REMOVE THIS 
    /*
    DGRAPH.PARSE_CALL_COUNT++;
    if (DGRAPH.PARSE_CALL_COUNT > DGRAPH.PARSE_CALL_LIMIT) {
        console.log("Parse call limit reached. Breaking to debug.");
        throw new Error("Parse call limit reached");
    } */   
    // Check if the tree is already cached
    if (DGRAPH.TREE_CACHE[filePath]) {
        console.log("Using cached tree for file: " + filePath);
        return DGRAPH.TREE_CACHE[filePath];
    }    
    //console.log("Parsing file: " + filePath);
    DGRAPH.PARSE_FILE_STACK.push(filePath); // Add file to the parse stack
    try{
        await DGRAPH.initParser();

        let theExt = path.extname(filePath).slice(1);

        let langSetter = null;
        let myAppPath = path.dirname(__dirname);
        if(typeof(STUBS) !== 'undefined'){
            myAppPath = STUBS.isPackaged ? STUBS.appPath : __dirname;
        }
        /*
      baseWasm: path.join(__dirname, 'jslib', 'wasm', 'tree-sitter-javascript.wasm'),
      jsWasm: path.join(__dirname, 'jslib', 'wasm', 'tree-sitter-javascript.wasm'),
      phpWasm: path.join(__dirname, 'jslib', 'wasm', 'tree-sitter-php.wasm'),
         */
        let languageToFile = { 
          'base' : path.join(myAppPath, 'jslib', 'wasm', 'treesitter-javascript_v2.wasm'),
          'javascript' : path.join(myAppPath, 'jslib', 'wasm', 'treesitter-javascript_v2.wasm'),
          'bash' : path.join(myAppPath, 'jslib', 'wasm',  'tree-sitter-bash.wasm'),
          'c': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-c_v2.wasm'),
          'c_sharp': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-csharp_v2.wasm'),
          'cpp': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-cpp_v2.wasm'),
          'css': path.join(myAppPath, 'jslib', 'wasm', 'tree-sitter-css.wasm'),
          'elm': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-elm_v2.wasm'),
          'embedded_template': path.join(myAppPath, 'jslib', 'wasm', 'tree-sitter-embedded_template.wasm'),
          'go': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-go_v2.wasm'),
          'html': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-html_v2.wasm'),
          'tpl': path.join(myAppPath, 'jslib', 'wasm', 'tree-sitter-smarty.wasm'),          
          'java': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-java_v2.wasm'),
          'json': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-json_v2.wasm'),          
          'lua': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-lua_v2.wasm'),
          'markdown': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-markdown_v2.wasm'),
          'php': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-php_v4.wasm'),
          'python': path.join(myAppPath, 'jslib', 'wasm', 'tree-sitter-python.wasm'),
          'ruby': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-ruby_v5.wasm'),
          'rust': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-rust_v2.wasm'),
          'systemrdl': path.join(myAppPath, 'jslib', 'wasm', 'tree-sitter-systemrdl.wasm'),
          'vue': path.join(myAppPath, 'jslib', 'wasm', 'tree-sitter-vue.wasm'),
          'dart': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-dart_v2.wasm'),          
          'yaml': path.join(myAppPath, 'jslib', 'wasm', 'tree-sitter-yaml.wasm'),
           'typescript': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-typescript_v2.wasm'),
           'tsx': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-tsx_v2.wasm'),
           'swift': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-swift_v2.wasm'),
           'kotlin': path.join(myAppPath, 'jslib', 'wasm', 'treesitter-kotlin_v2.wasm')
        };      
        if (theExt == 'js') {         
          console.log("loading javascript wasm");
          langSetter = await Parser.Language.load(languageToFile['javascript']);
        } else if (theExt == 'jsx') {
          console.log("loading javascript wasm");            
          langSetter = await Parser.Language.load(languageToFile['javascript']);
        } else if (theExt == 'php') {
            console.log("loading php wasm");
          langSetter = await Parser.Language.load(languageToFile['php']);
        } else if (theExt == 'sh') {
            console.log("loading bash wasm");
          langSetter = await Parser.Language.load(languageToFile['bash']);
        } else if (theExt == 'c') {
            console.log("loading c wasm");
          langSetter = await Parser.Language.load(languageToFile['c']);
        } else if (theExt == 'cs') {
            console.log("loading c_sharp wasm");
          langSetter = await Parser.Language.load(languageToFile['c_sharp']);
        } else if (theExt == 'cpp' || theExt == 'cc' || theExt == 'cxx' || theExt == 'hpp' || theExt == 'hxx') {
            console.log("loading cpp wasm");
          langSetter = await Parser.Language.load(languageToFile['cpp']);
        } else if (theExt == 'css') {
          langSetter = await Parser.Language.load(languageToFile['css']);
        } else if (theExt == 'elm') {
          langSetter = await Parser.Language.load(languageToFile['elm']);
        } else if (theExt == 'html' || theExt == 'htm') {
          langSetter = await Parser.Language.load(languageToFile['html']);
        } else if (theExt == 'tpl') {
          langSetter = await Parser.Language.load(languageToFile['tpl']);
        } else if (theExt == 'java') {
          langSetter = await Parser.Language.load(languageToFile['java']);
        } else if (theExt == 'lua') {
          langSetter = await Parser.Language.load(languageToFile['lua']);
        } else if (theExt == 'md' || theExt == 'markdown') {
          langSetter = await Parser.Language.load(languageToFile['markdown']);
        } else if (theExt == 'py') {
          langSetter = await Parser.Language.load(languageToFile['python']);
        } else if (theExt == 'rb') {
            console.log("loading ruby " + languageToFile['ruby']);
          langSetter = await Parser.Language.load(languageToFile['ruby']);
        } else if (theExt == 'rs') {
          langSetter = await Parser.Language.load(languageToFile['rust']);
        } else if (theExt == 'json') {
          langSetter = await Parser.Language.load(languageToFile['json']);
        } else if (theExt == 'vue') {
          langSetter = await Parser.Language.load(languageToFile['vue']);
        } else if (theExt == 'yaml' || theExt == 'yml') {
          langSetter = await Parser.Language.load(languageToFile['yaml']);
        } else if (theExt == 'dart' || theExt == 'dart') {
          langSetter = await Parser.Language.load(languageToFile['dart']);
        } else if (theExt == 'go') {
            langSetter = await Parser.Language.load(languageToFile['go']);
        } else if (theExt == 'ts') {
            langSetter = await Parser.Language.load(languageToFile['typescript']);
        }else if (theExt == 'tsx') {
            langSetter = await Parser.Language.load(languageToFile['typescript']);
        }else if (theExt == 'swift') {
            langSetter = await Parser.Language.load(languageToFile['swift']);
        }else if (theExt == 'kt') {
            langSetter = await Parser.Language.load(languageToFile['kotlin']);
        }              
        else {
          console.log("Unknown language: " + theExt + " from file " + filePath);
          return;
        }
        
      if (langSetter) {
            DGRAPH.parser.setLanguage(langSetter);
            let tree = await DGRAPH.parser.parse(fileData);
            return tree;
        } else {
            console.log("Unknown or unsupported language: " + theExt + " from file " + filePath);
            return;
        }        
    }catch(exxx){
        console.error('Error initializing web-tree-sitter for path: ' + filePath, exxx);
        return;        
    }
    
    return;
};

var extractFunctionFromLine = function(lineText, extension = 'js') {
    let functionName = lineText.trim();
    
    const languageRules = {
        'js': {
            keywords: ['var', 'let', 'const', 'async', 'function', 'class', 'static', 'get', 'set', 'return'],
            extraCleanup: ['=>']
        },
        'jsx': {
            keywords: ['var', 'let', 'const', 'async', 'function', 'class', 'static', 'get', 'set', 'return'],
            extraCleanup: ['=>']
        },
        'ts': {
            keywords: ['var', 'let', 'const', 'async', 'function', 'class', 'static', 'get', 'set', 'return', 'interface', 'type'],
            extraCleanup: ['=>']
        },
        'tsx': {
            keywords: ['var', 'let', 'const', 'async', 'function', 'class', 'static', 'get', 'set', 'return', 'interface', 'type'],
            extraCleanup: ['=>']
        },
        'php': {
            keywords: ['function', 'public', 'private', 'protected', 'static', 'final', 'abstract'],
            extraCleanup: ['function']
        },
        'go': {
            keywords: ['func'],
            extraCleanup: []
        },
        'py': {
            keywords: ['def', 'class', 'async'],
            extraCleanup: []
        },
        'rb': {
            keywords: ['def', 'class', 'module', 'private', 'protected', 'public'],
            extraCleanup: []
        },
        'rs': {
            keywords: ['fn', 'pub', 'async', 'const', 'static', 'unsafe'],
            extraCleanup: []
        },
        'swift': {
            keywords: ['func', 'class', 'struct', 'enum', 'protocol', 'extension', 'public', 'private', 'fileprivate', 'internal', 'open'],
            extraCleanup: []
        },
        'java': {
            keywords: ['public', 'private', 'protected', 'static', 'final', 'abstract', 'synchronized', 'native', 'default', 'void'],
            extraCleanup: []
        },
        'dart': {
            keywords: ['void', 'Function', 'static', 'final', 'const', 'var', 'late', 'async', 'sync*', 'async*'],
            extraCleanup: []
        },
        'kt': {
            keywords: ['fun', 'class', 'object', 'interface', 'val', 'var', 'suspend', 'private', 'protected', 'public', 'internal'],
            extraCleanup: []
        },
        'cpp': {
            keywords: ['void', 'int', 'float', 'double', 'char', 'bool', 'auto', 'const', 'static', 'virtual', 'inline', 'explicit', 'friend', 'volatile', 'class', 'struct', 'enum'],
            extraCleanup: []
        },
        'c': {
            keywords: ['void', 'int', 'float', 'double', 'char', 'auto', 'const', 'static', 'inline', 'volatile', 'struct', 'enum'],
            extraCleanup: []
        },
        'cs': {
            keywords: ['public', 'private', 'protected', 'internal', 'static', 'virtual', 'abstract', 'override', 'async', 'void', 'int', 'string', 'bool', 'var'],
            extraCleanup: []
        },
        'lua': {
            keywords: ['function', 'local'],
            extraCleanup: []
        }
    };

    // Default to JavaScript rules if the extension is not recognized
    const rules = languageRules[extension] || languageRules['js'];

    functionName = functionName.split('(')[0];
    
    rules.keywords.forEach(keyword => {
        functionName = replaceAll(functionName, keyword + ' ', '');
    });

    rules.extraCleanup.forEach(item => {
        functionName = replaceAll(functionName, item, '');
    });

    // Common cleanup for all languages
    const commonCleanup = ['\n', '\r', ':', '=', '}', ',', 'setTimeout'];
    commonCleanup.forEach(item => {
        functionName = replaceAll(functionName, item, '');
    });
    
    return functionName.trim();
};

  function getFunctionNameForNode(node, fileContent) {
    const nodeType = node.type;
    const startLine = node.startPosition.row;
    const endLine = node.endPosition.row;
    const lineText = fileContent.split(/\r?\n/)[startLine];

    if (nodeType === 'function_declaration') {
      const functionNameNode = node.childForFieldName('name');
      if (functionNameNode) {
        const functionName = functionNameNode.text;
        return functionName;
      }
      else{
        let functionName = extractFunctionFromLine(lineText);
        if(functionName != ''){
            return functionName;
        }
      }      
    } 
    else if(nodeType == 'function'){
      const functionNameNode = node.childForFieldName('name');
      if (functionNameNode) {
        const functionName = functionNameNode.text;
        return functionName;
      }
      else{
        let functionName = extractFunctionFromLine(lineText);
        if(functionName != ''){
            return functionName;
        }
      }         
    }
    else if (nodeType === 'method_definition') {
      const methodNameNode = node.childForFieldName('name');
      if (methodNameNode) {
        let functionName = methodNameNode.text;
        return functionName;
      }
      else{
        let functionName = extractFunctionFromLine(lineText);
        return functionName;
      }
  }
  return null;

}

function getFunctionNameForNode2(node, fileContent) {
  const nodeType = node.type;

  if (nodeType === 'function_declaration' || nodeType === 'method_definition' || nodeType === 'async_function_definition') {
    const functionNameNode = node.childForFieldName('name');
    if (functionNameNode) {
      return functionNameNode.text;
    }
  } else if (nodeType === 'variable_declaration') {
    const declarator = node.childForFieldName('declarator');
    if (declarator) {
      const nameNode = declarator.childForFieldName('name');
      if (nameNode) {
        return nameNode.text;
      }
    }
  } else if (nodeType === 'variable_declarator' && node.parent.type === 'variable_declaration') {
    const nameNode = node.childForFieldName('name');
    if (nameNode) {
      return nameNode.text;
    }
  }
  
  return extractFunctionFromLine(fileContent.split(/\r?\n/)[node.startPosition.row]);
}

// Function to extract function calls from a file
DGRAPH.extractFunctionCallsJavascript = async function(file) {
  const content = readFileWithCache(file);
  const tree = await DGRAPH.parseAnyFile(file, content);
  
  const functionCalls = [];
  if (!tree || !tree.rootNode) {
    console.error(`Failed to parse file: ${file}`);
    return functionCalls;
  }  

  // Traverse the AST to find function calls
  tree.rootNode.descendantsOfType('call_expression').forEach((callNode) => {
    const functionNameNode = callNode.childForFieldName('function');
    const jqueryFunctionCalls = [
      ".css", ".data", ".html", ".attr", ".val", ".text", ".append", ".prepend", ".on(",
      ".after", ".before", ".remove", ".empty", ".show", ".hide", ".toggle",
      ".addClass", ".removeClass", ".toggleClass", ".hasClass", ".width",
      ".height", ".offset", ".position", ".scrollLeft", ".scrollTop", ".animate"
    ];

    const javascriptCalls = [
      "console.log", "console.error", "console.warn", "console.info", "console.debug",
      "console.trace", ".preventDefault", ".stopPropagation", ".addEventListener", ".removeEventListener", ".dispatchEvent", ".setTimeout", ".clearTimeout", ".setInterval", ".clearInterval",
      ".requestAnimationFrame", ".cancelAnimationFrame", ".getElementById", ".getElementsByClassName", ".getElementsByTagName",
      ".querySelector", ".querySelectorAll", ".createElement", ".createTextNode", ".createDocumentFragment",
      ".appendChild", ".insertBefore", ".removeChild", ".replaceChild", ".cloneNode", ".contains", ".compareDocumentPosition",
      ".getAttribute", ".setAttribute", ".removeAttribute", ".hasAttribute", ".focus", ".blur", ".click", ".submit",
      ".getElementsByName", ".item", ".namedItem", ".getBoundingClientRect", ".getClientRects",".getComputedStyle", ".matchMedia",
      ".getSelection", ".find", ".findIndex",".forEach", ".map", ".filter", ".reduce", ".reduceRight", ".some", ".every",
      ".indexOf", ".lastIndexOf", ".includes", ".sort", ".reverse",".stringify", ".parse", ".push", ".pop", ".shift", ".unshift",
      ".slice", ".splice", ".concat", ".join", ".split",".startsWith", ".endsWith", ".substring", ".substr", ".replace", ".trim",
      ".toLowerCase", ".toUpperCase", ".charCodeAt", ".fromCharCode",".parseInt", ".parseFloat", ".isNaN", ".isFinite",
      "Math.abs", "Math.ceil", "Math.floor", "Math.round", "Math.max", "Math.min",
      "Math.random", "Date.now", "Date.parse", "Date.UTC",".getFullYear", ".getMonth", ".getDate", ".getDay", ".getHours", ".getMinutes",
      ".getSeconds", ".getMilliseconds", ".getTime", ".getTimezoneOffset", ".toDateString", ".toISOString", ".toJSON",
      ".toLocaleDateString", ".toLocaleString", ".toLocaleTimeString", ".toString",
      ".toTimeString", ".toUTCString", ".valueOf"
    ];    

    if (functionNameNode) {
      const functionName = functionNameNode.text;
      const callerName = getFunctionName(callNode);
      const startLine = callNode.startPosition.row;
      const endLine = callNode.endPosition.row;
      const lineText = content.split(/\r?\n/)[startLine];

      // Check if the line contains any jQuery function calls
      const isJQueryCall = jqueryFunctionCalls.some((call) => lineText.includes(call));
      const isJavascriptCall = javascriptCalls.some((call) => lineText.includes(call));
      let parentFunctionName = "";
      let currentNode = callNode.parent;
      while (currentNode) {
        let funName = getFunctionNameForNode(currentNode, content);
        if(funName != null){
            //console.log("Found function name " + funName);
            parentFunctionName = funName;
            break;
        }
        currentNode = currentNode.parent;
      }   
      

      if (!isJQueryCall && !isJavascriptCall) {
        functionCalls.push({
          caller: callerName,
          callee: functionName,
          start_line: startLine,
          end_line: endLine,
          line_text: lineText,
          parent_function: parentFunctionName
        });
        
                
      }
    }
  });

  return functionCalls;
};

// Function to extract function calls from a file
// Function to extract function calls from a PHP file
DGRAPH.extractFunctionCallsPHP = async function(file) {
  const content = readFileWithCache(file);
  const tree = await DGRAPH.parseAnyFile(file, content);
  
  const functionCalls = [];
  if (!tree || !tree.rootNode) {
    console.error(`Failed to parse file: ${file}`);
    return functionCalls;
  }

    const phpBuiltInFunctions = [
      'abs', 'acos', 'acosh', 'addcslashes', 'addslashes', 'array_change_key_case',
      'array_chunk', 'array_column', 'array_combine', 'array_count_values', 'array_diff',
      'array_fill', 'array_filter', 'array_flip', 'array_intersect', 'array_key_exists',
      'array_keys', 'array_map', 'array_merge', 'array_multisort', 'array_pad',
      'array_pop', 'array_product', 'array_push', 'array_rand', 'array_reduce',
      'array_replace', 'array_reverse', 'array_search', 'array_shift', 'array_slice',
      'array_splice', 'array_sum', 'array_unique', 'array_unshift', 'array_values',
      'array_walk', 'arsort', 'asin', 'asinh', 'asort', 'assert', 'atan',
      'atan2', 'base64_decode', 'base64_encode', 'basename', 'bin2hex', 'call_user_func',
      'call_user_func_array', 'ceil', 'chdir', 'checkdate', 'chmod', 'chop',
      'chr', 'chunk_split', 'class_exists', 'closedir', 'compact', 'constant',
      'copy', 'cos', 'cosh', 'count', 'crc32', 'crypt', 'current',
      'date', 'decbin', 'dechex', 'decoct', 'define', 'defined', 'dirname',
      'each', 'empty', 'end', 'error_log', 'error_reporting', 'escapeshellarg',
      'escapeshellcmd', 'exec', 'exp', 'explode', 'fclose', 'feof',
      'file_exists', 'file_get_contents', 'file_put_contents', 'fileatime', 'filemtime',
      'filesize', 'floor', 'flush', 'fopen', 'fread', 'fseek',
      'fsockopen', 'ftp_connect', 'ftp_login', 'ftp_close', 'func_get_arg', 'func_get_args',
      'func_num_args', 'function_exists', 'fwrite', 'get_class', 'get_class_methods', 'get_class_vars',
      'get_object_vars', 'get_parent_class', 'getdate', 'getenv', 'gethostbyaddr', 'gethostbyname',
      'gethostbynamel', 'getimagesize', 'getrandmax', 'gettype', 'glob', 'gmdate',
      'gmmktime', 'gzcompress', 'gzdeflate', 'gzencode', 'gzinflate', 'gzopen',
      'gzwrite', 'hash', 'header', 'hex2bin', 'hexdec', 'htmlentities',
      'htmlspecialchars', 'http_build_query', 'idate', 'implode', 'in_array', 'intval',
      'ip2long', 'is_array', 'is_bool', 'is_callable', 'is_dir', 'is_file',
      'is_float', 'is_int', 'is_null', 'is_numeric', 'is_object', 'is_readable',
      'is_resource', 'is_string', 'is_writable', 'isset', 'join', 'json_decode',
      'json_encode', 'key', 'krsort', 'ksort', 'lcfirst', 'levenshtein',
      'log', 'log10', 'long2ip', 'ltrim', 'mail', 'max',
      'md5', 'memory_get_peak_usage', 'memory_get_usage', 'microtime', 'min', 'mkdir',
      'move_uploaded_file', 'mt_rand', 'natcasesort', 'natsort', 'next', 'nl2br',
      'number_format', 'opendir', 'ord', 'pack', 'parse_url', 'pathinfo',
      'pow', 'preg_match', 'preg_match_all', 'preg_replace', 'preg_split', 'prev',
      'print_r', 'printf', 'rand', 'range', 'readdir', 'readfile',
      'realpath', 'rename', 'reset', 'round', 'rsort', 'rtrim',
      'serialize', 'session_destroy', 'session_start', 'session_unset', 'sha1', 'sin',
      'sinh', 'sizeof', 'sleep', 'sort', 'sprintf', 'sqrt',
      'srand', 'str_pad', 'str_repeat', 'str_replace', 'str_split', 'strcasecmp',
      'strchr', 'strcmp', 'strip_tags', 'stripos', 'stripslashes', 'stristr',
      'strlen', 'strnatcasecmp', 'strnatcmp', 'strpos', 'strrev', 'strrpos',
      'strstr', 'strtolower', 'strtotime', 'strtoupper', 'strtr', 'strval',
      'substr', 'tan', 'tanh', 'time', 'trim', 'ucfirst',
      'ucwords', 'uniqid', 'unlink', 'unpack', 'unserialize', 'urldecode',
      'urlencode', 'usort', 'var_dump', 'version_compare'
    ];

  // Traverse the AST to find function calls
  tree.rootNode.descendantsOfType('call_expression').forEach((callNode) => {
    const functionNameNode = callNode.childForFieldName('function');
    const startLine = callNode.startPosition.row;
    const endLine = callNode.endPosition.row;
    const lineText = content.split(/\r?\n/)[startLine];

    if (functionNameNode) {
      const functionName = functionNameNode.text;
      const callerName = getFunctionName(callNode);

      // Check if the function is a built-in PHP function
      const isBuiltInFunction = phpBuiltInFunctions.includes(functionName);

      if (!isBuiltInFunction) {
        functionCalls.push({
          caller: callerName,
          callee: functionName,
          start_line: startLine,
          end_line: endLine,
          line_text: lineText,
          parent_function: ""
        });
      }
    }
  });

  return functionCalls;
};
// Function to extract selectors from an HTML file
DGRAPH.extractSelectorsFromHtml = function(file) {
  if(!fs.existsSync(file)){
      console.log("HTML FILE MISSING: " + file);
      return [];
  }
  const content = readFileWithCache(file);
  const $contentHtml = cheerio.load(content);
  const selectors = [];
  const lines = content.split(/\r?\n/);

  $contentHtml.root().find("*").each(function(index, element){
      //console.log("processing element: " );
      //console.log($contentHtml.html(element));
    const $element = $contentHtml(element);
    const id = $element.attr('id');
    const classes = $element.attr('class');
    
    const startIndex = $contentHtml.html().indexOf($contentHtml.html(element));
    const lineNumber = content.substr(0, startIndex).split(/\r?\n/).length;

    if (id) {
        //console.log("FOUND ID " + id + " on line " + lineNumber);
      selectors.push({ selector: `#${id}`, line: lineNumber });
    }

    if (classes) {  
      classes.split(' ').forEach((className) => {
        //console.log("FOUND CLASS " + className + " on line " + lineNumber);    
        selectors.push({ selector: `.${className}`, line: lineNumber });
      });
    }
  });

  return selectors;
};
DGRAPH.isMinifiedFile = function(fileContent) {
    const lines = fileContent.split(/\r?\n/);
    let inMultiLineComment = false;
    
    for (const line of lines) {
        const trimmedLine = line.trim();
        
        // Skip empty lines
        if (trimmedLine === '') continue;
        
        // Handle multi-line comments
        if (inMultiLineComment) {
            if (trimmedLine.includes('*/')) {
                inMultiLineComment = false;
            }
            continue;
        }
        
        // Check for start of multi-line comment
        if (trimmedLine.startsWith('/*')) {
            inMultiLineComment = true;
            continue;
        }
        
        // Skip single-line comments
        if (trimmedLine.startsWith('//')) continue;
        
        // If we've reached this point, we have our first non-comment, non-empty line
        return trimmedLine.length > 1000;
    }
    
    // If we've gone through all lines without finding a non-comment line, it's not minified
    return false;
};

DGRAPH.extractSelectorsFromCss = function(file) {
    
    if(!fs.existsSync(file)){
        console.log("MISSING CSS FILE: " + file);
        return [];
    }
    
   var getLineIndex = function(selector, minLineIndex, totalLines, lines) {
    if (minLineIndex >= totalLines) {
      return totalLines;
    }

    const nextOccurrence = lines.slice(minLineIndex).findIndex(line => line.includes(selector));

    if (nextOccurrence === -1) {
      return minLineIndex;
    }

    return minLineIndex + nextOccurrence;
  };    
  const content = readFileWithCache(file);
  const lines = content.split(/\r?\n/);
  const totalLines = lines.length;
  const parser = new cssjs.cssjs();
  let selectors = [];
  let parsedCssNew = parser.parseCSS(content);
  let minLineIndex = 0;

  for (var v = 0; v < parsedCssNew.length; v++) {
    //console.log("index" + v);
    //console.log(parsedCssNew[v]);

    if (parsedCssNew[v]['selector'].indexOf('@media') >= 0) {
      if ('subStyles' in parsedCssNew[v] && parsedCssNew[v]['subStyles'].length > 0) {
        for (let s = 0; s < parsedCssNew[v]['subStyles'].length; s++) {
          const selector = parsedCssNew[v]['subStyles'][s]['selector'];
          const lineIndex = getLineIndex(selector, minLineIndex, totalLines, lines);
          selectors.push({ selector, line: lineIndex });

          if ('rules' in parsedCssNew[v]['subStyles'][s] && parsedCssNew[v]['subStyles'][s]['rules'].length > 0) {
            minLineIndex += parsedCssNew[v]['subStyles'][s]['rules'].length;
          }
        }
      } else {
        const selector = parsedCssNew[v]['selector'];
        const lineIndex = getLineIndex(selector, minLineIndex, totalLines, lines);
        selectors.push({ selector, line: lineIndex });
      }
    } else {
      const selector = parsedCssNew[v]['selector'];
      const lineIndex = getLineIndex(selector, minLineIndex, totalLines, lines);
      selectors.push({ selector, line: lineIndex });
    }

    if ('rules' in parsedCssNew[v] && parsedCssNew[v]['rules'].length > 0) {
      minLineIndex += parsedCssNew[v]['rules'].length;
    }

    minLineIndex++;
  }

  return selectors;
};

DGRAPH.getControlStructureForCode = async function(filepath, code) {
   
  let sPathExtFull = path.extname(filepath);
  let sPathExt = sPathExtFull.slice(1);        
  if(DGRAPH.markupExtensions.includes(sPathExt)){ 
      //console.log("returning markup for control structure");
      return "markup";
  } 
  // Remove leading and trailing whitespace from the code
  code = code.trim();

  // Find the first non-empty line that is not a comment
  const lines = code.split(/\r?\n/);
  let firstLine = '';
  for (const line of lines) {
    const trimmedLine = line.trim();
    if (trimmedLine !== '' && !trimmedLine.startsWith('//') && !trimmedLine.startsWith('/*')) {
      firstLine = trimmedLine;
      break;
    }
  }

  // If no valid first line is found, return 'empty'
  if (firstLine === '') {
    return 'empty';
  }

  await DGRAPH.initParser();
  const tree = await DGRAPH.parseAnyFile(filepath, code);

  if (!tree || !tree.rootNode) {
    console.error('Failed to parse code', code);
    return 'parse_error';
  }

  let result = 'top_level';
  let isFullFunction = false;

  function traverseNode(node, level) {
    const startLine = node.startPosition.row;
    const endLine = node.endPosition.row;

    //console.log("On startline " + startLine + " with " + node.type + " at level " + level);

    if (startLine === 0) {
      if (node.type === 'function_declaration' || node.type === 'method_definition' || node.type === 'class_declaration' || node.type === 'function' || node.type === 'async_function_definition') {
        result = 'start_function';

        // Check if the function or class ends at the last line of the code
        if (endLine === lines.length - 1) {
          isFullFunction = true;
        }
      } else if (node.type === 'if_statement' || node.type === 'for_statement' || node.type === 'while_statement') {
        result = 'control_structure';
      } else if (node.type === 'variable_declaration') {
        for (let child of node.namedChildren) {
          if (child.startPosition.row === startLine) {
            traverseNode(child, level + 1);
            if (result === 'start_function') {
              break;
            }
          }
        }
      } else {
        // Traverse deeper to find a function declaration
        for (const child of node.namedChildren) {
          if (child.startPosition.row === startLine) {
            traverseNode(child, level + 1);
            if (result === 'start_function') {
              break;
            }
          }
        }
      }
    }

    if (!isFullFunction) {
      for (const child of node.namedChildren) {
        traverseNode(child, level + 1);
      }
    }
  }

  traverseNode(tree.rootNode, 0);

  if (isFullFunction) {
    result = 'full_function';
  }

  return result;
};

function countIndentation(line) {
  let count = 0;
  for (let i = 0; i < line.length; i++) {
    if (line[i] === ' ') {
      count++;
    } else if (line[i] === '\t') {
      count += 4; // Assuming a tab is equivalent to 4 spaces
    } else {
      break;
    }
  }
  return count;
}

// Function to process files and build the dependency graph

// Function to process files and build the dependency graph
DGRAPH.buildDependencyGraph = function(files, functionDataStructure, functionCallsDataStructure, selectorDataStructure) {
  const dependencyGraph = {};

  files.forEach((file) => {
    if (!dependencyGraph[file]) {
      dependencyGraph[file] = {};
    }      
    const fileExtension = path.extname(file);
    const theExt = fileExtension.slice(1);  

    if (DGRAPH.codeExtensions.includes(theExt)) {
      const functionCalls = functionCallsDataStructure[file];

      functionCalls.forEach((functionCall) => {
        const callerName = functionCall.caller;
        const functionName = functionCall.callee;   
        try {
            // Find the file where the called function is defined
            const calleeFile = findFunctionFile(functionName, functionDataStructure);

            if (calleeFile) {
              // Add the function call to the dependency graph
              if (!dependencyGraph[file]) {
                dependencyGraph[file] = {};
              }
              if (!dependencyGraph[file][callerName]) {
                dependencyGraph[file][callerName] = [];
              }
              dependencyGraph[file][callerName].push(`${calleeFile}/${functionName}`);
            }
        }catch (error) {
          console.error(`Error processing function call: ${functionName}`, error);
          // Continue with the next iteration
        }
      });
    } else if (DGRAPH.markupExtensions.includes(theExt)) {
      const selectors = selectorDataStructure[file];

      selectors.forEach((selector) => {
        if (!dependencyGraph[file]) {
          dependencyGraph[file] = {};
        }
        if (!dependencyGraph[file][selector.selector]) {
          dependencyGraph[file][selector.selector] = [];
        }
        dependencyGraph[file][selector.selector].push(`${file}:${selector.line}`);
      });
    } else if (DGRAPH.styleExtensions.includes(theExt)) {
      const selectors = selectorDataStructure[file];

      selectors.forEach((selector) => {
        if (!dependencyGraph[file]) {
          dependencyGraph[file] = {};
        }
        if (!dependencyGraph[file][selector.selector]) {
          dependencyGraph[file][selector.selector] = [];
        }
        dependencyGraph[file][selector.selector].push(`${file}:${selector.line}`);
      });
    }
    
  });

  return dependencyGraph;
};


DGRAPH.buildDependencyGraphArray = function(files, functionDataStructure, functionCallsDataStructure, functionSelectors, selectorDataStructure, projectPath) {
  const dependencies = [];
  const selectorMatchCache = new Map();
  const MAX_SELECTORS_PER_FILE = 1000;

  files.forEach((file) => {
    const fileExtension = path.extname(file);
    const theExt = fileExtension.slice(1);  
    let shortFilepath = path.normalize(file.replace(projectPath, '')).replace(/\\/g, '/');
    //console.log("dep graph short path " + shortFilepath);
    let cssSelCount = 0;
    let htmlSelCount = 0;    

    if (DGRAPH.codeExtensions.includes(theExt)) {
      //console.log("IN JS FILE " + file);
      const functionCalls = functionCallsDataStructure[file];
      const mySelectors = functionSelectors[file];
      const addedDependencies = []; // Array to keep track of added dependencies for the current file
      
      if(file in functionCallsDataStructure && typeof(functionCallsDataStructure[file]) !== 'undefined'){
          
      }else{
          console.log("function calls in " + file + " invalid");
          return;
      }
      
      
      //function to function in js files
      functionCalls.forEach((functionCall) => {
        const callerName = functionCall.caller;
        const functionName = functionCall.callee;
        const lineText = functionCall.line_text;

        // Find the file where the called function is defined
        const calleeFile = findFunctionFile(functionName, functionDataStructure);
        if (calleeFile) {
          let shortCalleePath = path.normalize(calleeFile.replace(projectPath, '')).replace(/\\/g, '/');            
          // Find the line number of the called function in the callee file
          const calleeFunction = functionDataStructure[calleeFile].find((func) => func.name === functionName);
          const calleeLineNumber = calleeFunction ? calleeFunction.start_line : null;
          // Check if the dependency already exists for the current file
          const dependencyExists = addedDependencies.some(dep => 
            dep.short_caller === shortFilepath && 
            dep.line_caller === functionCall.start_line &&
            dep.short_callee === shortCalleePath
          );
  
          const dependencyExists2 = addedDependencies.some(dep => 
            dep.short_caller === shortFilepath && 
            dep.line_callee === calleeLineNumber &&
            dep.short_callee === shortCalleePath
          );  
          
          if (!dependencyExists && !dependencyExists2) {
            dependencies.push({
              caller: file,
              short_caller: shortFilepath,
              line_caller: functionCall.start_line,
              callee: calleeFile,
              short_callee: shortCalleePath,
              line_callee: calleeLineNumber,
              type: 'function',
              name: functionName,
              parent_function: functionCall.parent_function
            });
            addedDependencies.push({
              short_caller: shortFilepath,
              line_caller: functionCall.start_line,
              line_callee: calleeLineNumber,
              short_callee: shortCalleePath
            });
          }                 
        }
      });
      
      //function to selector in .css or .html files
      mySelectors.forEach((functionSel) => {
          console.log("functionSel ");
          console.log(functionSel);          
          
            for(let f = 0; f < files.length; f++){
              let styleFile = files[f];
              let shortStylePath = path.normalize(styleFile.replace(projectPath, '')).replace(/\\/g, '/');
              let styleExt = path.extname(styleFile);
              let shortStyleExt = styleExt.slice(1); 
              if(DGRAPH.markupExtensions.includes(shortStyleExt)){
                 const cacheKey = `${shortFilepath}-${shortStylePath}-${functionSel.selector}`;
                 if (selectorMatchCache.has(cacheKey)) {
                     console.log("already processed " . cacheKey);
                     continue; // Skip if already processed this combination
                 }                  
                 if(styleFile in selectorDataStructure && typeof(selectorDataStructure[styleFile]) !== 'undefined'){

                 }else{
                    console.log("selector structure in " + styleFile + " invalid");
                    return;
                 }
                  
                 let htmlSelectorArr = selectorDataStructure[styleFile];
                 htmlSelectorArr = DGRAPH.shuffleArray([...htmlSelectorArr]);
                 for(let h = 0; h < htmlSelectorArr.length; h++){
                     if (htmlSelCount >= MAX_SELECTORS_PER_FILE) {
                        break; // Stop processing if the limit is reached
                     }                     
                     let hSel = htmlSelectorArr[h];
                     //console.log("testing html");
                     //console.log(hSel);
                     if(hSel['selector'] == functionSel['selector']){
                        const dependencyExists = addedDependencies.some(dep => 
                          (dep.short_caller == shortFilepath && 
                          dep.line_caller == functionSel.line &&
                          dep.short_callee == shortStylePath)
                        );
                        const dependencyExists2 = addedDependencies.some(dep => 
                          (dep.short_caller == shortFilepath && 
                          dep.line_callee == hSel.line &&
                          dep.short_callee == shortStylePath)
                        );                  

                        if (!dependencyExists && !dependencyExists2) {
                          dependencies.push({
                            caller: file,
                            short_caller: shortFilepath,
                            line_caller: functionSel.line,
                            callee: styleFile,
                            short_callee: shortStylePath,
                            line_callee: hSel.line,
                            type: 'selector',
                            name: functionSel['selector'],
                            parent_function: ""
                          });
                          addedDependencies.push({
                            short_caller: shortFilepath,
                            line_caller: functionSel.line,
                            line_callee: hSel.line,
                            short_callee: shortStylePath
                          });
                          htmlSelCount++; 
                        }  
                        
                        //console.log("added");
                        //console.log(dependencies);
                     }
                 }
                 selectorMatchCache.set(cacheKey, true);
                 
              }else if(DGRAPH.styleExtensions.includes(shortStyleExt)){
                 let cssSelectorArr = selectorDataStructure[styleFile];
                 cssSelectorArr = DGRAPH.shuffleArray([...cssSelectorArr]);
                 for(let c = 0; c < cssSelectorArr.length; c++){
                     if (cssSelCount >= MAX_SELECTORS_PER_FILE) {
                        break; // Stop processing if the limit is reached
                     }                             
                     let cSel = cssSelectorArr[c];
                     //console.log("testing css");
                     //console.log(cSel);
                     
                     if(cSel['selector'].indexOf(functionSel['selector']) >= 0){
                        const dependencyExists = addedDependencies.some(dep => 
                          (dep.short_caller == shortFilepath && 
                          dep.line_caller == functionSel.line &&
                          dep.short_callee == shortStylePath)
                        );
                        const dependencyExists2 = addedDependencies.some(dep => 
                          (dep.short_caller == shortFilepath && 
                          dep.line_callee == cSel.line &&
                          dep.short_callee == shortStylePath)
                        );                     

                        if (!dependencyExists && !dependencyExists2) {
                          dependencies.push({
                            caller: file,
                            short_caller: shortFilepath,
                            line_caller: functionSel.line,
                            callee: styleFile,
                            short_callee: shortStylePath,
                            line_callee: cSel.line,
                            type: 'selector',
                            name: functionSel['selector'],
                            parent_function: ""
                          });
                          addedDependencies.push({
                            short_caller: shortFilepath,
                            line_caller: functionSel.line,
                            line_callee: cSel.line,
                            short_callee: shortStylePath
                          });
                          cssSelCount++; 
                        }   
                        
                        //console.log("added");
                        //console.log(dependencies);
                        
                     }
                 }
              }
            }
      });
      
    } else if (DGRAPH.markupExtensions.includes(theExt)) {
        //FUCK THIS DEPENDENCY GRAPH BS
        /*
      const selectors = selectorDataStructure[file];      
      const addedDependencies = []; // Array to keep track of added dependencies for the current file
      selectors.forEach((selector) => {
        let htmlSelector = selector.selector;
        let htmlLine = selector.line;
        for(let f = 0; f < files.length; f++){
            let styleFile = files[f];
            let shortStylePath = path.normalize(styleFile.replace(projectPath, '')).replace(/\\/g, '/');
            let styleExt = path.extname(styleFile);         
            let shortStyleExt = styleExt.slice(1); 
            if(DGRAPH.styleExtensions.includes(shortStyleExt)){
               let cssSelectorArr = selectorDataStructure[styleFile];
               for(let c = 0; c < cssSelectorArr.length; c++){
         
                   let cSel = cssSelectorArr[c];                        
                   if(cSel['selector'].indexOf(htmlSelector) >= 0){
                        const dependencyExists = addedDependencies.some(dep => 
                          (dep.short_caller == shortFilepath && 
                          dep.line_caller == htmlLine &&
                          dep.short_callee == shortStylePath)
                        );
                        const dependencyExists2 = addedDependencies.some(dep =>
                          (dep.short_caller == shortFilepath &&
                          dep.line_callee == cSel.line &&
                          dep.short_callee == shortStylePath)
                        );    
                        if (!dependencyExists && !dependencyExists2) {
                            //console.log("WE HAVE TO FIX THIS WHY ");
                            
                          dependencies.push({
                            caller: file,
                            short_caller: shortFilepath,
                            line_caller: htmlLine,
                            callee: styleFile,
                            short_callee: shortStylePath,
                            line_callee: cSel.line,
                            type: 'selector',
                            name: htmlSelector,
                            parent_function: cSel['selector']
                          }); 
                                    
                        }
                   }
               }
            }              
        }
      });
            */
    } else if (fileExtension === '.css') {
      //const selectors = selectorDataStructure[file];
    }
  });

  return dependencies;
};

DGRAPH.shuffleArray = function(array) {
    for (let i = array.length - 1; i > 0; i--) {
        const j = Math.floor(Math.random() * (i + 1));
        [array[i], array[j]] = [array[j], array[i]];
    }
    return array;
};

DGRAPH.searchDependencyGraph = async function(projectName, fileName, functionName) {
  const dependencyGraph = await getDependencyGraph(projectName);
  const fileNames = await getFileNamesFromDependencyGraph(projectName);

  if (!fileNames.includes(fileName)) {
    console.log(`File ${fileName} not found in the dependency graph.`);
    return [];
  }

  const result = [];

  function traverseDependencies(file, func) {
    const dependencies = dependencyGraph.filter(dep => dep.caller === file && dep.name === func);

    dependencies.forEach(dep => {
      result.push(dep);
      traverseDependencies(dep.callee, dep.name);
    });
  }

  traverseDependencies(fileName, functionName);

  return result;
}



// Function to find the file where a function is defined
function findFunctionFile(functionName, functionDataStructure) {
  for (const file in functionDataStructure) {
    const functions = functionDataStructure[file];
    if(!functions){
        let theExt = path.extname(file).slice(1);  
        if(DGRAPH.codeExtensions.includes(theExt)){
            console.log("File no function? " + file + " trying to find " + functionName);
        }
        continue;
    }
    for (const func of functions) {
      if (func.name === functionName) {
        return file;
      }
    }
  }
  return null;
}

/*
function groupFilesByDependency(dGraphArr, files, maxGroupSize = 5) {
  const groups = [];
  const ungroupedFiles = new Set(files);

  // Create a map to store the dependency count between files
  const dependencyCountMap = new Map();

  // Iterate over the dependencies and count the occurrences
  dGraphArr.forEach((dependency) => {
    const { caller, callee } = dependency;
    const key = `${caller}-${callee}`;

    if (!dependencyCountMap.has(key)) {
      dependencyCountMap.set(key, 0);
    }

    dependencyCountMap.set(key, dependencyCountMap.get(key) + 1);
  });

  while (ungroupedFiles.size > 0) {
    const currentGroup = new Set();

    // Find the file with the highest dependency count
    let maxDependencyCount = 0;
    let maxDependencyFile = null;

    for (const file of ungroupedFiles) {
      let fileDependencyCount = 0;

      for (const [key, count] of dependencyCountMap) {
        const [caller, callee] = key.split('-');

        if (caller === file || callee === file) {
          fileDependencyCount += count;
        }
      }

      if (fileDependencyCount > maxDependencyCount) {
        maxDependencyCount = fileDependencyCount;
        maxDependencyFile = file;
      }
    }

    if (maxDependencyFile) {
      currentGroup.add(maxDependencyFile);
      ungroupedFiles.delete(maxDependencyFile);

      // Find files with high dependency count with the current file
      const relatedFiles = new Set();
      for (const [key, count] of dependencyCountMap) {
        const [caller, callee] = key.split('-');

        if (caller === maxDependencyFile && ungroupedFiles.has(callee)) {
          relatedFiles.add(callee);
        } else if (callee === maxDependencyFile && ungroupedFiles.has(caller)) {
          relatedFiles.add(caller);
        }
      }

      for (const file of relatedFiles) {
        if (currentGroup.size < maxGroupSize) {
          currentGroup.add(file);
          ungroupedFiles.delete(file);
        } else {
          break;
        }
      }
    }

    // Ensure each group has at least one type of each file (.js, .html, .css) if available
    const fileTypes = ['.js', '.html', '.css'];
    for (const fileType of fileTypes) {
      if (currentGroup.size < maxGroupSize) {
        for (const file of ungroupedFiles) {
          if (file.endsWith(fileType)) {
            currentGroup.add(file);
            ungroupedFiles.delete(file);
            break;
          }
        }
      }
    }

    if (currentGroup.size > 0) {
      groups.push(Array.from(currentGroup));
    }

    if (ungroupedFiles.size > 0 && maxDependencyCount === 0) {
      // If there are no more dependencies, add the remaining files to separate groups
      const remainingFiles = Array.from(ungroupedFiles);
      const remainingGroups = [];

      while (remainingFiles.length > 0) {
        const group = new Set();

        // Ensure each remaining group has at least one type of each file (.js, .html, .css) if available
        for (const fileType of fileTypes) {
          for (let i = 0; i < remainingFiles.length; i++) {
            if (remainingFiles[i].endsWith(fileType)) {
              group.add(remainingFiles[i]);
              remainingFiles.splice(i, 1);
              break;
            }
          }
        }

        // Add remaining files to the group until maxGroupSize is reached
        while (group.size < maxGroupSize && remainingFiles.length > 0) {
          group.add(remainingFiles.shift());
        }

        remainingGroups.push(Array.from(group));
      }

      groups.push(...remainingGroups);
      ungroupedFiles.clear();
    }
  }

  return groups;
}

*/
DGRAPH.groupFilesByDependency = function(dGraphArr, files, maxGroupSize = 5) {
  const groups = [];
  const ungroupedFiles = new Set(files);

  // Create a map to store the dependency count between files
  const dependencyCountMap = new Map();

  // Iterate over the dependencies and count the occurrences
  dGraphArr.forEach((dependency) => {
    const { caller, callee } = dependency;
    const key = `${caller}-${callee}`;

    if (!dependencyCountMap.has(key)) {
      dependencyCountMap.set(key, 0);
    }

    dependencyCountMap.set(key, dependencyCountMap.get(key) + 1);
  });

  while (ungroupedFiles.size > 0) {
    const currentGroup = [];

    // Find the file with the highest dependency count
    let maxDependencyCount = 0;
    let maxDependencyFile = null;

    for (const file of ungroupedFiles) {
      let fileDependencyCount = 0;

      for (const [key, count] of dependencyCountMap) {
        const [caller, callee] = key.split('-');

        if (caller === file || callee === file) {
          fileDependencyCount += count;
        }
      }

      if (fileDependencyCount > maxDependencyCount) {
        maxDependencyCount = fileDependencyCount;
        maxDependencyFile = file;
      }
    }

    if (maxDependencyFile) {
      currentGroup.push(maxDependencyFile);
      ungroupedFiles.delete(maxDependencyFile);

      // Find files with high dependency count with the current file
      const relatedFiles = new Set();
      for (const [key, count] of dependencyCountMap) {
        const [caller, callee] = key.split('-');

        if (caller === maxDependencyFile && ungroupedFiles.has(callee)) {
          relatedFiles.add(callee);
        } else if (callee === maxDependencyFile && ungroupedFiles.has(caller)) {
          relatedFiles.add(caller);
        }
      }

      for (const file of relatedFiles) {
        if (currentGroup.length < maxGroupSize) {
          currentGroup.push(file);
          ungroupedFiles.delete(file);
        } else {
          break;
        }
      }
    }

    if (currentGroup.length > 0) {
      groups.push(currentGroup);
    }

    if (ungroupedFiles.size > 0 && maxDependencyCount === 0) {
      // If there are no more dependencies, add the remaining files to separate groups
      const remainingFiles = Array.from(ungroupedFiles);
      const remainingGroups = [];

      while (remainingFiles.length > 0) {
        const group = remainingFiles.splice(0, maxGroupSize);
        remainingGroups.push(group);
      }

      groups.push(...remainingGroups);
      ungroupedFiles.clear();
    }
  }

  return groups;
}


// Function to get the name of the function containing a call expression
function getFunctionName(callNode) {
  let currentNode = callNode;
  while (currentNode && currentNode.type !== 'function_declaration') {
    currentNode = currentNode.parent;
  }
  if (currentNode) {
    const functionNameNode = currentNode.childForFieldName('name');
    if (functionNameNode) {
      return functionNameNode.text;
    }
  }
  return 'toplevel';
}

// Function to print the dependency graph
// Function to print the dependency graph
function printDependencyGraph(dependencyGraphArray) {
  const fileMap = new Map();

  // Group dependencies by file
  dependencyGraphArray.forEach((dependency) => {
    const { caller, short_caller, callee, short_callee, line_caller, line_callee, type, name, parent_function } = dependency;
    if (!fileMap.has(caller)) {
      fileMap.set(caller, []);
    }

    fileMap.get(caller).push({ caller, short_caller, callee, short_callee, line_caller, line_callee, type, name, parent_function });
  });

  // Print the dependency graph
  for (const [file, dependencies] of fileMap) {
    console.log(`${file}:`);

    dependencies.forEach((dependency) => {
      const { caller, short_caller, callee, short_callee, line_caller, line_callee, type, name, parent_function } = dependency;
      console.log(`  ${name} (${type}) on line ${line_caller} => ${callee}:${line_callee}`);
    });

    console.log();
  }
}

DGRAPH.getDependencyGraphArray = function(dependencyGraphArray) {
  const result = [];

  const fileMap = new Map();
  const fileReverseMap = new Map();

  // Group dependencies by file
  dependencyGraphArray.forEach((dependency) => {
    const { caller, short_caller, callee, short_callee, line_caller, line_callee, type, name, parent_function } = dependency;

    if (!fileMap.has(caller)) {
      fileMap.set(caller, []);
    }
    
    fileMap.get(caller).push({ caller, short_caller, callee, short_callee, line_caller, line_callee, type, name, parent_function });

    // Build the reverse map
    if (!fileReverseMap.has(callee)) {
      fileReverseMap.set(callee, []);
    }
    
    fileReverseMap.get(callee).push({ caller, short_caller, callee, short_callee, line_caller, line_callee, type, name, parent_function });
  });

  // Build the result array
  for (const [file, dependencies] of fileMap) {
    const dependencyLines = dependencies.map((dependency) => {
      const { caller, short_caller, callee, short_callee, line_caller, line_callee, type, name, parent_function } = dependency;
      let extraFunction = "";
      if(parent_function != ''){
         extraFunction = " in " + parent_function;
      }      
      return `${short_caller} ${extraFunction} => ${name} (${line_caller}) in ${short_callee} (${line_callee})`;
    });

    const calledByLines = fileReverseMap.get(file) || [];
    const calledByLinesFormatted = calledByLines.map((dependency) => {
      const { caller, short_caller, callee, short_callee, line_caller, line_callee, type, name, parent_function } = dependency;
      let extraFunction = "";
      if(parent_function != ''){
         extraFunction = " in " + parent_function;
      }
      
      return ` ${short_caller} ${extraFunction} => ${name}`;
    });
    
    const mergedDependencyLines = [...dependencyLines, ...calledByLinesFormatted];

    result.push({
      filepath: file,
      mergedLines: mergedDependencyLines,
      dependencyLines: mergedDependencyLines,
      calledByLines: calledByLinesFormatted
    });
  }

  return result;
};

DGRAPH.getRelatedFunctionsForStubFunctions = async function(stubFunctionNames, projectName) {
  try {
    //console.log("Get related functions for stub functions");
    const dependencyGraph = await dexieDB.getDependencyGraph(projectName);
    
    let relatedFunctions = new Set();
    
    for (const dependency of dependencyGraph) {
      //console.log("Testing dependency");
      //console.log(dependency);
      if (stubFunctionNames.includes(dependency.name)) {          
          //console.log("FOUND DEPENDENCY");
        relatedFunctions.add({
          name: dependency.caller,
          filepath: dependency.caller_filepath,
          line: dependency.line_caller
        });
        relatedFunctions.add({
          name: dependency.callee, 
          filepath: dependency.callee_filepath,
          line: dependency.line_callee
        });
      }
    }
    
    return Array.from(relatedFunctions);
  } catch (err) {
    console.error('Error getting related functions:', err);
    return [];
  }
};

DGRAPH.getDependenciesForProject = async function(filepathArr, projectPath){
   console.log("project path: " + projectPath);
   console.log("analyzing files");
   console.log(filepathArr);
    
    
  // First pass: Build the function data structure
  let files = filepathArr;
  const dataStructures = await DGRAPH.buildFunctionDataStructure(files);
  let functionDataStructure = dataStructures[0];
  let functionSelectors = dataStructures[1];
  let selectorDataStructure = {};
  
    //console.log("function data structure");
    //console.log(functionDataStructure);  
  
  const functionCallsDataStructure = {};
  async function processFiles(files) {
      const promises = files.map(async (file) => {
        const fileExtension = path.extname(file);

        if (fileExtension === '.js') {
          const functionCalls = await DGRAPH.extractFunctionCallsJavascript(file);
          functionCallsDataStructure[file] = functionCalls;
          //console.log(functionCalls);
        } else if(fileExtension === '.php'){
          const functionCalls = await DGRAPH.extractFunctionCallsPHP(file);
          functionCallsDataStructure[file] = functionCalls;
          //console.log(functionCalls);            
        }
        else if (fileExtension === '.html') {
          const selectors = await DGRAPH.extractSelectorsFromHtml(file);
          selectorDataStructure[file] = selectors;
        } else if (fileExtension === '.css') {
          const selectors = await DGRAPH.extractSelectorsFromCss(file);
          selectorDataStructure[file] = selectors;
        }
      });

      await Promise.all(promises);
  }  
  await processFiles(files);
  
  //console.log("function data structure");
  //console.log(functionDataStructure);

  // Second pass: Build the dependency graph
    //const dependencyGraph = DGRAPH.buildDependencyGraph(files, functionDataStructure, functionCallsDataStructure, selectorDataStructure);

    let dGraphArr = DGRAPH.buildDependencyGraphArray(filepathArr, functionDataStructure, functionCallsDataStructure, functionSelectors, selectorDataStructure, projectPath);
    
    //console.log("d graph array");
    //console.log(dGraphArr);
    
    //console.log("total dependencies");
    //console.log(dGraphArr.length);
    
    // Print the dependency graph
    //printDependencyGraph(dGraphArr);
    
    let dependencyGroups = DGRAPH.groupFilesByDependency(dGraphArr, filepathArr);
    //console.log("Dependency Groups:");
    for (let d = 0; d < dependencyGroups.length; d++) {
      //console.log(`Group ${d + 1}:`);
      //console.log(dependencyGroups[d]);
      //console.log("===========================");
    }    
    return [dGraphArr, dependencyGroups];
};


DGRAPH.isFunctionDefinition = function(filename, line) {
  // Remove leading and trailing whitespace from the line
  line = line.trim();
  const extension = filename.split('.').pop().toLowerCase();

  // Check if the file extension is supported
  if (extension === 'js' || extension === 'ts') {
    if (line.endsWith(';') && !line.includes('}')) {
      return false;
    }
    if (line.includes('function')) {
      return true;
    }
  } else if (extension === 'py') {
    // Python file
    if (line.startsWith('def ')) {
      return true;
    }
  } else if (extension === 'java') {
    // Java file
    if (line.includes('public') || line.includes('private') || line.includes('protected')) {
      if (line.includes('(') && line.includes(')')) {
        return true;
      }
    }
  } else if (extension === 'dart') {
    // Dart file
    if (line.startsWith('void ') || 
        (line.includes('(') && line.includes(')') && !line.includes(';'))) {
      return true;
    }
  }

  // If none of the above conditions are met, assume it's not a function definition
  return false;
};

DGRAPH.getControlStructureForLine = async function(filePath, searchLine, fileContent, bestMatch = null, preParsedTree = null) {
    console.log("DGRAPH.getControlStructureForLine control structure for " + searchLine);
    console.log("in " + filePath);
    const lines = fileContent.split(/\r?\n/);
    let firstOccurrence = lines.findIndex(line => line.includes(searchLine));  

    if (firstOccurrence < 0) {
        if (bestMatch) {
            let bestLine = bestMatch['line'];
            firstOccurrence = lines.findIndex(line => line.includes(bestLine));  
            //console.log("bestline " );
            //console.log(firstOccurrence);
            
            if (firstOccurrence < 0) {
                //console.log("failed to find line that should exist: ", bestLine);
                return null;            
            }
        } else {
            console.log("failed to find line with no bestMatch: ");
            console.log(searchLine);
            return null;
        }
    }
    let startLine = firstOccurrence;

    const tree = preParsedTree || await DGRAPH.parseAnyFile(filePath, fileContent);

    if (!tree || !tree.rootNode) {
        console.error('Failed to parse file content: ' + filePath);
        return null;
    }
    //console.log("DGRAPH.getControlStructureForLine line index: " + firstOccurrence);
    
    function isControlStructure(node) {
        return ['function', 'function_declaration', 'function_definition', 'method_definition', 'method_declaration', 'async_function_definition', 'if_statement', 'for_statement', 'while_statement', 'do_statement', 'switch_statement', 'try_statement', 'catch_clause', 'call_expression', 'expression_statement'].includes(node.type);
    }    
    function findLargestControlStructureForLine(rootNode, targetLine) {
        let bestScope = null;
        const MAX_DEPTH = 100;

        function traverseEverything(node, depth) {
            if (depth > MAX_DEPTH) {
                //console.log(`Max depth reached at node: ${node.id}`);
                return;
            }
            
            //console.log(`DGRAPH.getControlStructureForLine Checking node: ${node.id} (${node.type}) at line ${node.startPosition.row}-${node.endPosition.row}, depth: ${depth}`);

            if (node.startPosition.row <= targetLine && node.endPosition.row >= targetLine) {
                if (isControlStructure(node)) {
                    if (!bestScope || 
                        (node.endPosition.row - node.startPosition.row) < 
                        (bestScope.endPosition.row - bestScope.startPosition.row)) {
                        bestScope = node;
                    }
                }

                for (const child of node.namedChildren) {
                    traverseEverything(child, depth + 1);
                }
            }
        }

        traverseEverything(rootNode, 0);
        return bestScope;
    }    
    
    
    function isObjectLiteralStart(line) {
        return line.trim().endsWith('{') && !line.trim().startsWith('{');
    }

    function findLineObjectLiteralEnd(lines, startIndex) {
        let braceCount = 1;
        for (let i = startIndex + 1; i < lines.length; i++) {
            braceCount += (lines[i].match(/{/g) || []).length;
            braceCount -= (lines[i].match(/}/g) || []).length;
            if (braceCount === 0) {
                return i + 1;
            }
        }
        return startIndex;
    }    
    
    

    function getStructureInfo(node) {
        let structureName = 'anonymous';
        let type = node.type;
        let startLine = node.startPosition.row;
        let endLine = node.endPosition.row;

        //console.log("getting structure info");
        //console.log("type " + type);
        //console.log("startLine " + startLine);
        //console.log("endLine " + endLine);
        
        // Special handling for object literals
        if (type === 'call_expression' && isObjectLiteralStart(lines[startLine])) {
            //console.log("trying to find object literal end");
            endLine = findLineObjectLiteralEnd(lines, startLine);
            type = 'object_literal';
        }        

        if (node.childForFieldName('name')) {
            structureName = node.childForFieldName('name').text;
        } else if (node.childForFieldName('declarator')) {
            const declarator = node.childForFieldName('declarator');
            if (declarator.childForFieldName('name')) {
                structureName = declarator.childForFieldName('name').text;
            }
        }

        return {
            name: structureName,
            startLine: startLine,
            endLine: endLine,
            type: type
        };
    }

    const relevantNode = findLargestControlStructureForLine(tree.rootNode, startLine);

    if (relevantNode) {
        //console.log("getting structure info for");
        //console.log(relevantNode);
        return getStructureInfo(relevantNode);
    }
    
    
    console.log("doing fallback for control structure1111");

    // Fallback method if Tree-sitter fails to find a suitable structure
    return DGRAPH.fallbackGetControlStructure(lines, startLine);
};

DGRAPH.getLargestControlStructureForLine = async function(filePath, searchLine, fileContent, bestMatch = null, preParsedTree = null) {
    console.log("DGRAPH.getLargestControlStructureForLine control structure for " + searchLine);
    console.log("in " + filePath);
    const lines = fileContent.split(/\r?\n/);
    let firstOccurrence = lines.findIndex(line => line.includes(searchLine));  

    if (firstOccurrence < 0) {
        if (bestMatch) {
            let bestLine = bestMatch['line'];
            firstOccurrence = lines.findIndex(line => line.includes(bestLine));  
            console.log("bestline " );
            console.log(firstOccurrence);
            
            if (firstOccurrence < 0) {
                //console.log("failed to find line that should exist: ", bestLine);
                return null;            
            }
        } else {
            console.log("failed to find line with no bestMatch: ");
            console.log(searchLine);
            return null;
        }
    }
    let startLine = firstOccurrence;

    const tree = preParsedTree || await DGRAPH.parseAnyFile(filePath, fileContent);

    if (!tree || !tree.rootNode) {
        console.error('Failed to parse file content');
        return null;
    }
    //console.log("line index: " + firstOccurrence);
    
    function isControlStructure(node) {
        return ['class_declaration', 'class_definition', 'function', 'function_declaration', 'function_definition', 'method_definition', 'method_declaration', 'async_function_definition', 'decorator', 'if_statement', 'for_statement', 'while_statement', 'do_statement', 'switch_statement', 'try_statement', 'catch_clause', 'call_expression', 'expression_statement'].includes(node.type);
    }    

    function isLargerStructure(node) {
        return ['class_declaration', 'class_definition', 'function', 'function_declaration', 'function_definition', 'method_definition', 'method_declaration', 'async_function_definition', 'decorator', 'if_statement', 'for_statement', 'while_statement', 'do_statement', 'switch_statement', 'try_statement', 'catch_clause'].includes(node.type);
    }

    function findControlStructureForLine(rootNode, targetLine, originalStartLine) {
        //console.log("target line is " + targetLine);
        let bestScope = null;
        let largerScope = null;
        const MAX_DEPTH = 100;

        function traverseEverything(node, depth) {
            if (depth > MAX_DEPTH) {
                console.log(`Max depth reached at node: ${node.id}`);
                return;
            }
            
            console.log(`DGRAPH.getLargestControlStructureForLine Checking node: ${node.id} (${node.type}) at line ${node.startPosition.row}-${node.endPosition.row}, depth: ${depth}`);

            if (node.startPosition.row <= targetLine && node.endPosition.row >= targetLine) {
                if (isControlStructure(node) && node.startPosition.row === originalStartLine) {
                    // If the node starts at the original search line, consider all control structures
                    if (!bestScope || 
                        (node.endPosition.row - node.startPosition.row) > 
                        (bestScope.endPosition.row - bestScope.startPosition.row)) {
                        console.log("set largest line match");
                        console.log(node);
                        bestScope = node;
                    }
                } 
                else if(node.type == 'array' && node.startPosition.row === originalStartLine && node.endPosition.row > targetLine){
                    if (!bestScope || 
                        (node.endPosition.row - node.startPosition.row) > 
                        (bestScope.endPosition.row - bestScope.startPosition.row)) {
                        console.log("set array as largest line match");
                        console.log(node);
                        bestScope = node;
                    }                    
                }
                else if (isLargerStructure(node)) {
                    if (!largerScope || 
                        (node.endPosition.row - node.startPosition.row) < 
                        (largerScope.endPosition.row - largerScope.startPosition.row)) {
                        console.log("used larger scope");
                        console.log(node);
                        largerScope = node;
                    }
                }

                for (const child of node.namedChildren) {
                    traverseEverything(child, depth + 1);
                }
            }
        }

        traverseEverything(rootNode, 0);
        
        if(!bestScope){
            bestScope = largerScope;
        }
        else if((bestScope.endPosition.row - bestScope.startPosition.row) <= 1){
            if(bestScope.type === 'call_expression' && isObjectLiteralStart(lines[bestScope.startPosition.row])){
                //console.log("multiline call expression");
            }
            else if(isArrayLiteralStart(lines[bestScope.startPosition.row])){
                //console.log("arrayliteralstart call expression with type " + bestScope.type);
            }else{
                //console.log("scope for line was just 1 line or less, so using larger");
                console.log(largerScope);
                bestScope = largerScope;
            }
        }
        
        return bestScope;
    }    
    
    function isObjectLiteralStart(line) {
        return line.trim().endsWith('{') && !line.trim().startsWith('{');
    }

    function findLineObjectLiteralEnd(lines, startIndex) {
        let braceCount = 1;
        for (let i = startIndex + 1; i < lines.length; i++) {
            braceCount += (lines[i].match(/{/g) || []).length;
            braceCount -= (lines[i].match(/}/g) || []).length;
            if (braceCount === 0) {
                return i;
            }
        }
        return startIndex;
    }    

    function isArrayLiteralStart(line) {
        return line.trim().endsWith('[') && !line.trim().startsWith('[');
    }

    const findLineArrayLiteralEnd = (codeblockLines, startLine) => {
      let bracketCount = 1;
      for (let i = startLine + 1; i < codeblockLines.length; i++) {
        bracketCount += (codeblockLines[i].match(/\[/g) || []).length;
        bracketCount -= (codeblockLines[i].match(/\]/g) || []).length;
        if (bracketCount === 0) {
          return i;
        }
      }
      return startLine;
    };
    

    function getStructureInfo(node, originalStartLine) {
        let structureName = 'anonymous';
        let type = node.type;
        let startLine = node.startPosition.row;
        let endLine = node.endPosition.row;

         console.log("getting structure info");
        console.log("type " + type);
        console.log("startLine " + startLine);
        console.log("endLine " + endLine);
        console.log(lines[startLine]);
        
        // Special handling for object literals
        if (type === 'call_expression' && isObjectLiteralStart(lines[startLine])) {
            //console.log("trying to find object literal end");
            endLine = findLineObjectLiteralEnd(lines, startLine);
            type = 'object_literal';
        }      
        
        // Special handling for array literals
        if (isArrayLiteralStart(lines[startLine])) {
            console.log("IS LITERAL START: " + lines[startLine]);
            console.log("type for literal start: " + type);
            endLine = findLineArrayLiteralEnd(lines, startLine);
            type = 'array_literal';
        }        

        if (node.childForFieldName('name')) {
            structureName = node.childForFieldName('name').text;
        } else if (node.childForFieldName('declarator')) {
            const declarator = node.childForFieldName('declarator');
            if (declarator.childForFieldName('name')) {
                structureName = declarator.childForFieldName('name').text;
            }
        }

        return {
            name: structureName,
            startLine: startLine,
            endLine: endLine,
            type: type
        };
    }

    const relevantNode = findControlStructureForLine(tree.rootNode, startLine, startLine);

    if (relevantNode) {
        //console.log("getting structure info for");
        //console.log(relevantNode);
        return getStructureInfo(relevantNode, startLine);
    }
    
    //console.log("doing fallback for control structure2222");
    
    //its just one line and that's ok
    if(searchLine.indexOf('{') < 0){        
        console.log("just one line");
        return {
            name: 'single_line',
            startLine: startLine,
            endLine: startLine,
            type: 'statement'
        };
    }
    
    
    console.log("using fallback");
    // Fallback method if Tree-sitter fails to find a suitable structure
    return DGRAPH.fallbackGetControlStructure(lines, startLine);
};

DGRAPH.fallbackGetControlStructure = function(lines, startLine) {
  let braceCount = 0;
  let inString = false;
  let stringDelimiter = '';
  let structureName = lines[startLine].trim().split(/[\s({=]/)[0];
  
  for (let i = startLine; i < lines.length; i++) {
    const line = lines[i];
    for (let j = 0; j < line.length; j++) {
      const char = line[j];
      if (!inString) {
        if (char === '{') braceCount++;
        else if (char === '}') {
          braceCount--;
          if (braceCount === 0) {
            let retStructure = {
              name: structureName,
              startLine: startLine,
              endLine: i + 1,
              type: 'block'
            };
            //console.log("fallback result");
            //console.log(retStructure);
            return retStructure;
          }
        } else if (char === "'" || char === '"' || char === '`') {
          inString = true;
          stringDelimiter = char;
        }
      } else {
        if (char === stringDelimiter && line[j-1] !== '\\') {
          inString = false;
        }
      }
    }
  }
  console.log("no fallback structure found");
  return null;
};


DGRAPH.getTopLevelScopes = async function(codeblockData, filename, originalFileContent) {
  let codeblockLines = codeblockData.split(/\r?\n/);

  let originalFilename = filename;
  let theExt = path.extname(filename).slice(1);  
  //parser reads from cache filename, make unique filename to not use cache
  filename = "not_a_real_file" + randomLetters(5) + "." + theExt; 
  console.log("DGRAPH.getTopLevelScopes filename " + originalFilename);
  let isPHP = false;
  let addedPHPTag = false;
  let skipLineIndex = -1;
  let skipFirstLine = false;
  if(filename.indexOf('.php') >= 0){
      //do not clean lines
      isPHP = true;
      if (!codeblockLines[0].trim().startsWith('<?php')) {
        codeblockLines[0] = '<?php ' + codeblockLines[0];
        codeblockData = "<?php " + codeblockData;
        addedPHPTag = true;
        //console.log("Added <?php tag to the first line");
      }      
      skipLineIndex = codeblockLines.findIndex(line => line.trim().startsWith('<?php'));
  } 
  let language = extToLang[theExt];
  if(language){
    let codeblockLangSpecifierPattern = new RegExp('^```\\s*' + language + '\\s*$');
    if (codeblockLines[0].trim().match(codeblockLangSpecifierPattern) || codeblockLines[0].trim() === language) {
        skipFirstLine = true;
    }
  }
  
  //console.log("cleaned");
  //console.log(codeblockLines);
  const scopes = [];
  const processedLines = new Set();
  let requestedLines = [];

  const isBreakingComment = (line) => {
    const trimmedLine = line.trim();
    return (trimmedLine.startsWith('//') || trimmedLine.startsWith('#')) && line.includes('...');
  };
  const isExistingCodeComment = (line) => {
      return line.trim() === '//... existing code';
  };  
  
  // New function to check if a line starts a multi-line comment
  const isStartOfMultilineComment = function(line) {
    const trimmedLine = line.trim();
    return trimmedLine.startsWith('/*') && !trimmedLine.endsWith('*/');
  };

  // New function to check if a line ends a multi-line comment
  const isEndOfMultilineComment = function(line) {
    const trimmedLine = line.trim();
    return trimmedLine.endsWith('*/') && !trimmedLine.startsWith('/*');
  };

  // New function to find the end of a multi-line comment
  const findEndOfMultilineComment = function(startLine) {
    for (let i = startLine; i < codeblockLines.length; i++) {
      if (codeblockLines[i].trim().endsWith('*/')) {
        return i;
      }
    }
    // If no end found, return the last line
    return codeblockLines.length - 1;
  };  
  
  const findFirstNonEmptyLine = (startIndex) => {
    for (let i = startIndex; i < codeblockLines.length; i++) {
      if (codeblockLines[i].trim() !== '') {
        return i;
      }
    }
    return startIndex;
  };
  const isObjectLiteralStart = (line) => {
    return line.trim().endsWith('{') && !line.trim().startsWith('{');
  };

  const findObjectLiteralEnd = (startLine) => {
    let bracketCount = 1;
    for (let i = startLine + 1; i < codeblockLines.length; i++) {
      bracketCount += (codeblockLines[i].match(/{/g) || []).length;
      bracketCount -= (codeblockLines[i].match(/}/g) || []).length;
      if (bracketCount === 0) {
        return i + 1;
      }
    }
    return startLine;
  };  
  
  
    function isArrayLiteralStart(line) {
        return line.trim().endsWith('[') && !line.trim().startsWith('[');
    }

    const findLineArrayLiteralEnd = (startLine) => {
      let bracketCount = 1;
      for (let i = startLine + 1; i < codeblockLines.length; i++) {
        bracketCount += (codeblockLines[i].match(/\[/g) || []).length;
        bracketCount -= (codeblockLines[i].match(/\]/g) || []).length;
        if (bracketCount === 0) {
          return i;
        }
      }
      return startLine;
    };

    
  const isCommentLine = function(line) {
    const trimmedLine = line.trim();
    return trimmedLine.startsWith('//') || trimmedLine.startsWith('#') || (trimmedLine.startsWith('/*') && trimmedLine.endsWith('*/'));
  };
  
  
  // Try to parse the file, but don't return if it fails
  await DGRAPH.initParser();
  let tree = null;
  try {
    tree = await DGRAPH.parseAnyFile(filename, codeblockData);
  } catch (error) {
    console.error('Failed to parse file:', filename, error);
    // We'll continue with tree as null, which will trigger the fallback method
  }

  let currentLine = 0;
  let skippedLargeFunction = false;
  while (currentLine < codeblockLines.length) {
    if (processedLines.has(currentLine)) {
        currentLine++;
        continue;
    }      
    if(currentLine < skipLineIndex){
        //console.log("skipping line " + codeblockLines[currentLine].trim() + " before <?php");
    }
    if(currentLine <= 1) {
      //the LLM is giving a filename at the start of the code.  So if it's our filename we'll assume it's just putting the filename in code accidentally
      //console.log("comparing " + codeblockLines[currentLine].trim() + " vs " + originalFilename);
      if(comparePaths(codeblockLines[currentLine].trim(), originalFilename)){
        currentLine++;
        continue;
      }     
    }
    // Skip the code block language specifier line if applicable
    if (currentLine === 0 && skipFirstLine) {
        currentLine++;
        continue;
    }     
    if (isBreakingComment(codeblockLines[currentLine])) {
      const allLines = codeblockLines[currentLine].trim();
      if (allLines !== '') {
        processedLines.add(currentLine);
        /*
            console.log("adding scope111");
            console.log({
          line_start: currentLine + 1,
          line_end: currentLine + 1,
          start_line: allLines,
          all_lines: allLines,
          category: 'comment'
        });
               
         */        
        scopes.push({
          line_start: currentLine + 1,
          line_end: currentLine + 1,
          start_line: allLines,
          all_lines: allLines,
          category: 'comment'
        });
      }
      currentLine++;      
    }          
    else if(isCommentLine(codeblockLines[currentLine])){
      const allLines = codeblockLines[currentLine].trim();
      if (allLines !== '') {
        processedLines.add(currentLine);
        scopes.push({
          line_start: currentLine + 1,
          line_end: currentLine + 1,
          start_line: allLines,
          all_lines: allLines,
          category: 'comment'
        });
      }
      currentLine++;        
    }
    else if (isStartOfMultilineComment(codeblockLines[currentLine])) {
      const startLine = currentLine;
      const endLine = findEndOfMultilineComment(currentLine);
      const allLines = codeblockLines.slice(startLine, endLine + 1).join('\n').trim();
      if (allLines !== '') {
        scopes.push({
          line_start: startLine + 1,
          line_end: endLine + 1,
          start_line: codeblockLines[startLine].trim(),
          all_lines: allLines,
          category: 'comment'
        });
      }
      for (let i = startLine; i <= endLine; i++) {
        processedLines.add(i);
      }
      currentLine = endLine + 1;
    }    
    else {
      if(requestedLines.includes(currentLine)){
          currentLine++;
          continue;
      }
      //console.log("DGRAPH.getTopLevelScopes Getting control structure for line:");  
      //console.log(codeblockLines[currentLine]);
      
      if(codeblockLines[currentLine].trim() == ''){
          //console.log("skipping empty line");
          currentLine++;
          continue;          
      }
      else if(isPHP){
          if(codeblockLines[currentLine].trim() == '<?php' || codeblockLines[currentLine].trim() == '?>'){
                //console.log("skipping php tag line");
                currentLine++;
                continue;                 
          }
      }
      
      
      let structure;
      if (tree) {
        structure = await DGRAPH.getLargestControlStructureForLine(filename, codeblockLines[currentLine], codeblockData, null, tree);
        requestedLines.push(currentLine);        
      }else{
          //console.log("no tree");
      }
      
      if (!structure) {
        //console.log("no structure");
        // Use fallback method if Tree-sitter parsing failed or didn't find a structure
        structure = DGRAPH.fallbackGetControlStructure(codeblockLines, currentLine);
      }
        // Remove the added <?php tag if it was added and it's the first line
      if (addedPHPTag) {
          codeblockLines[0] = codeblockLines[0].replace(/^<\?php\s/, '');          
      }      
      
      //console.log("structure");
      //console.log(structure);
      //console.log("startLine must match");
      //console.log(currentLine);
      
      
      
      if (structure && (structure.startLine == currentLine || structure.startLine == (currentLine + 1))) {
        //console.log("It does match");
        
        
        if(structure.startLine == structure.endLine && isObjectLiteralStart(codeblockLines[currentLine])) {
             //console.log("object literal match");
              const startLine = currentLine;
              const endLine = findObjectLiteralEnd(startLine);
              const allLines = codeblockLines.slice(startLine, endLine + 1).join('\n').trim();
              if (allLines !== '') {
                //console.log("adding scope444");
                scopes.push({
                  line_start: startLine + 1,
                  line_end: endLine + 1,
                  start_line: codeblockLines[startLine].trim(),
                  all_lines: allLines,
                  category: 'object_literal'
                });
              }
              for (let i = startLine; i <= endLine; i++) {
                processedLines.add(i);
              }
              currentLine = endLine + 1;
        }else if(structure.startLine == structure.endLine && isArrayLiteralStart(codeblockLines[currentLine])) {
             console.log("array literal match");
              const startLine = currentLine;
              const endLine = findLineArrayLiteralEnd(startLine);
              const allLines = codeblockLines.slice(startLine, endLine + 1).join('\n').trim();
              if (allLines !== '') {
                console.log("adding scope777");
                scopes.push({
                  line_start: startLine + 1,
                  line_end: endLine + 1,
                  start_line: codeblockLines[startLine].trim(),
                  all_lines: allLines,
                  category: 'array_literal'
                });
              }
              for (let i = startLine; i <= endLine; i++) {
                processedLines.add(i);
              }
              currentLine = endLine + 1;
        }else{                       
            const firstNonEmptyLine = findFirstNonEmptyLine(structure.startLine);
            if(currentLine <= 1) {
                //structure.startLine
                //originalFileContent
                //codeblockLines[firstNonEmptyLine].trim();

                //find first non empty line in the originalFileContent.  If exact match see how big of a block it is.  
                //if the block is over 300 lines then skip this line (and end line for languages that have a closing brace })
                //that way only inside of block will be processed and create separate scopes - the container that was skipped will be used as reference when asking
                //for full code changes but it doesn't need to be part of scope

                // New code to check for large functions
                const currentLineContent = codeblockLines[currentLine].trim();
                const originalLines = originalFileContent.split(/\r?\n/);
                const originalLineIndex = originalLines.findIndex(line => line.trim() === currentLineContent);

                if (originalLineIndex !== -1) {
                    //console.log("Found matching line in original file at index:", originalLineIndex);

                    // Use DGRAPH.getLargestControlStructureForLine to get the structure in the original file
                    const originalStructure = await DGRAPH.getLargestControlStructureForLine(
                      originalFilename,
                      currentLineContent,
                      originalFileContent
                    );

                    if (originalStructure && originalStructure.endLine - originalStructure.startLine > 300) {
                        // Determine the language based on the file extension
                        //console.log("Skipping large function (over 300 lines) starting at line", currentLineContent);
                        // Languages with explicit end brackets/braces
                        const bracketExtensions = ['js', 'c', 'cs', 'cpp', 'java', 'php', 'rs', 'go'];

                        // Skip the starting line of the large function
                        processedLines.add(currentLine);
                        currentLine++;                        
                        const fileExt = path.extname(filename).slice(1);
                        if (bracketExtensions.includes(fileExt)) {                        
                            processedLines.add(structure.startLine);
                            let endLineSkip = codeblockLines[structure.endLine].trim();
                            //console.log("skipping the end line " + endLineSkip);
                            processedLines.add(structure.endLine);
                        }
                        skippedLargeFunction = true;
                        continue;
                    }
                }
            }            
            
            const allLines = codeblockLines.slice(structure.startLine, structure.endLine + 1).join('\n').trim();
            if (allLines !== '') {
                
                //if(isCommentLine(codeblockLines[currentLine])){          
                //}
                //console.log("adding scope222");      
                /*
                console.log({
                line_start: structure.startLine + 1,
                line_end: structure.endLine + 1,
                start_line: codeblockLines[firstNonEmptyLine].trim(),
                all_lines: allLines,
                category: structure.type
              });*/
              scopes.push({
                line_start: structure.startLine + 1,
                line_end: structure.endLine + 1,
                start_line: codeblockLines[firstNonEmptyLine].trim(),
                all_lines: allLines,
                category: structure.type
              });
            }
            for (let i = structure.startLine; i <= structure.endLine; i++) {
                processedLines.add(i);
            }        
            currentLine = Math.max(currentLine + 1, structure.endLine + 1);
            //console.log("currentline now at " + currentLine);
        }
      } else {
        //console.log("group consecutive single lines that are not part of control structure");
        // Group consecutive single lines that are not part of any control structure
        const startLine = currentLine;
        let firstNonEmptyLine = startLine;

        while (currentLine < codeblockLines.length && !codeblockLines[currentLine].trim().endsWith('{')) {
          if (codeblockLines[firstNonEmptyLine].trim() === '' && currentLine > firstNonEmptyLine) {
            firstNonEmptyLine = currentLine;
          }
          currentLine++;
        }

        if (startLine < currentLine) {
          const allLines = codeblockLines.slice(startLine, currentLine).join('\n').trim();
          if (allLines !== '') {
              /*
            console.log("adding scope333");
            console.log({
              line_start: startLine + 1,
              line_end: currentLine,
              start_line: codeblockLines[firstNonEmptyLine].trim(),
              all_lines: allLines,
              category: 'lines'
            });
            */
            scopes.push({
              line_start: startLine + 1,
              line_end: currentLine,
              start_line: codeblockLines[firstNonEmptyLine].trim(),
              all_lines: allLines,
              category: 'lines'
            });
            // Mark these lines as processed
            for (let i = startLine; i < currentLine; i++) {
              processedLines.add(i);
            }            
          }
        }
      }
    }
  }

  return scopes;
};


if(!randomLetters){
    var randomLetters = function(length){
        var letters = "abcdefghijklmnopqrstovwxyzABCDEFGHIJKLMNOPQRSTUVWYXZ1234567890";
        var len = letters.length - 1;
        var retString = "";
        for(var x =0; x < length; x++){
            var rand = Math.floor(Math.random() * len);
            retString = retString + letters.substr(rand, 1);
        }
        return retString;
    };   
}
if(!comparePaths){
    function comparePaths(path1, path2) {
      const normalizedPath1 = path.normalize(path1).replace(/\\/g, '/');
      const normalizedPath2 = path.normalize(path2).replace(/\\/g, '/');

      const pathPieces1 = normalizedPath1.split('/').filter(piece => piece !== '');
      const pathPieces2 = normalizedPath2.split('/').filter(piece => piece !== '');

      if (pathPieces1.length !== pathPieces2.length) {
        return false;
      }

      for (let i = 0; i < pathPieces1.length; i++) {
        if (pathPieces1[i] !== pathPieces2[i]) {
          return false;
        }
      }

      return true;
    }   
}

if (!process.type) {
    module.exports = DGRAPH;
}
//testDependencies();