let watcher;
const langToExt = {
  'javascript': 'js',
  'bash': 'sh',
  'c': 'c',
  'c_sharp': 'cs',
  'cpp': 'cpp',
  'css': 'css',
  'elm': 'elm',
  'embedded_template': 'html',
  'go': 'go',
  'html': 'html',
  'java': 'java',
  'lua': 'lua',
  'markdown': 'md',
  'php': 'php',
  'python': 'py',
  'smarty': 'tpl',
  'ruby': 'rb',
  'rust': 'rs',
  'systemrdl': 'sv',
  'toml': 'toml',
  'vue': 'vue',
  'yaml': 'yaml',
  'dart': 'dart',
  'json': 'json',
  'typescript': 'ts',
  'tsx': 'tsx'
};

const extToLang = {
  'js': 'javascript',
  'jsx' : 'javascript',
  'sh': 'bash',
  'c': 'c',
  'cs': 'c_sharp',
  'cpp': 'cpp',
  'css': 'css',
  'elm': 'elm',
  'html': 'embedded_template',
  'tpl': 'smarty',
  'go': 'go',
  'java': 'java',
  'lua': 'lua',
  'md': 'markdown',
  'php': 'php',
  'py': 'python',
  'rb': 'ruby',
  'rs': 'rust',
  'sv': 'systemrdl',
  'toml': 'toml',
  'vue': 'vue',
  'yaml': 'yaml',
  'dart': 'dart',  
  'json': 'json',
  'ts': 'typescript',
  'tsx': 'tsx'  
};
function initializeWatcher() {
  watcher = chokidar.watch([], {
    persistent: true,
    ignoreInitial: true
  });

  watcher
    .on('change', (filePath) => {
      console.log(`File ${filePath} has been changed`);
      STUBS.FILE_CACHE.delete(filePath);
    })
    .on('unlink', (filePath) => {
      console.log(`File ${filePath} has been removed`);
      STUBS.FILE_CACHE.delete(filePath);
    });

  // Listen for window unload event
  window.addEventListener('unload', () => {
    if (watcher) {
      watcher.close();
    }
  });
}

// Initialize the watcher when your app starts
initializeWatcher();


// Modified readFileWithCache function
function readFileWithCache(filePath) {
  if (STUBS.platform === 'darwin' && !filePath.startsWith('/')) {
    filePath = '/' + filePath;
    //console.log("new path " + filePath);
  }
  
  //console.log("chokidar " + filePath);
  //console.log("in");
  //console.log(STUBS.FILE_CACHE);

  if (STUBS.FILE_CACHE.has(filePath)) {
    return STUBS.FILE_CACHE.get(filePath);
  }  
  //console.log('it is not');

  let content = fs.readFileSync(filePath, 'utf-8');

  // Normalize line endings to Unix style (\n)
  content = content.replace(/\r\n/g, '\n');

  STUBS.FILE_CACHE.set(filePath, content);

  // Add the file to the watcher
  watcher.add(filePath);

  return content;
}


// You might also want to add a function to manually close the watcher
function closeWatcher() {
  if (watcher) {
    watcher.close();
    watcher = null;
  }
}

function printCacheKeys() {
  console.log("Current files in cache:");
  if (STUBS.FILE_CACHE.size === 0) {
    console.log("  (Cache is empty)");
  } else {
    STUBS.FILE_CACHE.forEach((value, key) => {
      console.log(`  - ${key}`);
    });
  }
  console.log(`Total files in cache: ${STUBS.FILE_CACHE.size}`);
}

function invalidateFileCache(filePath = null) {
  if (filePath === null) {
    STUBS.FILE_CACHE.clear();
    console.log('Entire file cache cleared');
    watcher.unwatch('*');
  } else {
    if (STUBS.FILE_CACHE.has(filePath)) {
      STUBS.FILE_CACHE.delete(filePath);
      watcher.unwatch(filePath);
      console.log(`File ${filePath} removed from cache`);
    } else {
      console.log(`File ${filePath} not found in cache`);
    }
  }
}