const usermysql = require('mysql');
const userutil = require('util');
let pool = null;
//user mysql
var UMYSQL = UMYSQL || {};  
UMYSQL.databaseConnected = false;

UMYSQL.initDatabase = async function(mysqlConfig){
    console.log("mysql config");
    console.log(mysqlConfig);
    if(pool == null){
        console.log("doing init");
        pool = usermysql.createPool({
          connectionLimit: 10,
          host: mysqlConfig['host'],
          user: mysqlConfig['user'],
          password: mysqlConfig['password'],
          database: mysqlConfig['database']
        });

        pool.query = userutil.promisify(pool.query);
    }
    console.log("done init");
};

UMYSQL.testConnection = async function(mysqlConfig) {
    try {
        let testPool = usermysql.createPool({
            connectionLimit: 1,
            host: mysqlConfig['host'],
            port: mysqlConfig['port'],
            user: mysqlConfig['user'],
            password: mysqlConfig['password'],
            database: mysqlConfig['database']
        });

        testPool.query = userutil.promisify(testPool.query);

        await testPool.query('SELECT 1');
        testPool.end();

        return true;
    } catch (error) {
        console.error('Error testing database connection:', error);
        return false;
    }
};


UMYSQL.getSchemas = async function() {
    try {
        const result = await pool.query('SHOW SCHEMAS');
        return result.map(row => row.Database);
    } catch (error) {
        console.error('Error retrieving schemas:', error);
        return [];
    }
};

UMYSQL.getTables = async function(schema) {
    console.log("getting tables for: " + schema);
    try {        
        await pool.query('USE ??', [schema]);
        const result = await pool.query('SHOW TABLES');
        console.log('Raw result:', result);
        const tables = result.map(row => Object.values(row)[0]);
        console.log('Parsed tables:', tables);
        return tables;
    } catch (error) {
        console.error(`Error retrieving tables for schema ${schema}:`, error);
        return [];
    }
};

UMYSQL.getTableStructure = async function(schema) {
    try {
        await pool.query(`USE \`${schema}\``);
        const tables = await UMYSQL.getTables(schema);
        let structure = '';
        for (const table of tables) {
            console.log("has table: " + table);
            structure += `${table}\n`;
            const columns = await pool.query(`DESCRIBE \`${table}\``);
            for (const column of columns) {
                structure += `- ${column.Field} (${column.Type}${column.Key === 'PRI' ? ', PRIMARY KEY' : ''}${column.Key === 'MUL' ? ', FOREIGN KEY' : ''})\n`;
            }
            structure += '\n';
        }
        return structure;
    } catch (error) {
        console.error(`Error retrieving table structure for schema ${schema}:`, error);
        return '';
    }
};

module.exports = UMYSQL;

