const DEBUG_ENDPOINT = 'http://localhost:5768/debug';

function sendDebugData(debugData) {
  fetch(DEBUG_ENDPOINT, {
    method: 'POST',
    headers: {
      'Content-Type': 'application/json',
    },
    body: JSON.stringify(debugData),
  })
    .then(response => {
      // Handle successful response if needed
    })
    .catch(error => {
      // Handle error if needed
    });
}

function captureMessage(project, messageType, messageBody) {
  const debugData = {
    project,
    messageType,
    messageBody: JSON.stringify(messageBody)
  };
  sendDebugData(debugData);
}

function captureError(project, error) {
  const debugData = {
    project,
    messageType: 'error',
    error: error.toString(),
    stackTrace: error.stack
  };
  sendDebugData(debugData);
}

function initializeDebugLogger(project) {
  // Override console methods
  const originalConsoleLog = console.log;
  console.log = function(...args) {
    originalConsoleLog.apply(console, args);
    captureMessage(project, 'log', args);
  };

  const originalConsoleError = console.error;
  console.error = function(...args) {
    originalConsoleError.apply(console, args);
    captureMessage(project, 'error', args);
  };

  // Capture unhandled promise rejections
  process.on('unhandledRejection', (reason, promise) => {
    captureError(project, reason);
  });

  // Capture uncaught exceptions
  process.on('uncaughtException', error => {
    captureError(project, error);
  });
}

//initializeDebugLogger("Stubs");