// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE

// Defines jumpToLine command. Uses dialog.js if present.

(function(mod) {
  if (typeof exports == "object" && typeof module == "object") // CommonJS
    mod(require("../../lib/codemirror"), require("../dialog/dialog"));
  else if (typeof define == "function" && define.amd) // AMD
    define(["../../lib/codemirror", "../dialog/dialog"], mod);
  else // Plain browser env
    mod(CodeMirror);
})(function(CodeMirror) {
  "use strict";

  // default search panel location
  CodeMirror.defineOption("search", {bottom: false});
/*
  function dialog(cm, text, shortText, deflt, f) {
    if (cm.openDialog) cm.openDialog(text, f, {value: deflt, selectValueOnOpen: true, bottom: cm.options.search.bottom});
    else f(prompt(shortText, deflt));
  }
*/
function createCustomDialog(options) {
  return new Promise((resolve) => {
    const dialog = document.createElement('div');
    dialog.style.cssText = `
      position: fixed;
      top: 50%;
      left: 50%;
      transform: translate(-50%, -50%);
      background: white;
      padding: 20px;
      border-radius: 5px;
      box-shadow: rgba(0, 0, 0, 0.3) 0px 0px 10px;
      z-index: 1000;
    `;

    const title = document.createElement('h3');
    title.textContent = options.title;
    dialog.appendChild(title);

    const input = document.createElement('input');
    input.type = 'input';
    input.className = 'stubs_form_input';
    input.value = options.value || '';
    input.style.cssText = 'width: 80%; margin-bottom: 10px; margin-top: 10px;';
    dialog.appendChild(input);

    const buttonContainer = document.createElement('div');
    buttonContainer.style.textAlign = 'right';

    const cancelButton = document.createElement('button');
    cancelButton.textContent = 'Cancel';
    cancelButton.className = 'stubs_btn stubs_btn_outline-primary stubs_btn-sm';
    cancelButton.style.marginRight = '10px';
    cancelButton.onclick = () => {
      document.body.removeChild(dialog);
      resolve(null);
    };
    buttonContainer.appendChild(cancelButton);

    const okButton = document.createElement('button');
    okButton.textContent = 'OK';
    okButton.className = 'stubs_btn-sm stubs_btn stubs_btn-primary';
    okButton.onclick = () => {
      document.body.removeChild(dialog);
      resolve(input.value);
    };
    buttonContainer.appendChild(okButton);

    dialog.appendChild(buttonContainer);
    document.body.appendChild(dialog);
    input.focus();
  });
}
    
function dialog(cm, text, shortText, deflt, f) {
  if (cm.openDialog) {
    cm.openDialog(text, f, {value: deflt, selectValueOnOpen: true, bottom: cm.options.search.bottom});
  } else {
    createCustomDialog({
      title: 'CodeMirror Search',
      label: shortText,
      value: deflt,
      type: 'input'
    }).then((result) => {
      if (result !== null) {
        f(result);
      }
    }).catch(console.error);
  }
}  

  function getJumpDialog(cm) {
    return cm.phrase("Jump to line:") + ' <input type="text" style="width: 10em" class="CodeMirror-search-field"/> <span style="color: #888" class="CodeMirror-search-hint">' + cm.phrase("(Use line:column or scroll% syntax)") + '</span>';
  }

  function interpretLine(cm, string) {
    var num = Number(string)
    if (/^[-+]/.test(string)) return cm.getCursor().line + num
    else return num - 1
  }

  CodeMirror.commands.jumpToLine = function(cm) {
    var cur = cm.getCursor();
    dialog(cm, getJumpDialog(cm), cm.phrase("Jump to line:"), (cur.line + 1) + ":" + cur.ch, function(posStr) {
      if (!posStr) return;

      var match;
      if (match = /^\s*([\+\-]?\d+)\s*\:\s*(\d+)\s*$/.exec(posStr)) {
        cm.setCursor(interpretLine(cm, match[1]), Number(match[2]))
      } else if (match = /^\s*([\+\-]?\d+(\.\d+)?)\%\s*/.exec(posStr)) {
        var line = Math.round(cm.lineCount() * Number(match[1]) / 100);
        if (/^[-+]/.test(match[1])) line = cur.line + line + 1;
        cm.setCursor(line - 1, cur.ch);
      } else if (match = /^\s*\:?\s*([\+\-]?\d+)\s*/.exec(posStr)) {
        cm.setCursor(interpretLine(cm, match[1]), cur.ch);
      }
    });
  };

  CodeMirror.keyMap["default"]["Alt-G"] = "jumpToLine";
});
