const { app, BrowserWindow, BrowserView, protocol, screen, ipcMain, globalShortcut, dialog, desktopCapturer, shell, Tray, Menu, nativeImage} = require('electron');
const storage = require('electron-json-storage'); // Add this line to import electron-json-storage
const path = require('path');
const express = require('express');
const bodyParser = require('body-parser');
const fetch = require('node-fetch-commonjs');
const keytar = require('keytar');
const initializeRealtimeClient = require('./main_realtime.js');
//const pty = require('node-pty');
//when ready to use pty do
//npm install @xterm/xterm @xterm/addon-fit node-pty

const os = require('os');


// Assign node-fetch to the global fetch variable
global.fetch = fetch;
global.Headers = fetch.Headers;
global.Request = fetch.Request;
global.Response = fetch.Response;
const mainEmbeddings = require('./main_embeddings.js');
const { generateDocumentEmbeddings } = mainEmbeddings;

require('./main_embeddings.js');

//const Parser = require('tree-sitter');
var win;
var transcriptionWindow = null;
let WIN_LOADED = false;
let EXPRESS_LOADED = false;
let DEV_TOOLS_ACTIVE = !app.isPackaged; 
let HAS_TRAY = false;
let tray = null;
let popupWindow = null;
let appPath = app.getAppPath();
let authWindow;
let authWindowType = "login";
//or payment
let realtimeWindow = null;

function createWindow() {
    
   const { width, height } = screen.getPrimaryDisplay().workAreaSize;    
   const windowWidth = Math.round(width * 0.9);
   const windowHeight = Math.round(height * 0.9);    
   
   
    // Create the window first with default dimensions
    win = new BrowserWindow({
        width: windowWidth,
        height: windowHeight,
        webPreferences: {
            nodeIntegration: true,
            contextIsolation: false,
            webSecurity: false,
            enableRemoteModule: true 
        },
        icon: path.join(appPath, 'images', 'stubslogo-colored_512x512.png'),
        titleBarStyle: 'hidden'
    });
    
    
    
    if(DEV_TOOLS_ACTIVE){      
        win.webContents.openDevTools();
    }    

    win.webContents.on('did-finish-load', () => {
      win.webContents.executeJavaScript(`
        delete window.navigator.serviceWorker;
      `);
    });
    win.loadFile(path.join(appPath, 'index.html'));

    // Asynchronously load window state and update dimensions
    storage.get('windowState', (error, data) => {
        if (error) {
            console.error('Error loading window state:', error);
            return;
        }

        if (data && data.width && data.height && data.x && data.y) {
            win.setBounds({
                x: data.x,
                y: data.y,
                width: data.width,
                height: data.height
            });
        }
    });

    win.on('resize', saveWindowState);
    win.on('move', saveWindowState);
    win.on('ready-to-show', () => {
        WIN_LOADED = true;            
    });     
    
}

function createTray() {
  const iconPath = path.join(appPath, 'images', 'stubs_circle_icon_512x512d_transparent.png');
  const trayIcon = nativeImage.createFromPath(iconPath);
  tray = new Tray(trayIcon);

  const contextMenu = Menu.buildFromTemplate([
    {
      label: 'Exit',
      click: () => {
        app.quit();
      }
    }
  ]);

  tray.setToolTip('Stubs');
  tray.setContextMenu(contextMenu);

  tray.on('click', () => {
      if (popupWindow) {
        popupWindow.close();
      } else {
        createPopupWindow();
      }
  });
}

function createPopupWindow() {
  if (popupWindow) {
    popupWindow.focus();
    return;
  }

  const trayBounds = tray.getBounds();
  const x = Math.round(trayBounds.x + (trayBounds.width / 2) - (300 / 2));
  const y = Math.round(trayBounds.y - 200);

  popupWindow = new BrowserWindow({
    width: 300,
    height: 200,
    x: x,
    y: y,
    alwaysOnTop: true,
    frame: false,
    resizable: false,
    show: false,
    webPreferences: {
      nodeIntegration: true,
      contextIsolation: false
    }
  });

  popupWindow.loadFile(path.join(appPath, 'views', 'app_tray_popup.html'));

  popupWindow.on('closed', () => {
    popupWindow = null;
  });
  /*
  popupWindow.on('blur', () => {
    popupWindow.close();
  });
  */
  popupWindow.on('ready-to-show', () => {
    popupWindow.show();
  });
    setTimeout(() => {
      if(DEV_TOOLS_ACTIVE){
          popupWindow.webContents.toggleDevTools();
      }
    }, 2000);       
}
app.commandLine.appendSwitch('enable-unsafe-webgpu');

app.whenReady().then(() => {
  createWindow();
  //setupExpressTranscriber();
  if(HAS_TRAY){
      createTray();
  }    
    
  app.setAsDefaultProtocolClient('stubs');

  protocol.registerHttpProtocol('stubs', (request, callback) => {
    console.log("here with request for stubs");
    console.log(request);
    if (request.url.startsWith('stubs://login-success')) {
      const token = request.url.split('?token=')[1];
      //save the token to the api credentials
      keytar.setPassword('api_credentials', 'user-token', token);
      console.log("token111111111 set to " + token);
      win.webContents.send('updateUserToken', token);
      //authWindow.webContents.send('should-close-window', '');
      //authWindow.webContents.send('login-success', token);
      if(authWindowType == 'login'){
          authWindow.close();
      }else{
          authWindow.webContents.send('pricing-page', "");
      }
    }
    console.log("No callback?");
    //callback({ statusCode: 200 });

  });
});


app.on('window-all-closed', () => {
  app.quit();
});

app.on('will-quit', () => {
  // Clean up resources here
  if (tray) {
    tray.destroy();
  }
  // Close any open windows
  BrowserWindow.getAllWindows().forEach(window => window.close());
});

/*
ipcMain.handle('login-success', async (event, token) => {
  try {
    await keytar.setPassword('api_credentials', 'user-token', token);
    console.log("token2222222222 set to " + token);
    authWindow.close();
  } catch (error) {
    console.error('Error saving user token:', error);
  }
});
*/

app.on('activate', () => {
  if (BrowserWindow.getAllWindows().length === 0) {
    createWindow();
  }
  
  if (!globalShortcut.isRegistered('Control+Shift+I')) {
    globalShortcut.register('Control+Shift+I', () => {
      // Toggle Developer Tools for the focused window
      let focusedWindow = BrowserWindow.getFocusedWindow();
      if (focusedWindow) {
        focusedWindow.webContents.toggleDevTools();
      }
    });  
  }  
});
function saveWindowState() {
  const bounds = win.getBounds();
  storage.set('windowState', { 
    width: bounds.width, 
    height: bounds.height,
    x: bounds.x,
    y: bounds.y
    }, (error) => {
      if (error) {
        console.error('Error saving window state:', error);
      }
  });
}


  var setupExpressDebug = async function(){
      if(EXPRESS_LOADED){
          return;
      }
      EXPRESS_LOADED = true;
        const expressApp = express();
        expressApp.use(bodyParser.json());

        expressApp.post('/debug', (req, res) => {
          const debugData = req.body;
          console.log("IN MAIN.JS with debug data:");
          console.log(debugData);

          const { project, messageType, messageBody, error, stackTrace } = debugData;
          console.log("project is: " + project);
          console.log("messageType is: " + messageType);
          console.log("messageBody is: " + messageBody);
          console.log("error is: " + error);
          console.log("stackTrace is: " + stackTrace);
          

          if(!WIN_LOADED){
              return;
          }
          win.webContents.send('storeDebugData', debugData);
          
          res.sendStatus(200);
        });
        
        console.log("LISTENING FOR DEBUG MESSAGES ");


        expressApp.listen(5768, () => {
          console.log('Debug Receive Server is running on port 5768');
        });  
      
  };
  
  /*
function setupExpressTranscriber() {
  expressApp = express();

  // Serve static files from the 'views' and 'js' directories
  expressApp.use(express.static(path.join(__dirname, 'views')));
  expressApp.use('/js', express.static(path.join(__dirname, 'js')));

  // Serve the worker bundle
  expressApp.use(express.static(path.join(__dirname, '.')));
  //json enable?
  expressApp.use(express.json());
  expressApp.use(express.urlencoded({ extended: true }));  

  // Route for the transcription page
  expressApp.get('/transcription', (req, res) => {
    res.sendFile(path.join(__dirname, 'views', 'transcription.html'));
  });
  
  // New endpoint for receiving transcription results
    expressApp.post('/transcription-result', (req, res) => {
        console.log('Received request with body:', req.body);
        try {
            // Assuming the body directly contains the transcription text
            const { transcription } = req.body;
            console.log('Transcription:', transcription);

            // Send the data to your renderer process if needed
            if (win && win.webContents) {
                win.webContents.send('transcriptionData', transcription);
            }
            res.status(200).send('Transcription received successfully');
        } catch (error) {
            console.error('Error handling transcription:', error);
            res.status(500).send('Server error');
        }
    });


  // Start the server
  const server = expressApp.listen(3000, () => {
    console.log('Express server running on http://localhost:3000');
  });
}
*/

ipcMain.on('transcription-result', (event, data) => {
    console.log("received ipc message for result");
    console.log(data);
  if (win && win.webContents) {
    win.webContents.send('transcriptionData', data);
  }
});

app.on('will-quit', () => {
  // Unregister all global shortcuts
  globalShortcut.unregisterAll();
});

ipcMain.on('minimizeApp', () => {
    console.log("see if output goes to command line");
  win.minimize();
});

ipcMain.on('maximizeRestoreApp', () => {
  if (win.isMaximized()) {
    win.restore();
  } else {
    win.maximize();
  }
});

ipcMain.handle('showOpenDialog', async (event, options) => {
  const result = await dialog.showOpenDialog(options);
  return result;
});

ipcMain.on('closeApp', () => {
  if(transcriptionWindow != null){ 
    transcriptionWindow.close();
  }
  win.close();
});

ipcMain.on('captureAllWindows', async (event) => {
  console.log("in ipc request");
  try {
      
      
    const sources = await desktopCapturer.getSources({
        types: ['window', 'screen'], // Add 'screen' to capture all windows including the current one
        thumbnailSize: {
            height: 768,
            width: 1366
        }
    });
    const screenshots = await Promise.all(
      sources.map(async (source) => {
        const screenshot = await source.thumbnail.toDataURL();
        return { title: source.name, screenshot: screenshot, fullimage: screenshot };
      })
    );
    event.reply('screenshotsCaptured', screenshots);
  } catch (error) {
    console.error('Error capturing screenshots:', error);
  }
});



ipcMain.handle('get-user-app-paths', () => {
   const appPath = app.getAppPath();
   const userDataPath = app.getPath('userData');
   const isPackaged = app.isPackaged;
   const appVersion = app.getVersion();
   const platform = process.platform;   
   const chromePath = process.execPath;
   return {
      userDataPath,
      appPath,
      isPackaged,
      appVersion,
      platform,
      chromePath
   };
});


ipcMain.on('setupDebugListener', () => {
    setupExpressDebug();
});

ipcMain.on('openTerminal', () => {
    createTerminalWindow();
});

ipcMain.on('open-external-link', (event, url) => {
  shell.openExternal(url);
});

ipcMain.on('popupText', (event, text) => {
  win.webContents.send('popupText', text);
  popupWindow.close();
});

ipcMain.handle('start-crawl', async (event, documentationEmbeddings) => {
  try {
    console.log("got event to start crawl");
    const results = await CRAWL.crawlDocumentationUrls(documentationEmbeddings);
    return results;
  } catch (error) {
    console.error('Crawling failed:', error);
    throw error;
  }
});


ipcMain.handle('get-lint-errors', async (event, theData) => {
    let filepath = theData.filepath;
    let fileContent = theData.filecontent;
      /*
    const eslint = new ESLint({ overrideConfigFile: path.join(appPath, '.eslintrc.js') });

    const results = await eslint.lintText(fileContent);

    // Output any syntax errors
    const syntaxErrors = results[0].messages.filter(message => message.severity === 2);

    if (syntaxErrors.length > 0) {
        console.log(`Syntax errors found in ${filepath}:`);
        syntaxErrors.forEach(error => {
            console.log(`Line ${error.line}: ${error.message}`);
        });
    } 
    else {
        console.log("No syntax errors found.");
    }
    */
    //return syntaxErrors;
});


ipcMain.handle('save-connection', async (event, connectionDetails) => {
    try {
        await keytar.setPassword('database_connections', connectionDetails.name, JSON.stringify(connectionDetails));
        return true;
    } catch (error) {
        console.error('Error saving connection details:', error);
        return false;
    }
});

ipcMain.handle('delete-connection', async (event, connectionName) => {
    try {
        await keytar.deletePassword('database_connections', connectionName);
        return true;
    } catch (error) {
        console.error('Error deleting connection:', error);
        return false;
    }
});

ipcMain.handle('get-connections', async () => {
    try {
        const connectionNames = await keytar.findCredentials('database_connections');
        const connections = {};

        for (const { account, password } of connectionNames) {
            connections[account] = JSON.parse(password);
        }

        return connections;
    } catch (error) {
        console.error('Error retrieving connection details:', error);
        return {};
    }
});

ipcMain.handle('logout-user', async () => {
  try {
    await keytar.deletePassword('api_credentials', 'user-token');    
    return true;
  } catch (error) {
    console.error('Error logging out user:', error);
    return false;
  }
});

// Example of saving other data (API credentials)
ipcMain.handle('save-api-credentials', async (event, apiCredentials) => {
    try {
        await keytar.setPassword('api_credentials', apiCredentials.name, JSON.stringify(apiCredentials));
        return true;
    } catch (error) {
        console.error('Error saving API credentials:', error);
        return false;
    }
});

ipcMain.handle('get-api-credentials', async () => {
    try {
        const openaiCredentials = await keytar.getPassword('api_credentials', 'openai');
        const geminiCredentials = await keytar.getPassword('api_credentials', 'gemini');
        const anthropicCredentials = await keytar.getPassword('api_credentials', 'anthropic');
        const groqCredentials = await keytar.getPassword('api_credentials', 'groq'); 
        const mistralCredentials = await keytar.getPassword('api_credentials', 'mistral'); 
        const apiCounterRecords = await keytar.getPassword('api_credentials', 'generic');
        const userTokenRecords = await keytar.getPassword('api_credentials', 'user-token');

        return {
            openai: openaiCredentials ? JSON.parse(openaiCredentials).key : '',
            gemini: geminiCredentials ? JSON.parse(geminiCredentials).key : '',
            anthropic: anthropicCredentials ? JSON.parse(anthropicCredentials).key : '',
            groq: groqCredentials ? JSON.parse(groqCredentials).key : '',
            mistral: mistralCredentials ? JSON.parse(mistralCredentials).key : '',
            generic: apiCounterRecords ? JSON.parse(apiCounterRecords) : [],
            user_token: userTokenRecords ? userTokenRecords : ''
        };
    } catch (error) {
        console.error('Error retrieving API credentials:', error);
        return {};
    }
});

ipcMain.handle('save-api', async (event, additionalInfo) => {
  try {
    let now = new Date();
    let fullDate = new Date().toISOString();
    let currentDate = fullDate.slice(0, 10);
    let usageRecord = {
      date: currentDate,
      full_date: fullDate,
      count: 1,
      info: additionalInfo
    };
    let apiCounterRecords = await keytar.getPassword('api_credentials', 'generic');
    let records = apiCounterRecords ? JSON.parse(apiCounterRecords) : [];
    
    // Remove records older than a week
    let twoDaysAgo = new Date();
    twoDaysAgo.setDate(twoDaysAgo.getDate() - 2);
    records = records.filter(record => new Date(record.date) >= twoDaysAgo);
    
    records.push(usageRecord);
    await keytar.setPassword('api_credentials', 'generic', JSON.stringify(records));
    return true;
  } catch (error) {
    console.error('Error incrementing API usage:', error);
    return false;
  }
});


function createAuthWindow(options) {
   
  let userToken = "";
  let buttonType = "login";
  //or upgrade
  
  if(options){
      if(options['user_token']){
          userToken = options['user_token'];
          console.log("token " + options['user_token']);
      }
      if(options['button_type']){
          buttonType = options['button_type'];
      }      
  }
  authWindowType = buttonType;
  //const uniquePartition = `persist:auth-${Date.now()}`;   
  //add to webPreferences
  //,
  //partition: uniquePartition

  authWindow = new BrowserWindow({
    width: 900,
    height: 750,
    autoHideMenuBar: false, // Set autoHideMenuBar to false
    menuBarVisibility: 'visible', // Set menuBarVisibility to 'visible'
    webPreferences: {
      nodeIntegration: false,
      preload: path.join(appPath, 'js', 'authwindow', 'preload.js')
    },
    frame: true 
  });
  
  authWindow.setMenuBarVisibility(false);

  authWindow.loadURL(`https://stubs.dev/signup?app=1&type=${buttonType}`);
  
  authWindow.webContents.on('did-finish-load', () => {
    if(DEV_TOOLS_ACTIVE){
        authWindow.webContents.toggleDevTools();
    }
    setTimeout(function(){
        if (userToken && userToken != "") {
          console.log("sending auto-login");
          authWindow.webContents.send('auto-login', userToken);
        }      
    }, 500);    

  });  

  authWindow.on('closed', () => {
    authWindow = null;
  });
  

}


ipcMain.handle('auth-window', async (event, additionalInfo) => {
    console.log("got additional info");
    console.log(additionalInfo);
    createAuthWindow(additionalInfo);
});
/*
ipcMain.handle('voice-commands', async (event, additionalInfo) => {
    console.log("doing voice commands");
    createTranscriptionView();
});
*/
ipcMain.handle('voice-electron', async (event, additionalInfo) => {
    console.log("doing voice commands");
    createElectronTranscriptionView();
});

function createTranscriptionView() {

    
  const transcriptionUrl = 'http://localhost:3000/transcription';
  
  shell.openExternal(transcriptionUrl)
    .then(() => {
      console.log('Transcription page opened in default browser');
    })
    .catch(err => {
      console.error('Failed to open transcription page:', err);
    });    
}

function createElectronTranscriptionView(){
    console.log("we are really doing browser window");
    transcriptionWindow = new BrowserWindow({
    width: 480,
    height: 520,
    webPreferences: {
      nodeIntegration: true,
      contextIsolation: false,
      webSecurity: false,
      enableRemoteModule: true,
      allowRunningInsecureContent: true,
      experimentalFeatures: true,
      enableBlinkFeatures: 'WebGPU',
      additionalArguments: ['--enable-unsafe-webgpu']
    }
  });
  transcriptionWindow.loadFile(path.join(appPath, 'views', 'transcription.html'));
  //transcriptionWindow.loadURL('http://localhost:3000/transcription');
  

  // Optional: Open DevTools for the transcription window
  if (DEV_TOOLS_ACTIVE) {
      setTimeout(function(){
        transcriptionWindow.webContents.openDevTools({ mode: 'detach' });          
      }, 100);
  }
  
  transcriptionWindow.on('closed', () => {
      transcriptionWindow = null;
  });  
  
}


function createRealtimeWindow() {
  realtimeWindow = new BrowserWindow({
    width: 600,
    height: 800,
    webPreferences: {
      nodeIntegration: false,
      contextIsolation: true,
      preload: path.join(appPath, 'js', 'realtime', 'preload_realtime.js'),
    },
  });

  realtimeWindow.loadFile(path.join(appPath, 'views', 'realtime_tasks.html'));

  if (DEV_TOOLS_ACTIVE) {
    realtimeWindow.webContents.openDevTools({ mode: 'detach' });
  }

  realtimeWindow.on('closed', () => {
    realtimeWindow = null;
  });

  // Initialize the RealtimeClient after the window is created
  keytar.getPassword('api_credentials', 'openai')
    .then(openaiCredentials => {
      if (openaiCredentials) {
        const apiKey = JSON.parse(openaiCredentials).key;
        initializeRealtimeClient(apiKey, () => realtimeWindow);
      } else {
        console.error('OpenAI API key not found');
      }
    })
    .catch(error => {
      console.error('Error retrieving OpenAI API key:', error);
    });
}

ipcMain.handle('open-realtime-window', async (event, projectData) => {
  if (!realtimeWindow) {
    createRealtimeWindow(projectData);
  } else {
    realtimeWindow.focus();
    realtimeWindow.webContents.send('project-data', projectData);
  }
});

ipcMain.handle('get-openai-api-key', async () => {
  try {
    const openaiCredentials = await keytar.getPassword('api_credentials', 'openai');
    const apiKey = openaiCredentials ? JSON.parse(openaiCredentials).key : '';
    return apiKey;
  } catch (error) {
    console.error('Error retrieving OpenAI API key:', error);
    return '';
  }
});


function createTerminalWindow() {
    console.log("disabled");
    /*
  const terminalWindow = new BrowserWindow({
    width: 800,
    height: 600,
    webPreferences: {
      nodeIntegration: true,
      contextIsolation: false, // Ensure this is set to false
    },
  });

  terminalWindow.loadFile(path.join(appPath, 'views', 'terminal.html'));

  if (DEV_TOOLS_ACTIVE) {
    terminalWindow.webContents.openDevTools();
  }

  let ptyProcess;

  // Start PTY when renderer is ready
  ipcMain.on('pty-start', (event) => {
    const shell = os.platform() === 'win32' ? 'powershell.exe' : 'bash';

    ptyProcess = pty.spawn(shell, [], {
      name: 'xterm-color',
      cwd: process.cwd(),
      env: process.env,
    });

    // Send PTY data to renderer
    ptyProcess.onData((data) => {
      event.sender.send('pty-data', data);
    });
  });

  // Receive user input from renderer
  ipcMain.on('pty-input', (event, data) => {
    if (ptyProcess) {
      ptyProcess.write(data);
    }
  });

  // Handle terminal resize
  ipcMain.on('pty-resize', (event, size) => {
    if (ptyProcess) {
      ptyProcess.resize(size.cols, size.rows);
    }
  });
    */
}



ipcMain.on('close-application', () => {
  app.quit();
});

ipcMain.handle('authenticate-user', async (event, token) => {
    try {
        const response = await fetch('https://stubs.dev/api/authenticate_user', {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json'
            },
            body: JSON.stringify({ user_token: token })
        });
        const data = await response.json();
        return data;
    } catch (error) {
        console.error('Error authenticating user:', error);
        return null;
    }
});


ipcMain.handle('handle-scrape-crawl', async (event, docName, url, type, userToken) => {
  try {
    const endpoint = type === 'crawl' ? 'handle_crawl_request' : 'handle_scrape_request';
    const response = await fetch(`https://stubs.dev/api/${endpoint}`, {
      method: 'POST',
      headers: {
        'Content-Type': 'application/json'
      },
      body: JSON.stringify({ url, user_token: userToken })
    });
    const data = await response.json();
    
    let retObject = {
        success: data['success']        
    };
    if(data['success'] == true){
    
        if(type == 'scrape'){
            //get vector embeddings for the file, put it on data object
            retObject['markdown'] = data['markdown'];
            retObject['doc_id'] = data['doc_id'];
            retObject['embedding_name'] = docName;
            if(data['metadata'] && data['metadata']['title']){
                retObject['document_name'] = data['metadata']['title'];
            }else if(data['metadata'] && data['metadata']['ogTitle']){
                retObject['document_name'] = data['metadata']['title'];
            }
            let embeddingArr = [retObject];
            try {
              const docEmbeddings = await generateDocumentEmbeddings(embeddingArr);
              console.log("embedding from one doc: ");
              console.log(docEmbeddings);
              let theEmbedding = docEmbeddings[0];
              retObject['vector_data'] = theEmbedding;
              
              //for scrape the url is just the same url
              retObject['url'] = url;
              
            } catch (error) {
              console.error('Error generating embeddings:', error);
              return null;
            }           
            console.log("returning retObject");
            console.log(retObject);
        }else{
            //return the crawl job info
        }
    }else{
        retObject['error'] = "Error with " + type + " request: " + data.toString();
    }
    
    return retObject;
  } catch (error) {
    console.error('Error in handle-scrape-crawl:', error);
    throw error;
  }
});

/*
ipcMain.handle('log-user', async (event, token) => {
    try {
        const response = await fetch('https://stubs.dev/api/log_open', {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json'
            },
            body: JSON.stringify({ user_token: '' })
        });
        const data = await response.json();
        return data;
    } catch (error) {
        console.error('Error logging open user:', error);
        return null;
    }
});
*/

ipcMain.handle('clear-all-keytar-data', async () => {
  try {
    const databaseCredentials = await keytar.findCredentials('database_connections');
    for (const { account } of databaseCredentials) {
      await keytar.deletePassword('database_connections', account);
    }

    const apiCredentials = await keytar.findCredentials('api_credentials');
    for (const { account } of apiCredentials) {
      await keytar.deletePassword('api_credentials', account);
    }

    return true;
  } catch (error) {
    console.error('Error clearing keytar data:', error);
    return false;
  }
});




