const fs = require('fs');
const path = require('path');
const Parser = require('web-tree-sitter');
const cssWhat = require('css-what');
var cssjs = require("jotform-css.js");
const cheerio = require('cheerio');

let baseWasm = path.join(__dirname, 'jslib', 'wasm', 'tree-sitter-javascript.wasm');

// Create a new parser instance
let parser;
var DGRAPH = DGRAPH || {};

// Function to extract functions from a file
// Function to extract functions from a file
DGRAPH.extractFunctions = function(filePath) {
  const content = fs.readFileSync(filePath, 'utf-8');
  const tree = parser.parse(content);
  const functions = [];
  
  var extractFunctionFromLine = function(lineText){
      let functionName = lineText.split('(')[0].trim();
      functionName = replaceAll(functionName, "var ", "");
      functionName = replaceAll(functionName, "async", "");
      functionName = replaceAll(functionName, "function", "");
      functionName = replaceAll(functionName, "\n", "");
      functionName = replaceAll(functionName, "\r", "");
      functionName = replaceAll(functionName, ":", "");
      functionName = replaceAll(functionName, "=", "");    
      functionName = replaceAll(functionName, "}", "");
      functionName = replaceAll(functionName, ",", "");      
      functionName = replaceAll(functionName, "setTimeout", "");
      functionName = functionName.trim();
      return functionName;     
  };

  function traverseNode(node) {
    const nodeType = node.type;
    const startLine = node.startPosition.row;
    const endLine = node.endPosition.row;
    const lineText = content.split('\n')[startLine];



    if (nodeType === 'function_declaration') {
      console.log(`1111 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');

      const functionNameNode = node.childForFieldName('name');
      if (functionNameNode) {
        const functionName = functionNameNode.text;
        functions.push({name: functionName,
            line_text: lineText,
            start_line: startLine,
            end_line: endLine,
            extract_needed: false,
            selectors: []
        });
      }
      else{
        let functionName = extractFunctionFromLine(lineText);
        if(functionName != ''){
            functions.push({name: functionName,
                line_text: lineText,
                start_line: startLine,
                end_line: endLine,
                extract_needed: true,
                selectors: [] 
            });
        }
      }      
    } 
    else if(nodeType == 'function'){
      
      console.log(`2222 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');
      let needExtract = false;
      const functionNameNode = node.childForFieldName('name');
      if (functionNameNode) {
        const functionName = functionNameNode.text;
        functions.push({name: functionName,
            line_text: lineText,
            start_line: startLine,
            end_line: endLine,
            extract_needed: false,
            selectors: [] 
        });
      }
      else{
        let functionName = extractFunctionFromLine(lineText);
        if(functionName != ''){
            functions.push({name: functionName,
                line_text: lineText,
                start_line: startLine,
                end_line: endLine,
                extract_needed: true,
                selectors: [] 
            });
        }
      }         
    }
    /*
    else if (nodeType === 'variable_declarator') {
      console.log(`2222 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');

      const identifierNode = node.childForFieldName('name');
      const initNode = node.childForFieldName('value');

      if (identifierNode && initNode && initNode.type === 'function') {
        const functionName = identifierNode.text;
        functions.push(functionName);
      } else if (identifierNode && initNode && initNode.type === 'arrow_function') {
        const functionName = identifierNode.text;
        functions.push(functionName);
      }
      else{
        functions.push(lineText);
      }      
    }*/
    else if (nodeType === 'method_definition') {
      console.log(`333 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');        
      const methodNameNode = node.childForFieldName('name');
      if (methodNameNode) {
        let functionName = methodNameNode.text;
        functions.push({name: functionName,
            line_text: lineText,
            start_line: startLine,
            end_line: endLine,
            extract_needed: false,
            selectors: [] 
        });
      }
      else{
        let functionName = extractFunctionFromLine(lineText);
        if(functionName != ''){
            functions.push({name: functionName,
                line_text: lineText,
                start_line: startLine,
                end_line: endLine,
                extract_needed: true,
                selectors: [] 
            });
        }
      }
      
    }/*    
    else if (nodeType === 'expression_statement') {
      console.log(`333 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');        
      const methodNameNode = node.childForFieldName('name');
      if (methodNameNode) {
        const methodName = methodNameNode.text;
        functions.push(methodName); // Assuming you want to capture the method name
      }
      else{
        functions.push(lineText);
      }
    }  */   
    else{
        //console.log(`Line Line: ${lineText}`);       
        //console.log(`Skip Line Text: ${lineText}`);
        //console.log(`Skip Node Type: ${nodeType}`);        
    }

    for (const child of node.namedChildren) {
      traverseNode(child);
    }
  }

  traverseNode(tree.rootNode);

  return functions;
}

var escapeRegExp = function(string) {
    return string.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1");
};

var replaceAll = function(string, find, replace) {
  return string.replace(new RegExp(escapeRegExp(find), 'g'), replace);
};

function extractStringsFromLine(line) {
  const regex = /['"]([\w-$.#[\]:>+~=,\s]+)['"]/g;
  const matches = line.match(regex);
  if (matches) {
    return matches.map(match => match.slice(1, -1));
  }
  return [];
}

// Function to process files and build the function data structure
DGRAPH.buildFunctionDataStructure = function(files) {
  const functionDataStructure = {};
  const selectorDataStructure = {};

  files.forEach((file) => {
    console.log("extracting file: " + file);
    const functions = DGRAPH.extractFunctions(file);
    selectorDataStructure[file] = new Array();
    for(let f = 0; f < functions.length; f++){
        let line = functions[f].line_text;
        let lineNum = functions[f].start_line;
        if (line.includes('querySelector') || line.includes('$') || line.includes('addEventListener')) {
            console.log("testing: " + functions[f].line_text);
            if(line.includes("$(" || line.includes('addEventListener'))){
                line.name = "event";
            }
            // Extract potential selectors from the line
            const selectors = extractStringsFromLine(line);
            let validSelectors = [];
            let invalidSelectors = ['body', 'click', 'mousedown','mouseup','mousemove','mouseover','mouseout','mouseenter','mouseleave','dblclick',
            'keydown','keyup','keypress','focus','blur','change','input','submit','reset','load','scroll','resize','touchstart','dragstart','drag','contextmenu'];

            // Check each selector for validity
            selectors.forEach(selector => {
              if (isValidSelector(selector) && !invalidSelectors.includes(selector)) {
                
                validSelectors.push(selector);
                let selectorObject = {line: lineNum,
                    selector: selector
                };
                
                selectorDataStructure[file].push(selectorObject);
              }
            });

            functions[f]['selectors'] = validSelectors;
            console.log("valid selectors");
            console.log(validSelectors);
        }
        if(line.includes('getElementById')){            
          let regex = /getElementById\((['"])(.+?)\1\)/;
          let match = line.match(regex);

          if (match) {
              let idSelector = "#" + match;
              functions[f]['selectors'].push(idSelector);
          }       
        }      

    }      
    functionDataStructure[file] = functions;
  });

  return [functionDataStructure, selectorDataStructure];
}


function extractStringsFromLine(line) {
  const regex = /['"]([\w-$.#[\]:>+~=,\s]+)['"]/g;
  const matches = line.match(regex);
  if (matches) {
    return matches.map(match => match.slice(1, -1));
  }
  return [];
}

function isValidSelector(selector) {
  try {
    cssWhat.parse(selector);
    return true;
  } catch (error) {
    return false;
  }
}

// Function to extract function calls from a file
DGRAPH.extractFunctionCalls = function(file) {
  const content = fs.readFileSync(file, 'utf-8');
  const tree = parser.parse(content);
  const functionCalls = [];

  // Traverse the AST to find function calls
  tree.rootNode.descendantsOfType('call_expression').forEach((callNode) => {
    const functionNameNode = callNode.childForFieldName('function');
    const jqueryFunctionCalls = [
      ".css", ".data", ".html", ".attr", ".val", ".text", ".append", ".prepend", ".on(",
      ".after", ".before", ".remove", ".empty", ".show", ".hide", ".toggle",
      ".addClass", ".removeClass", ".toggleClass", ".hasClass", ".width",
      ".height", ".offset", ".position", ".scrollLeft", ".scrollTop", ".animate"
    ];

    const javascriptCalls = [
      "console.log", "console.error", "console.warn", "console.info", "console.debug",
      "console.trace", ".preventDefault", ".stopPropagation", ".addEventListener", ".removeEventListener", ".dispatchEvent", ".setTimeout", ".clearTimeout", ".setInterval", ".clearInterval",
      ".requestAnimationFrame", ".cancelAnimationFrame", ".getElementById", ".getElementsByClassName", ".getElementsByTagName",
      ".querySelector", ".querySelectorAll", ".createElement", ".createTextNode", ".createDocumentFragment",
      ".appendChild", ".insertBefore", ".removeChild", ".replaceChild", ".cloneNode", ".contains", ".compareDocumentPosition",
      ".getAttribute", ".setAttribute", ".removeAttribute", ".hasAttribute", ".focus", ".blur", ".click", ".submit",
      ".getElementsByName", ".item", ".namedItem", ".getBoundingClientRect", ".getClientRects",".getComputedStyle", ".matchMedia",
      ".getSelection", ".find", ".findIndex",".forEach", ".map", ".filter", ".reduce", ".reduceRight", ".some", ".every",
      ".indexOf", ".lastIndexOf", ".includes", ".sort", ".reverse",".stringify", ".parse", ".push", ".pop", ".shift", ".unshift",
      ".slice", ".splice", ".concat", ".join", ".split",".startsWith", ".endsWith", ".substring", ".substr", ".replace", ".trim",
      ".toLowerCase", ".toUpperCase", ".charCodeAt", ".fromCharCode",".parseInt", ".parseFloat", ".isNaN", ".isFinite",
      "Math.abs", "Math.ceil", "Math.floor", "Math.round", "Math.max", "Math.min",
      "Math.random", "Date.now", "Date.parse", "Date.UTC",".getFullYear", ".getMonth", ".getDate", ".getDay", ".getHours", ".getMinutes",
      ".getSeconds", ".getMilliseconds", ".getTime", ".getTimezoneOffset", ".toDateString", ".toISOString", ".toJSON",
      ".toLocaleDateString", ".toLocaleString", ".toLocaleTimeString", ".toString",
      ".toTimeString", ".toUTCString", ".valueOf"
    ];    

    if (functionNameNode) {
      const functionName = functionNameNode.text;
      const callerName = getFunctionName(callNode);
      const startLine = callNode.startPosition.row;
      const endLine = callNode.endPosition.row;
      const lineText = content.split('\n')[startLine];

      // Check if the line contains any jQuery function calls
      const isJQueryCall = jqueryFunctionCalls.some((call) => lineText.includes(call));
      const isJavascriptCall = javascriptCalls.some((call) => lineText.includes(call));

      if (!isJQueryCall && !isJavascriptCall) {
        functionCalls.push({
          caller: callerName,
          callee: functionName,
          start_line: startLine,
          end_line: endLine,
          line_text: lineText
        });
      }
    }
  });

  return functionCalls;
}

// Function to extract selectors from an HTML file
DGRAPH.extractSelectorsFromHtml = function(file) {
  const content = fs.readFileSync(file, 'utf-8');
  const $contentHtml = cheerio.load(content);
  const selectors = [];
  const lines = content.split('\n');

  $contentHtml.root().find("*").each(function(index, element){
      console.log("processing element: " );
      console.log($contentHtml.html(element));
    const $element = $contentHtml(element);
    const id = $element.attr('id');
    const classes = $element.attr('class');
    
    const startIndex = $contentHtml.html().indexOf($contentHtml.html(element));
    const lineNumber = content.substr(0, startIndex).split('\n').length;

    if (id) {
        console.log("FOUND ID " + id + " on line " + lineNumber);
      selectors.push({ selector: `#${id}`, line: lineNumber });
    }

    if (classes) {  
      classes.split(' ').forEach((className) => {
        console.log("FOUND CLASS " + className + " on line " + lineNumber);    
        selectors.push({ selector: `.${className}`, line: lineNumber });
      });
    }
  });

  return selectors;
}
DGRAPH.extractSelectorsFromCss = function(file) {
    
   var getLineIndex = function(selector, minLineIndex, totalLines, lines) {
    if (minLineIndex >= totalLines) {
      return totalLines;
    }

    const nextOccurrence = lines.slice(minLineIndex).findIndex(line => line.includes(selector));

    if (nextOccurrence === -1) {
      return minLineIndex;
    }

    return minLineIndex + nextOccurrence;
  }    
  const content = fs.readFileSync(file, 'utf-8');
  const lines = content.split('\n');
  const totalLines = lines.length;
  const parser = new cssjs.cssjs();
  let selectors = [];
  let parsedCssNew = parser.parseCSS(content);
  let minLineIndex = 0;

  for (var v = 0; v < parsedCssNew.length; v++) {
    console.log("index" + v);
    console.log(parsedCssNew[v]);

    if (parsedCssNew[v]['selector'].indexOf('@media') >= 0) {
      if ('subStyles' in parsedCssNew[v] && parsedCssNew[v]['subStyles'].length > 0) {
        for (let s = 0; s < parsedCssNew[v]['subStyles'].length; s++) {
          const selector = parsedCssNew[v]['subStyles'][s]['selector'];
          const lineIndex = getLineIndex(selector, minLineIndex, totalLines, lines);
          selectors.push({ selector, line: lineIndex });

          if ('rules' in parsedCssNew[v]['subStyles'][s] && parsedCssNew[v]['subStyles'][s]['rules'].length > 0) {
            minLineIndex += parsedCssNew[v]['subStyles'][s]['rules'].length;
          }
        }
      } else {
        const selector = parsedCssNew[v]['selector'];
        const lineIndex = getLineIndex(selector, minLineIndex, totalLines, lines);
        selectors.push({ selector, line: lineIndex });
      }
    } else {
      const selector = parsedCssNew[v]['selector'];
      const lineIndex = getLineIndex(selector, minLineIndex, totalLines, lines);
      selectors.push({ selector, line: lineIndex });
    }

    if ('rules' in parsedCssNew[v] && parsedCssNew[v]['rules'].length > 0) {
      minLineIndex += parsedCssNew[v]['rules'].length;
    }

    minLineIndex++;
  }

  return selectors;
}



// Function to process files and build the dependency graph

// Function to process files and build the dependency graph
DGRAPH.buildDependencyGraph = function(files, functionDataStructure, functionCallsDataStructure, selectorDataStructure) {
  const dependencyGraph = {};

  files.forEach((file) => {
    if (!dependencyGraph[file]) {
      dependencyGraph[file] = {};
    }      
    const fileExtension = path.extname(file);

    if (fileExtension === '.js') {
      const functionCalls = functionCallsDataStructure[file];

      functionCalls.forEach((functionCall) => {
        const callerName = functionCall.caller;
        const functionName = functionCall.callee;

        // Find the file where the called function is defined
        const calleeFile = findFunctionFile(functionName, functionDataStructure);

        if (calleeFile) {
          // Add the function call to the dependency graph
          if (!dependencyGraph[file]) {
            dependencyGraph[file] = {};
          }
          if (!dependencyGraph[file][callerName]) {
            dependencyGraph[file][callerName] = [];
          }
          dependencyGraph[file][callerName].push(`${calleeFile}/${functionName}`);
        }
      });
    } else if (fileExtension === '.html') {
      const selectors = selectorDataStructure[file];

      selectors.forEach((selector) => {
        if (!dependencyGraph[file]) {
          dependencyGraph[file] = {};
        }
        if (!dependencyGraph[file][selector.selector]) {
          dependencyGraph[file][selector.selector] = [];
        }
        dependencyGraph[file][selector.selector].push(`${file}:${selector.line}`);
      });
    } else if (fileExtension === '.css') {
      const selectors = selectorDataStructure[file];

      selectors.forEach((selector) => {
        if (!dependencyGraph[file]) {
          dependencyGraph[file] = {};
        }
        if (!dependencyGraph[file][selector.selector]) {
          dependencyGraph[file][selector.selector] = [];
        }
        dependencyGraph[file][selector.selector].push(`${file}:${selector.line}`);
      });
    }
    
  });

  return dependencyGraph;
}


DGRAPH.buildDependencyGraphArray = function(files, functionDataStructure, functionCallsDataStructure, functionSelectors, selectorDataStructure) {
  const dependencies = [];

  files.forEach((file) => {
    const fileExtension = path.extname(file);

    if (fileExtension === '.js') {
        console.log("IN JS FILE " + file);
      const functionCalls = functionCallsDataStructure[file];
      const mySelectors = functionSelectors[file];
      //function to function in js files
      functionCalls.forEach((functionCall) => {
        const callerName = functionCall.caller;
        const functionName = functionCall.callee;
        const lineText = functionCall.line_text;

        // Find the file where the called function is defined
        const calleeFile = findFunctionFile(functionName, functionDataStructure);

        if (calleeFile) {
          // Find the line number of the called function in the callee file
          const calleeFunction = functionDataStructure[calleeFile].find((func) => func.name === functionName);
          const calleeLineNumber = calleeFunction ? calleeFunction.start_line : null;
          dependencies.push({
            caller: file,
            line_caller: functionCall.start_line,
            callee: calleeFile,
            line_callee: calleeLineNumber,
            type: 'function',
            name: functionName
          });                   
        }
      });
      
      //function to selector in .css or .html files
      mySelectors.forEach((functionSel) => {
          //console.log("functionSel ");
          //console.log(functionSel);
          
            for(let f = 0; f < files.length; f++){
              let styleFile = files[f];
              let styleExt = path.extname(styleFile);
              if(styleExt == '.html'){
                 let htmlSelectorArr = selectorDataStructure[styleFile];
                 for(let h = 0; h < htmlSelectorArr.length; h++){
                     let hSel = htmlSelectorArr[h];
                     //console.log("testing html");
                     //console.log(hSel);
                     if(hSel['selector'] == functionSel['selector']){
                        dependencies.push({
                          caller: file,
                          line_caller: functionSel.line,
                          callee: styleFile,
                          line_callee: hSel.line,
                          type: 'selector',
                          name: functionSel['selector']
                        });     
                        //console.log("added");
                        //console.log(dependencies);
                     }
                 }
              }else if(styleExt == '.css'){
                 let cssSelectorArr = selectorDataStructure[styleFile];
                 for(let c = 0; c < cssSelectorArr.length; c++){
                     let cSel = cssSelectorArr[c];
                     //console.log("testing css");
                     //console.log(cSel);
                     
                     if(cSel['selector'].indexOf(functionSel['selector']) >= 0){
                        dependencies.push({
                          caller: file,
                          line_caller: functionSel.line,
                          callee: styleFile,
                          line_callee: cSel.line,
                          type: 'selector',
                          name: functionSel['selector']
                        });                
                        
                        //console.log("added");
                        //console.log(dependencies);
                        
                     }
                 }
              }
            }
      });
      
    } else if (fileExtension === '.html') {
      const selectors = selectorDataStructure[file];      
      selectors.forEach((selector) => {
        let htmlSelector = selector.selector;
        let htmlLine = selector.line;
        for(let f = 0; f < files.length; f++){
            let styleFile = files[f];
            let styleExt = path.extname(styleFile);         
            if(styleExt == '.css'){
               let cssSelectorArr = selectorDataStructure[styleFile];
               for(let c = 0; c < cssSelectorArr.length; c++){
                   let cSel = cssSelectorArr[c];
                   if(cSel['selector'].indexOf(htmlSelector) >= 0){
                      dependencies.push({
                        caller: file,
                        line_caller: htmlLine,
                        callee: styleFile,
                        line_callee: cSel.line,
                        type: 'selector',
                        name: htmlSelector
                      });                         
                   }
               }
            }              
        }
      });
    } else if (fileExtension === '.css') {
      //const selectors = selectorDataStructure[file];
    }
  });

  return dependencies;
}


// Function to find the file where a function is defined
function findFunctionFile(functionName, functionDataStructure) {
  for (const file in functionDataStructure) {
    const functions = functionDataStructure[file];
    for (const func of functions) {
      if (func.name === functionName) {
        return file;
      }
    }
  }
  return null;
}

/*
function groupFilesByDependency(dGraphArr, files, maxGroupSize = 5) {
  const groups = [];
  const ungroupedFiles = new Set(files);

  // Create a map to store the dependency count between files
  const dependencyCountMap = new Map();

  // Iterate over the dependencies and count the occurrences
  dGraphArr.forEach((dependency) => {
    const { caller, callee } = dependency;
    const key = `${caller}-${callee}`;

    if (!dependencyCountMap.has(key)) {
      dependencyCountMap.set(key, 0);
    }

    dependencyCountMap.set(key, dependencyCountMap.get(key) + 1);
  });

  while (ungroupedFiles.size > 0) {
    const currentGroup = new Set();

    // Find the file with the highest dependency count
    let maxDependencyCount = 0;
    let maxDependencyFile = null;

    for (const file of ungroupedFiles) {
      let fileDependencyCount = 0;

      for (const [key, count] of dependencyCountMap) {
        const [caller, callee] = key.split('-');

        if (caller === file || callee === file) {
          fileDependencyCount += count;
        }
      }

      if (fileDependencyCount > maxDependencyCount) {
        maxDependencyCount = fileDependencyCount;
        maxDependencyFile = file;
      }
    }

    if (maxDependencyFile) {
      currentGroup.add(maxDependencyFile);
      ungroupedFiles.delete(maxDependencyFile);

      // Find files with high dependency count with the current file
      const relatedFiles = new Set();
      for (const [key, count] of dependencyCountMap) {
        const [caller, callee] = key.split('-');

        if (caller === maxDependencyFile && ungroupedFiles.has(callee)) {
          relatedFiles.add(callee);
        } else if (callee === maxDependencyFile && ungroupedFiles.has(caller)) {
          relatedFiles.add(caller);
        }
      }

      for (const file of relatedFiles) {
        if (currentGroup.size < maxGroupSize) {
          currentGroup.add(file);
          ungroupedFiles.delete(file);
        } else {
          break;
        }
      }
    }

    // Ensure each group has at least one type of each file (.js, .html, .css) if available
    const fileTypes = ['.js', '.html', '.css'];
    for (const fileType of fileTypes) {
      if (currentGroup.size < maxGroupSize) {
        for (const file of ungroupedFiles) {
          if (file.endsWith(fileType)) {
            currentGroup.add(file);
            ungroupedFiles.delete(file);
            break;
          }
        }
      }
    }

    if (currentGroup.size > 0) {
      groups.push(Array.from(currentGroup));
    }

    if (ungroupedFiles.size > 0 && maxDependencyCount === 0) {
      // If there are no more dependencies, add the remaining files to separate groups
      const remainingFiles = Array.from(ungroupedFiles);
      const remainingGroups = [];

      while (remainingFiles.length > 0) {
        const group = new Set();

        // Ensure each remaining group has at least one type of each file (.js, .html, .css) if available
        for (const fileType of fileTypes) {
          for (let i = 0; i < remainingFiles.length; i++) {
            if (remainingFiles[i].endsWith(fileType)) {
              group.add(remainingFiles[i]);
              remainingFiles.splice(i, 1);
              break;
            }
          }
        }

        // Add remaining files to the group until maxGroupSize is reached
        while (group.size < maxGroupSize && remainingFiles.length > 0) {
          group.add(remainingFiles.shift());
        }

        remainingGroups.push(Array.from(group));
      }

      groups.push(...remainingGroups);
      ungroupedFiles.clear();
    }
  }

  return groups;
}

*/
function groupFilesByDependency(dGraphArr, files, maxGroupSize = 5) {
  const groups = [];
  const ungroupedFiles = new Set(files);

  // Create a map to store the dependency count between files
  const dependencyCountMap = new Map();

  // Iterate over the dependencies and count the occurrences
  dGraphArr.forEach((dependency) => {
    const { caller, callee } = dependency;
    const key = `${caller}-${callee}`;

    if (!dependencyCountMap.has(key)) {
      dependencyCountMap.set(key, 0);
    }

    dependencyCountMap.set(key, dependencyCountMap.get(key) + 1);
  });

  while (ungroupedFiles.size > 0) {
    const currentGroup = [];

    // Find the file with the highest dependency count
    let maxDependencyCount = 0;
    let maxDependencyFile = null;

    for (const file of ungroupedFiles) {
      let fileDependencyCount = 0;

      for (const [key, count] of dependencyCountMap) {
        const [caller, callee] = key.split('-');

        if (caller === file || callee === file) {
          fileDependencyCount += count;
        }
      }

      if (fileDependencyCount > maxDependencyCount) {
        maxDependencyCount = fileDependencyCount;
        maxDependencyFile = file;
      }
    }

    if (maxDependencyFile) {
      currentGroup.push(maxDependencyFile);
      ungroupedFiles.delete(maxDependencyFile);

      // Find files with high dependency count with the current file
      const relatedFiles = new Set();
      for (const [key, count] of dependencyCountMap) {
        const [caller, callee] = key.split('-');

        if (caller === maxDependencyFile && ungroupedFiles.has(callee)) {
          relatedFiles.add(callee);
        } else if (callee === maxDependencyFile && ungroupedFiles.has(caller)) {
          relatedFiles.add(caller);
        }
      }

      for (const file of relatedFiles) {
        if (currentGroup.length < maxGroupSize) {
          currentGroup.push(file);
          ungroupedFiles.delete(file);
        } else {
          break;
        }
      }
    }

    if (currentGroup.length > 0) {
      groups.push(currentGroup);
    }

    if (ungroupedFiles.size > 0 && maxDependencyCount === 0) {
      // If there are no more dependencies, add the remaining files to separate groups
      const remainingFiles = Array.from(ungroupedFiles);
      const remainingGroups = [];

      while (remainingFiles.length > 0) {
        const group = remainingFiles.splice(0, maxGroupSize);
        remainingGroups.push(group);
      }

      groups.push(...remainingGroups);
      ungroupedFiles.clear();
    }
  }

  return groups;
}


// Function to get the name of the function containing a call expression
function getFunctionName(callNode) {
  let currentNode = callNode;
  while (currentNode && currentNode.type !== 'function_declaration') {
    currentNode = currentNode.parent;
  }
  if (currentNode) {
    const functionNameNode = currentNode.childForFieldName('name');
    if (functionNameNode) {
      return functionNameNode.text;
    }
  }
  return 'toplevel';
}

// Function to print the dependency graph
// Function to print the dependency graph
function printDependencyGraph(dependencyGraphArray) {
  const fileMap = new Map();

  // Group dependencies by file
  dependencyGraphArray.forEach((dependency) => {
    const { caller, callee, line_caller, line_callee, type, name } = dependency;

    if (!fileMap.has(caller)) {
      fileMap.set(caller, []);
    }

    fileMap.get(caller).push({ callee, line_caller, line_callee, type, name });
  });

  // Print the dependency graph
  for (const [file, dependencies] of fileMap) {
    console.log(`${file}:`);

    dependencies.forEach((dependency) => {
      const { callee, line_caller, line_callee, type, name } = dependency;
      console.log(`  ${name} (${type}) on line ${line_caller} => ${callee}:${line_callee}`);
    });

    console.log();
  }
}

async function testDependencies() {
  try {
    await Parser.init();
    const lan = await Parser.Language.load(baseWasm);
    parser = new Parser();
    parser.setLanguage(lan);
  } catch (error) {
    console.error('Error initializing web-tree-sitter:', error);
    return;
  }

  //let file1 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\tests\\simple_chatgpt.js";
  //let file2 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\tests\\simple_ui_handlers.js";
  
  let file1 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\chatgpt.js";
  let file2 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\codemirror_diff.js";
  let file3 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\dexiedatabase.js";
  let file4 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\directories.js";
  let file5 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\mydatabase.js";
  let file6 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\screenshots.js";
  let file7 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\stub_generator.js";
  let file8 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\tabmanager.js";
  let file9 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\ui_handlers.js";
  let file10 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\ui_modal_handlers.js";  
  let file11 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\index.html";  
  let file12 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\views\\chat_response_container.html";  
  let file13 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\views\\modal_addedit_project.html";  
  let file14 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\views\\modal_settings.html";  
  let file15 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\css\\styles.css";  
  let file16 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\css\\chatgpt_textarea.css";  
  let file17 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\css\\diff-panel.css";  
  
  //console.log("selectors from html");
  //let htmlSelectors = extractSelectorsFromHtml(file11);
   //console.log(htmlSelectors);
   //console.log("selectors from css");
   //let cssSelectors = extractSelectorsFromCss(file16);
   //console.log(cssSelectors);
   //
//return;
  
  //let functionCalls = extractFunctionCalls(file2);
  

  
  //const functions = extractFunctions(file2);
  //console.log("functions:");
  //console.log(functions);
  //console.log("test done");
  
  // Usage example
  const files = [file1, file2, file3, file4, file5, file6, file7, file8, file9, file10, file11, file12, file13, file14, file15, file16, file17];
  

  // First pass: Build the function data structure
  const dataStructures = DGRAPH.buildFunctionDataStructure(files);
  let functionDataStructure = dataStructures[0];
  let functionSelectors = dataStructures[1];
  let selectorDataStructure = {};
  
  const functionCallsDataStructure = {};
  files.forEach((file) => {
         //if file extension js do function calls
         //if css use jotform
         //if html use cheerio to get ids/classes
        //console.log("file: " + file);
        const fileExtension = path.extname(file);

        if (fileExtension === '.js') {
          const functionCalls = DGRAPH.extractFunctionCalls(file);
          functionCallsDataStructure[file] = functionCalls;
        } else if (fileExtension === '.html') {
          const selectors = DGRAPH.extractSelectorsFromHtml(file);
          selectorDataStructure[file] = selectors;
        } else if (fileExtension === '.css') {
          const selectors = DGRAPH.extractSelectorsFromCss(file);
          selectorDataStructure[file] = selectors;
        }
        //console.log(functionCalls);
  });  
  
  
    console.log("files");
    console.log(files);
    console.log("function data structure");
    console.log(functionDataStructure);
    
    console.log("functionCallsDataStructure");
    console.log(functionCallsDataStructure);
    console.log("functionSelectors");
    console.log(functionSelectors);
    console.log("selectorDataStructure");
    console.log(selectorDataStructure);

  // Second pass: Build the dependency graph
    //const dependencyGraph = DGRAPH.buildDependencyGraph(files, functionDataStructure, functionCallsDataStructure, selectorDataStructure);
/*
    let dGraphArr = DGRAPH.buildDependencyGraphArray(files, functionDataStructure, functionCallsDataStructure, functionSelectors, selectorDataStructure);
    
    console.log("d graph array");
    console.log(dGraphArr);
    
    console.log("total dependencies");
    console.log(dGraphArr.length);
    
    // Print the dependency graph
    printDependencyGraph(dGraphArr);
    
    let dependencyGroups = groupFilesByDependency(dGraphArr, files);
    console.log("Dependency Groups:");
    for (let d = 0; d < dependencyGroups.length; d++) {
      console.log(`Group ${d + 1}:`);
      console.log(dependencyGroups[d]);
      console.log("===========================");
    }     
 */
}

testDependencies();