const fs = require('fs');
const path = require('path');
const Parser = require('web-tree-sitter');
const cssWhat = require('css-what');

let baseWasm = path.join(__dirname, 'jslib', 'wasm', 'tree-sitter-javascript.wasm');

// Create a new parser instance
let parser;

// Function to extract functions from a file
function extractFunctions(filePath) {
  const content = fs.readFileSync(filePath, 'utf-8');
  const tree = parser.parse(content);
  const functions = [];
  
  var extractFunctionFromLine = function(lineText){
      let functionName = lineText.split('(')[0].trim();
      functionName = replaceAll(functionName, "var ", "");
      functionName = replaceAll(functionName, "async", "");
      functionName = replaceAll(functionName, "function", "");
      functionName = replaceAll(functionName, "\n", "");
      functionName = replaceAll(functionName, "\r", "");
      functionName = replaceAll(functionName, ":", "");
      functionName = replaceAll(functionName, "=", "");    
      functionName = replaceAll(functionName, "}", "");
      functionName = replaceAll(functionName, ",", "");      
      functionName = replaceAll(functionName, "setTimeout", "");
      functionName = functionName.trim();
      return functionName;     
  };

  function traverseNode(node) {
    const nodeType = node.type;
    const startLine = node.startPosition.row;
    const endLine = node.endPosition.row;
    const lineText = content.split('\n')[startLine];



    if (nodeType === 'function_declaration') {
      console.log(`1111 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');

      const functionNameNode = node.childForFieldName('name');
      if (functionNameNode) {
        const functionName = functionNameNode.text;
        functions.push({name: functionName,
            line_text: lineText,
            start_line: startLine,
            end_line: endLine,
            extract_needed: false,
            selectors: []
        });
      }
      else{
        let functionName = extractFunctionFromLine(lineText);
        if(functionName != ''){
            functions.push({name: functionName,
                line_text: lineText,
                start_line: startLine,
                end_line: endLine,
                extract_needed: true,
                selectors: [] 
            });
        }
      }      
    } 
    else if(nodeType == 'function'){
      
      console.log(`2222 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');
      let needExtract = false;
      const functionNameNode = node.childForFieldName('name');
      if (functionNameNode) {
        const functionName = functionNameNode.text;
        functions.push({name: functionName,
            line_text: lineText,
            start_line: startLine,
            end_line: endLine,
            extract_needed: false,
            selectors: [] 
        });
      }
      else{
        let functionName = extractFunctionFromLine(lineText);
        if(functionName != ''){
            functions.push({name: functionName,
                line_text: lineText,
                start_line: startLine,
                end_line: endLine,
                extract_needed: true,
                selectors: [] 
            });
        }
      }         
    }
    /*
    else if (nodeType === 'variable_declarator') {
      console.log(`2222 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');

      const identifierNode = node.childForFieldName('name');
      const initNode = node.childForFieldName('value');

      if (identifierNode && initNode && initNode.type === 'function') {
        const functionName = identifierNode.text;
        functions.push(functionName);
      } else if (identifierNode && initNode && initNode.type === 'arrow_function') {
        const functionName = identifierNode.text;
        functions.push(functionName);
      }
      else{
        functions.push(lineText);
      }      
    }*/
    else if (nodeType === 'method_definition') {
      console.log(`333 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');        
      const methodNameNode = node.childForFieldName('name');
      if (methodNameNode) {
        let functionName = methodNameNode.text;
        functions.push({name: functionName,
            line_text: lineText,
            start_line: startLine,
            end_line: endLine,
            extract_needed: false,
            selectors: [] 
        });
      }
      else{
        let functionName = extractFunctionFromLine(lineText);
        if(functionName != ''){
            functions.push({name: functionName,
                line_text: lineText,
                start_line: startLine,
                end_line: endLine,
                extract_needed: true,
                selectors: [] 
            });
        }
      }
      
    }/*    
    else if (nodeType === 'expression_statement') {
      console.log(`333 Node Type: ${nodeType}`);
      console.log(node);
      console.log(`Start Line: ${startLine}`);
      console.log(`End Line: ${endLine}`);
      console.log(`Line Text: ${lineText}`);
      console.log('---');        
      const methodNameNode = node.childForFieldName('name');
      if (methodNameNode) {
        const methodName = methodNameNode.text;
        functions.push(methodName); // Assuming you want to capture the method name
      }
      else{
        functions.push(lineText);
      }
    }  */   
    else{
        //console.log(`Line Line: ${lineText}`);       
        //console.log(`Skip Line Text: ${lineText}`);
        //console.log(`Skip Node Type: ${nodeType}`);        
    }

    for (const child of node.namedChildren) {
      traverseNode(child);
    }
  }

  traverseNode(tree.rootNode);

  return functions;
}

var escapeRegExp = function(string) {
    return string.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1");
};

var replaceAll = function(string, find, replace) {
  return string.replace(new RegExp(escapeRegExp(find), 'g'), replace);
};

function extractStringsFromLine(line) {
  const regex = /['"]([\w-$.#[\]:>+~=,\s]+)['"]/g;
  const matches = line.match(regex);
  if (matches) {
    return matches.map(match => match.slice(1, -1));
  }
  return [];
}

function isValidSelector(selector) {
  try {
    cssWhat.parse(selector);
    return true;
  } catch (error) {
    return false;
  }
}

async function testExtract() {
  try {
    await Parser.init();
    const lan = await Parser.Language.load(baseWasm);
    parser = new Parser();
    parser.setLanguage(lan);
  } catch (error) {
    console.error('Error initializing web-tree-sitter:', error);
    return;
  }

  let file1 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\ui_handlers.js";
  let file2 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\chatgpt.js";
  let file3 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\directories.js";
  let file4 = "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\js\\codemirror_diff.js";
  
  let simplefile =  "C:\\wamp\\www\\stubs\\stubs-ai-assistant\\tests\\test_simple_functions.js";
  
  const functions = extractFunctions(file1);
  
  for(let f = 0; f < functions.length; f++){
      let line = functions[f].line_text;
      if (line.includes('querySelector') || line.includes('$') || line.includes('addEventListener')) {
          console.log("testing: " + functions[f].line_text);
          if(line.includes("$(" || line.includes('addEventListener'))){
              line.name = "event";
          }
          // Extract potential selectors from the line
          const selectors = extractStringsFromLine(line);
          let validSelectors = [];
          let invalidSelectors = ['body', 'click', 'mousedown','mouseup','mousemove','mouseover','mouseout','mouseenter','mouseleave','dblclick',
          'keydown','keyup','keypress','focus','blur','change','input','submit','reset','load','scroll','resize','touchstart','dragstart','drag','contextmenu'];
                  
          // Check each selector for validity
          selectors.forEach(selector => {
            if (isValidSelector(selector) && !invalidSelectors.includes(selector)) {
              validSelectors.push(selector);
            }
          });
          
          functions[f]['selectors'] = validSelectors;
          console.log("valid selectors");
          console.log(validSelectors);
      }
      if(line.includes('getElementById')){            
        let regex = /getElementById\((['"])(.+?)\1\)/;
        let match = line.match(regex);

        if (match) {
            let idSelector = "#" + match;
            functions[f]['selectors'].push(idSelector);
        }       
      }      
      
  }  
  
  console.log("functions:");
  console.log(functions);  
  
  return;

}




testExtract();